-- -----------------------------------------------------------------------------
--------------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- CircularBuffer class
--------------------------------------------------------------------------------

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_CircularBuffer.lua" )

-- Meta class
-- head and tail always increase with the exception being when they reach math.maxinteger
-- rather than wrapping each time head or tail exceeds the max size of the buffer.
-- by simply increasing these and only wrapping at maxinteger we simplify some of the logic:
-- - inserts always happen at head % size
-- - empty is only when tail == head (no modulus necessary)
-- - full is when:
--   - head - tail >= size
-- - we don't need to store a count because it's easily derivable from head - tail
-- - when we reach maxInteger, we just subtract maxInteger from both head and tail, no
--   wrapping. This works even if tail is negative because in Lua the % operator is
--   actual modulo, not just remainder and will always be positive.
local CircularBuffer = { }
CircularBuffer.__index = CircularBuffer
CircularBuffer.maxInteger = bit.lshift( 1, 31 )

-- Derived class method new
function CircularBuffer:print( text )
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( text );
	end
end

function CircularBuffer:new( maxSize_ )
	-- disallow trailing nil parameters
	if ( maxSize_ == nil ) then
		self:print( "maxSize_ == nil!")
		return nil
	end

	local b = setmetatable( { head = 0, tail = 0, maxSize = maxSize_ }, CircularBuffer )
	return b
end

function CircularBuffer:incIndex( index )
	-- only wrap if we are at maximum integer value already
	if ( index == self.maxInteger ) then
		index = 0
	else
		index = index + 1
	end
	return index
end

function CircularBuffer:boolToStr( value )
	if ( not value ) then
		return "false"
	else
		return "true"
	end
end

function CircularBuffer:nilToStr( value )
	if ( value == nil ) then
		return "nil"
	else
		return value
	end
end

function CircularBuffer:count()
	return self.head - self.tail
end

function CircularBuffer:isFull()
	return self:count() >= self.maxSize
end

function CircularBuffer:isEmpty()
	return self.head - self.tail == 0
end

function CircularBuffer:add( val )
	if ( self:isFull() ) then
		return false
	end

	local index = self.head % self.maxSize
	self[index] = val

	if ( self.head == self.maxInteger ) then
		self.head = self.head - self.maxInteger
		self.tail = self.tail - self.maxInteger
	end

	self.head = self.head + 1

	--HoJ:DebugMsg( "added " .. tostring( val ) .. " at " .. index )

	return true
end

function CircularBuffer:remove()
	if ( self:isEmpty() ) then
		return nil
	end

	local index = self.tail % self.maxSize
	local value = self[index]
	self[index] = nil
	self.tail = self.tail + 1
	if ( self.tail == self.maxInteger ) then
		self.tail = self.tail - maxInteger
	end
	return value
end

function CircularBuffer:empty()
	local count = self:count()
	for i = 0, count - 1 do
		local index = ( self.tail + i ) % self.maxSize
		self[index] = nil
	end

	self.head = 0
	self.tail = 0
end

-- moves the tail up by count items, marking anything that was in the shrunken area for garbage collection
function CircularBuffer:shrink( count )
	local remaining = self.maxInteger - self.tail
	local newTail
	if ( remaining < count ) then
		self.head = self.head - self.maxInteger
		newTail = ( self.tail - maxInteger ) + count
	else
		newTail = self.tail + count
	end
--[[
	HoJ:DebugMsg( "shrink: tail = " .. self.tail .. 
			", newTail = " .. newTail .. 
			", count = " .. count .. 
			", remaining = " .. remaining )

	if ( newTail >= self.head ) then
		-- make the list empty if we try to shrink past the head
		self:empty()
		HoJ:DebugMsg( "shrink -- emptied list" )
		return
	end
--]]
	-- mark everything from current tail to new tail for garbage collection
	for i = 0, count - 1 do
		local index = ( self.tail + i ) % self.maxSize
		self[index] = nil
	end

	self.tail = newTail
end

function CircularBuffer:get( index )
	local realIndex = ( self.tail + index ) % self.maxSize
	return self[realIndex]
end

HoJ.CircularBuffer = CircularBuffer