-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
-- Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Health.lua
-- DESC: Health tracking over time.

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Health.lua" )

-- -----------------------------------------------------------------------------
-- HoJ:FindStartIndexForTime
-- -----------------------------------------------------------------------------
function HoJ:FindStartIndexForTime( buffer, minTime )
	local count = buffer:count()
	if ( count == 0 ) then
		return -1
	end
	
	local left = 0
	local right = count - 1

	local mid = -1

	while( left <= right ) do
		local range = right - left
		mid = left + math.floor( range / 2 )
		local entry = buffer:get( mid )
		if ( entry._time == minTime ) then
			return mid
		elseif ( entry._time < minTime ) then
			left = mid + 1
		else 
			right = mid - 1
		end
	end

	-- we return the last mid if we get here, because we're just doing a search
	-- for the first entry > minTime
	return mid
end

-- -----------------------------------------------------------------------------
-- HoJ:ProcessBuffer
-- returns
-- numTotalEvents - total number of events in the buffer
-- numRelevantEvents - total number of events within the interval
-- totalAmount - total amount of all the events
-- totalTimeSpan - total time between the earliest and latest events
-- -----------------------------------------------------------------------------
function HoJ:ProcessBuffer( buffer, curTime, duration )
	
	if ( buffer == nil ) then
		return 0, 0, 0, 0
	end
	
	local count = buffer:count()
	if ( count == 0 ) then
		return 0, 0, 0, 0
	end
	
	local startTime = math.huge
	local endTime = -1.0
	local totalAmount = 0
	local numEvents = 0
	local minTime = curTime - duration

	local startIndex = HoJ:FindStartIndexForTime( buffer, minTime )
	if ( startIndex >= 0 ) then
--[[
		if ( HoJ._debugHealth ) then
			HoJ:DebugMsg( "startIndex = " .. startIndex .. ", count = " .. count )
		end
--]]
		for i = startIndex, count - 1 do
--[[
			if ( HoJ._debugHealth ) then
				HoJ:DebugMsg( "i = " .. i )
			end
--]]
			local entry = buffer:get( i )
			if ( entry == nil ) then
				HoJ:DebugMsg( "entry " .. i .. " == nil, count = " .. count )
			elseif ( entry._time >= minTime ) then
--[[
				if ( HoJ._debugHealth ) then
					HoJ:DebugMsg( "minTime = " .. minTime .. ", entry._time = " .. entry._time )
				end
--]]
				numEvents = numEvents + 1
				totalAmount = totalAmount + entry._amount 

				if ( entry._time < startTime ) then
					startTime = entry._time
				end

				if ( entry._time > endTime ) then
					endTime = entry._time
				end
			end
		end
	end

	local totalTimeSpan = endTime - startTime

	return count, numEvents, totalAmount, totalTimeSpan
end

-- -----------------------------------------------------------------------------
-- HoJ:CalcPerSecondStats
-- returns 
-- numEvents - total number of events
-- numRelevantEvents - number of relevant events
-- totalAmount - total of all events
-- totalTimeSpan - total time space of all events
-- amountPerSec - amount per second over the time span
-- -----------------------------------------------------------------------------
function HoJ:CalcPerSecondStats( buffer, curTime, duration )
	local numEvents, numRelevantEvents, totalAmount, totalTimeSpan = 
		self:ProcessBuffer( buffer, curTime, duration )
	local amountPerSec = 0.0
	if ( totalTimeSpan > 0 ) then
		amountPerSec = totalAmount / totalTimeSpan
	end
	return numEvents, numRelevantEvents, totalAmount, totalTimeSpan, amountPerSec
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateHealthHistory2
-- -----------------------------------------------------------------------------
function HoJ:UpdateHealthHistory2( curTime )
	-- local totalHits, relevantHits, totalDamage, totalDamageDuration, dps = 0, 0, 0.0, 0.0, 0.0
	local totalHits, relevantHits, totalDamage, totalDamageDuration, dps = 
			HoJ:CalcPerSecondStats( self._incomingDamageBuffer, curTime, 5.0 )
	if ( totalHits > relevantHits ) then
		if ( relevantHits == 0 ) then
			-- flush the buffer if nothing is relevant
			self._incomingDamageBuffer:empty()
--[[
			if ( HoJ._debugHealth ) then
				HoJ:DebugMsg( "Flushed damage buffer" )
			end
--]]
		else
			local shrinkBy = totalHits - relevantHits
			self._incomingDamageBuffer:shrink( shrinkBy )
--[[
			if ( HoJ._debugHealth ) then
				HoJ:DebugMsg( "Shrank damage buffer by " .. shrinkBy )
			end
--]]
		end
	end

	--local heals, totalHealing, totalHealsDuration, hps = 0, 0, 0.0, 0.0, 0.0
	local totalHeals, relevantHeals, totalHealing, totalHealsDuration, hps = 
			HoJ:CalcPerSecondStats( self._incomingHealsBuffer, curTime, 10.0 )
	if ( totalHeals > relevantHeals ) then
		if ( relevantHeals == 0 ) then
			-- flush the buffer if nothing is relevant
			self._incomingHealsBuffer:empty()
--[[
			if ( HoJ._debugHealth ) then
				HoJ:DebugMsg( "Flushed heals buffer" )
			end
--]]
		else
			local shrinkBy = totalHeals - relevantHeals
			self._incomingHealsBuffer:shrink( shrinkBy )
--[[
			if ( HoJ._debugHealth ) then
				HoJ:DebugMsg( "Shrank heals buffer by " .. shrinkBy )
			end
--]]
		end
	end
	
	return totalHits, relevantHits, totalDamage, totalDamageDuration, dps, 
			totalHeals, relevantHeals, totalHealing, totalHealsDuration, hps
end

-- -----------------------------------------------------------------------------
-- UpdateAverageDps
-- -----------------------------------------------------------------------------
function UpdateAverageDps( curTime, damagePerSec, buffer )
	local incomingDpsInfo = {
		_time = curTime,
		_dps = damagePerSec,
	}
	-- HoJ:Log( format( "Added %.2f dps, %.2f", damagePerSec, incomingDpsInfo._dps ) )
	buffer:add( incomingDpsInfo )

	-- average all of the incoming dps for up to the last couple of seconds
	local count = buffer:count()
	if ( count == 0 ) then
		HoJ:Log( "No incomingDps events" )
		return 0.0
	end

	local total = 0
	local minTime = math.huge
	local maxTime = -1.0
	for i = 0, count -1 do
		local t = buffer:get( i )
		HoJ:LogD( t == nil, "entry " .. i .. " == nil, count = " .. count )
		total = total + t._dps
		if ( t._time < minTime ) then
			minTime = t._time
		end
		if ( t._time > maxTime ) then
			maxTime = t._time
		end
	end
	local totalTime = maxTime - minTime
	local avgDps = total / totalTime
	
	-- HoJ:Log( format( "AIDPS: count = %.0f, duration = %.2f, total = %.0f, dps = %.0f", count, totalTime, total, avgDps ) )
	return avgDps
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateHealth2
-- -----------------------------------------------------------------------------
function HoJ:UpdateHealth2( curTime )
	local maxHealth = UnitHealthMax( "player" )
	local curHealth = UnitHealth( "player" )

	self._playerMaxHealth = maxHealth
	self._playerHealth = curHealth

	local startMs = HoJ.ProfileBegin2( "UpdateHealthHistory2")

	local totalHits, relevantHits, totalDamage, totalDamageDuration, damagePerSec, 
			totalHeals, relevantHeals, totalHealing, totalHealsDuration, healsPerSec =
			HoJ:UpdateHealthHistory2( curTime )
	
	HoJ.ProfileEnd( "UpdateHealthHistory2", endMs )
--[[
	if ( HoJ._debugHealth ) then
		HoJ:DebugMsg( "total hits = " .. totalHits .. 
				", hits = " .. relevantHits ..
				", damage = " .. totalDamage .. 
				", duration = " .. string.format( "%.2f", totalDamageDuration ) .. 
				", DPS = " .. string.format( "%.2f", damagePerSec ) )
		HoJ:DebugMsg( "total heals = " .. totalHeals ..
				", heals = " .. relevantHeals ..
				", healing = " .. totalHealing .. 
				", duration = " .. string.format( "%.2f", totalHealsDuration ) .. 
				", HPS = " .. string.format( "%.2f", healsPerSec ) )
	end
--]]

	self._playerHealthRatio = curHealth / maxHealth

	self._incomingDps = damagePerSec
	self._healsPerSec = healsPerSec

	self._incomingDpsRatio = damagePerSec / maxHealth
	self._healsPerSecRatio = healsPerSec / maxHealth
	
	local netIncomingDps = ( damagePerSec - healsPerSec )
	self._netIncomingDps = netIncomingDps
	
	-- HoJ:Log( format( "netIncomingDps = %.2f", netIncomingDps ) )

--[[
	if ( HoJ._inCombat ) then
		HoJ._averageIncomingDps = UpdateAverageDps( curTime, netIncomingDps, HoJ._incomingDpsBuffer )
	end
--]]

	if ( netIncomingDps > 0.0 ) then
		local timeToLive = curHealth / damagePerSec
		self._playerTimeToLive = timeToLive
		if ( self._debugHealth and totalDamage > (maxHealth * 0.05 ) ) then
			HoJ:DebugMsg( format( "_netIncomingDps = %.2f", netIncomingDps ) )
			HoJ:DebugMsg( format( "curHealth = %.2f", curHealth ) )
			local netIncomingDpsCHR = netIncomingDps / curHealth
			HoJ:DebugMsg( format( "netIncomingDpsCHR = %.2f", netIncomingDpsCHR ) )
			HoJ:DebugMsg( format( "_playerTimeToLive = %.2f / %.2f = %.2f", curHealth, damagePerSec, timeToLive ) )
			HoJ:DebugMsg( format( "totalDamage = %i, totalDamageDuration = %.2f", totalDamage, totalDamageDuration ) )
			HoJ:DebugMsg( format( "totalHits = %i, relevantHits = %i", totalHits, relevantHits ) )
		end
		local healthAboveNearDeath = curHealth - ( maxHealth * 0.1 )
		if ( healthAboveNearDeath <= 0.0 ) then
			self._playerTimeToNearDeath = 0.0
		else
			local timeToNearDeath = healthAboveNearDeath / damagePerSec
			self._playerTimeToNearDeath = timeToNearDeath
		end
	else
		self._playerTimeToLive = 999.0
		self._playerTimeToNearDeath = 999.0
	end

--[[
	if ( self._debugHealth ) then
		if ( self._playerTimeToNearDeath > 0.0 and self._playerTimeToNearDeath < 999.0 ) then
			self:DebugMsg( string.format( "Player time to near death = %.2f seconds", self._playerTimeToNearDeath ) )
		end
		if ( self._playerTimeToLive > 0.0 and self._playerTimeToLive < 999.0 ) then
			self:DebugMsg( string.format( "Player time to live = %2.f seconds", self._playerTimeToLive ) )
		end
		if ( damagePerSec > 0 or healsPerSec > 0 ) then
			self:DebugMsg( string.format( "Incoming DPS = %.2f, HPS = %.2f", damagePerSec, healsPerSec ) )
		end
		if ( damagePerSec > 0 or healsPerSec > 0 ) then
			self:DebugMsg( "PMH: " .. maxHealth .. 
			string.format( ", DPSR = %.1f%%, HPSR = %.1f%%", self._incomingDpsRatio * 100, self._healsPerSecRatio * 100 ) )
		end
	end
--]]
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetHealth
-- -----------------------------------------------------------------------------
function HoJ:ResetHealth()
	self._healthBuffer:empty()
	
	self._incomingDamageBuffer:empty()
	self._incomingHealsBuffer:empty()
	-- self._incomingDpsBuffer:empty()
 
	self._incomingDps = 0.0
	self._incomingDpsRatio = 0.0
	self._healsPerSec = 0.0
	self._healsPerSecRatio = 0.0

	-- self._averageIncomingDps = 0.0
	self._netIncomingDps = 0.0
--[[
	self._healthBuffer:add( { time = GetTime(), health = UnitHealth( "player" ) } )
	local h = self._healthBuffer:remove()
	if ( h ~= nil ) then
		self:DebugMsg( "h.time = " .. h.time )
		self:DebugMsg( "h.health = " .. h.health )
	else
		self:DebugMsg( "h == nil!" )
	end
]]
end
