-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
-- Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Messaging.lua
-- DESC: Chat and addon messaging functionality.

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Messaging.lua" )

-- -----------------------------------------------------------------------------
-- HoJ_Filter_Whisper
-- -----------------------------------------------------------------------------
function HoJ_Filter_Whisper( self, event, msg, player, _, _, _, gm )
	local blocked = false;
	if ( HoJ._blockedNames[ player ] ~= nil ) then
		blocked = true;
	end

	if ( blocked == true ) then
		-- HoJ:Print( type( HoJ._ignoreHistory[ player ] ) );
		if ( HoJ._ignoreHistory[ player ] == nil ) then
			HoJ._ignoreHistory[ player ] = { };
		end
		table.insert( HoJ._ignoreHistory[ player ], msg );
		if ( HoJ._silentIgnore == false ) then
			if ( HoJ._stealthIgnore == true ) then
				HoJ:Print( HoJ._stealthIgnoreMessage );
			else
				HoJ:Print( "Whisper from " .. player .. " blocked." );
			end
		end
		return true;
	end
	-- do default behavior
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ:AddMessageFilter
-- -----------------------------------------------------------------------------
function HoJ:AddMessageFilter()
    ChatFrame_AddMessageEventFilter( "CHAT_MSG_WHISPER", HoJ_Filter_Whisper );
end

-- -----------------------------------------------------------------------------
-- HoJ:InitAddonMessaging
-- -----------------------------------------------------------------------------
function HoJ:InitAddonMessaging()
	if ( not C_ChatInfo.IsAddonMessagePrefixRegistered( HoJ._addonPrefix ) ) then
		local success = C_ChatInfo.RegisterAddonMessagePrefix( HoJ._addonPrefix )
		if ( success ) then
			HoJ_MainFrame:RegisterEvent( "CHAT_MSG_ADDON" )
			HoJ:ChannelInfo( "Joined addon versioning channel: " .. HoJ._addonPrefix )
		else
			HoJ:ChannelError( "Error joining addon versioning channel." )
		end
	end
	HoJ._lastUpdateMessageTime = 0.0
end

-- -----------------------------------------------------------------------------
-- HoJ:SendAddonMessageToCustomChannel
-- -----------------------------------------------------------------------------
function HoJ:SendAddonMessageToCustomChannel( channelId, msg )
	local success = false
	local index = GetChannelName( channelId )
	if ( index ~= nil ) then
		success = C_ChatInfo.SendAddonMessage( HoJ._addonPrefix, msg, "CHANNEL", index )
		if ( success ) then
			HoJ:ChannelInfo( format( "Successfully sent addon message to channel %s(%i)", tostring( channelId ), index ) )
		else
			HoJ:ChannelInfo( format( "Failed to send message to channel %s(%i)", tostring( channelId ), index ) )
		end
	else
		HoJ:ChannelInfo( "Could not find channel " .. channelName )
	end
	return success
end

-- -----------------------------------------------------------------------------
-- HoJ:SendAddonMessage
-- -----------------------------------------------------------------------------
function HoJ:SendAddonMessage( msg )
	if ( not C_ChatInfo.IsAddonMessagePrefixRegistered( HoJ._addonPrefix ) ) then
		return false
	end

	local success = false

	--success = HoJ:SendAddonMessageToCustomChannel( 1, msg )
	--success = HoJ:SendAddonMessageToCustomChannel( "General", msg )
	--success = HoJ:SendAddonMessageToCustomChannel( "HOJ", msg )

	local isGuildMember = IsGuildMember( "player" )
	if ( isGuildMember ) then
		success = C_ChatInfo.SendAddonMessage( HoJ._addonPrefix, msg, "GUILD" )
		if ( not success ) then
			HoJ:ChannelInfo( "Failed to send message to addon channel GUILD." )
			return false
		end
	end

	local isInRaid = IsInRaid()
	if ( isInRaid ) then
		success = C_ChatInfo.SendAddonMessage( HoJ._addonPrefix, msg, "RAID" )
		if ( not success ) then
			HoJ:ChannelInfo( "Failed to send message to addon channel RAID." )
			return false
		end
	end

	local isBattleground = C_PvP.IsBattleground()
	if ( isBattleground ) then
		success = C_ChatInfo.SendAddonMessage( HoJ._addonPrefix, msg, "BATTLEGROUND" )
		if ( not success ) then
			HoJ:ChannelInfo( "Failed to send message to addon channel BATTLEGROUND." )
			return false
		end
	end

	return success
end

-- -----------------------------------------------------------------------------
-- HoJ:BroadcastVersion
-- -----------------------------------------------------------------------------
function HoJ:BroadcastVersion()
	if ( self._versionBroadcastEnabled == false ) then
		return
	end

	if ( self._curTime >= self._versionBroadcastNextTime ) then
		local versionMsg = "version = " .. HoJ.MAJOR_VERSION .. "." .. HoJ.MINOR_VERSION .. "." .. HoJ.BUILD_VERSION
		if ( HoJ:SendAddonMessage( versionMsg ) ) then
			HoJ:ChannelInfo( "Sent HOJ version info" )
		end
		self._versionBroadcastNextTime = self._curTime + self._versionBroadcastInterval
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:HandleVersionBroadcast
-- -----------------------------------------------------------------------------
function HoJ:HandleVersionBroadcast( parsed )
	if type( parsed ) ~= "table" then
		HoJ:ChannelError( "parsed parameter is not a table." )
	end

	if ( parsed[1] == "version" and parsed[2] == "=" ) then
		if ( self._versionChecksEnabled == false ) then
			return
		end

		local ver = HoJ:SplitString( parsed[3], "." )
		local major = tonumber( ver[1] )
		local minor = tonumber( ver[2] )
		local build = tonumber( ver[3] )
		HoJ:ChannelInfo( "Got version " .. major .. "." .. minor .. "." .. build )

		local upgrade = false
		if ( major > HoJ.MAJOR_VERSION ) then
			upgrade = true
		end
		if ( major == HoJ.MAJOR_VERSION and minor > HoJ.MINOR_VERSION ) then
			upgrade = true
		end
		if ( major == HoJ.MAJOR_VERSION and minor == HoJ.MINOR_VERSION and build > HoJ.BUILD_VERSION ) then
			upgrade = true
		end

		if ( upgrade ) then
			if ( self._versionMessageShown == false or self._curTime >= self._versionMessageNextTime ) then
				HoJ:Print( HoJ.MSG_PREFIX .. "a new version is available. Download it from curseforge.com or wow-interface.com." )
				self._versionMessageShown = true
				self._versionMessageNextTime = self._curTime + self._versionMessageInterval
			end
		end
	else
		HoJ:ChannelError( "Malformed version received." )
		for i, v in ipairs( parsed ) do
			HoJ:ChannelError( "parsed[" .. i .. "] = '" .. v .. "'" )
		end
	end
end
