-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Rogue.lua
-- DESC: Rogue-specific functions and tables.

--[[
TODO:
- many rogue skills seem to rely on combos. For instance, Symbols of Death + Shadow Dance or
  Vanish + Shadowstrike. There's not really a great way of chaining these right nowwithout 
  ensuring two rules sync up.
- Pooling. Need to wait for energy to pool for skills that aren't chosen currently but should 
  be chosen next. We currently end up with Backstab appearing in these gaps when we should just 
  be waiting to pool energy.
-- feint when taking AoE damage. Unforutunately it looks like we cannot determine AoE damage without a huge list of spells

BUGS:

NOTES:
--]]

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Rogue.lua" )

local rogueSpellInfo = HoJ._rogueSpellInfos

-- -----------------------------------------------------------------------------
-- HoJ:Outlaw_Vanish_Succeeded
-- -----------------------------------------------------------------------------
function HoJ:Outlaw_Vanish_Succeeded( rule )
    HoJ:Log( format( "Rule '%s' succeeded, threat is %s", 
        HoJ:GetRuleDebugName( rule ), 
        tostring( HoJ._threatTypeNames[HoJ._threat] ) ) )

    local buffTimeRemaining = HoJ:GetUnitBuffTimeRemaining( HoJ._curTime, "target", rogueSpellInfo.cheapShot.id, "PLAYER|HARMFUL" )
    HoJ:Log( format( "Cheapshot time remaining = %.1f", buffTimeRemaining ) )
end

-- -----------------------------------------------------------------------------
-- HoJ:Outlaw_BladeRush_Failed
-- -----------------------------------------------------------------------------
function HoJ:Outlaw_BladeRush_Failed( rule )
    HoJ:Log( format( "_power = %.1f, _powerMax = %.1f, overhead = %.1f, Blade Rush Energy = %.1f",
        self._power, 
        self._curSpecInfo._powerMax, 
        self._curSpecInfo._powerMax - self._power,
        rule.spellInfo._powerGen ) )
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_Garrote_Failed
-- -----------------------------------------------------------------------------
function HoJ:Rogue_Garrote_Failed( rule )
    local remaining = HoJ:GetUnitBuffTimeRemaining( HoJ._curTime, "target", rogueSpellInfo.garrote.id, "PLAYER|HARMFUL" )
    local gcdRemaining = HoJ:GetSpellCooldown( rogueSpellInfo.globalCooldown )

    HoJ:Log( format( "%s FAILED: _power = %.1f, _powerMax = %.1f, overhead = %.1f, Garrote Energy = %.1f, buff remainig = %.1f",
            HoJ:GetRuleDebugName( rule ),
            self._power, 
            self._curSpecInfo._powerMax, 
            self._curSpecInfo._powerMax - self._power,
            rule.spellInfo._powerGen,
            remaining ) )
    
    HoJ:Log( format( "Garotte buff remainig = %.1f, GCD remaining = %.1f",
            remaining, gcdRemaining ) )
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_Garotte_Succeeded
-- -----------------------------------------------------------------------------
function HoJ:Rogue_Garotte_Succeeded( rule )
    local remaining = HoJ:GetUnitBuffTimeRemaining( HoJ._curTime, "target", rogueSpellInfo.garrote.id, "PLAYER|HARMFUL" )
    local gcdRemaining = HoJ:GetSpellCooldown( rogueSpellInfo.globalCooldown )

    HoJ:Log( format( "%s SUCCEEDED: Garotte buff remainig = %.1f, GCD remaining = %.1f",
            HoJ:GetRuleDebugName( rule ), remaining, gcdRemaining ) )
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckBuff
--
-- returns
--   buffUsable: true if the buff can be used
--   hasBuff: true if the player has the buff
-- -----------------------------------------------------------------------------
function HoJ:CheckBuff( spellInfo, activeBlessingsList )
    local usable = HoJ._level >= spellInfo.minLevel and not C_SpellBook.IsSpellDisabled( spellInfo.id )
    local hasBuff = false
    if ( usable ) then
        local buff = activeBlessingsList[spellInfo.id]
        if ( buff ) then
--            if ( spellInfo.id == rogueSpellInfo.woundPoison.id ) then
--                HoJ:Log( format( "buff.start = %.2f", buff.start ) )
--            end
            hasBuff = buff.start > 0
        end
    end
--    if ( spellInfo.id == rogueSpellInfo.woundPoison.id ) then
--        HoJ:Log( format( "wp usable = %s, hasBuff = %s ", tostring( usable ), tostring( hasBuff ) ) )
--    end
    return usable, hasBuff
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_CheckPoisons
-- -----------------------------------------------------------------------------
function HoJ:Rogue_CheckPoisons( activeBlessingsList )
    if ( not self._showBuffsFrame ) then
        return
    end

    -- this has to be called every time we check blessings to update non-party member blessings
    if ( self._updateFrameNum >= self._nextUpdateBlessingFrame ) then
        self:UpdateBlessings( HoJ._curTime, activeBlessingsList )
    end

    -- self:DebugMsg( "CheckPoisons" )

    -- dont show any poison buffs for a second after entering world
    if ( self._updateFrameNum < self._enteredWorldUpdateFrameNum + HoJ.UPDATES_PER_SEC ) then
        return
    end

    -- check lethal poisons
    local instantPoisonUsable, hasInstantPoisonBuff = self:CheckBuff( rogueSpellInfo.instantPoison, activeBlessingsList )
    local woundPoisonUsable, hasWoundPoisonBuff = self:CheckBuff( rogueSpellInfo.woundPoison, activeBlessingsList )
    local deadlyPoisonUsable, hasDeadlyPoisonBuff = self:CheckBuff( rogueSpellInfo.deadlyPoison, activeBlessingsList )

    -- check non-lethal poisons
    local numbingPoisonUsable, hasNumbingPoisonBuff = self:CheckBuff( rogueSpellInfo.numbingPoison, activeBlessingsList )
    local cripplingPoisonUsable, hasCripplingPoisonBuff = self:CheckBuff( rogueSpellInfo.cripplingPoison, activeBlessingsList )

    --[[
    self:Log( format( "ip = %s, cp = %s, wp = %s, np = %s, dp = %s", 
        tostring( hasInstantPoisonBuff ),
        tostring( hasCripplingPoisonBuff ),
        tostring( hasWoundPoisonBuff ),
        tostring( hasNumbingPoisonBuff ),
        tostring( hasDeadlyPoisonBuff ) ) )
    self:Log( format( "ip = %s, cp = %s, wp = %s, np = %s, dp = %s", 
        tostring( instantPoisonUsable ),
        tostring( cripplingPoisonUsable ),
        tostring( woundPoisonUsable ),
        tostring( numbingPoisonUsable ),
        tostring( deadlyPoisonUsable ) ) )
    --]]
        
    if ( not hasInstantPoisonBuff and not hasWoundPoisonBuff and not hasDeadlyPoisonBuff ) then
        if ( self._curSpecInfo._spec == HoJ.SPEC_ROGUE_ASSASSINATION ) then
            HoJ:AddIcon( self._cooldowns, rogueSpellInfo.deadlyPoison )
        else
            HoJ:AddIcon( self._cooldowns, rogueSpellInfo.instantPoison )
        end
    end

    if ( ( not hasCripplingPoisonBuff and not hasNumbingPoisonBuff ) and 
        ( cripplingPoisonUsable or numbingPoisonUsable ) ) then
        HoJ:AddIcon( self._cooldowns, rogueSpellInfo.cripplingPoison )
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_SelectSingleOrMultitarget
-- returns
--   rotationRules
--   interruptRules
--   mitigationRules
--   dpsCooldownRules
-- -----------------------------------------------------------------------------
function HoJ:Rogue_SelectSingleOrMultitarget( curSpecInfo, numAttackers )
    if ( self._numMeleeAttackers > curSpecInfo._multiTargetThreshold ) then
        return curSpecInfo._rotationRules._multiTarget, 
                curSpecInfo._interruptRules._multiTarget,
                curSpecInfo._mitigationRules._multiTarget,
                curSpecInfo._dpsCooldownRules._multiTarget
    else
        return curSpecInfo._rotationRules._singleTarget, 
                curSpecInfo._interruptRules._singleTarget,
                curSpecInfo._mitigationRules._singleTarget,
                curSpecInfo._dpsCooldownRules._singleTarget
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_Update
-- -----------------------------------------------------------------------------
function HoJ:Rogue_Update( time, curSpecInfo, updateCooldowns, updateBuffs, updateRotation )
    -- HoJ:DebugMsg( "Updating spec " .. specInfo._name );
    local rotationRules, interruptRules, mitigationRules, dpsCooldownRules = curSpecInfo._selectSingleOrMultiTargetFn( 
        HoJ, curSpecInfo,
        self._numMeleeAttackers )

    local rotationState = HoJ:ConstructCurrentRotationState( HoJ._curTime )

    -- update buffs    
    if ( updateBuffs ) then
        local startBuffsMs = HoJ.ProfileBegin2( "Buffs" )
        
        HoJ:ClearIconList( HoJ._buffs );

        HoJ:CheckProcdBuffs( rotationState, curSpecInfo._activeBuffList, curSpecInfo._procdBuffList );

        HoJ.ProfileFunction( "UpdateIcons2:Buffs2:", HoJ.UpdateIcons2, self, time, HoJ._buffs, 
        curSpecInfo._buffSortPriorities, HoJ._buffFrames, true )
        
        HoJ.ProfileEnd( "Buffs", startBuffsMs )        
    end
    
    -- update cooldowns
    if ( updateCooldowns ) then
        local startCooldownsMs = HoJ.ProfileBegin2( "Cooldowns" )
        
        HoJ.ProfileFunction( "ClearIconList", HoJ.ClearIconList, self, HoJ._cooldowns )
        
        HoJ.ProfileFunction( "CheckRules(dpsCooldowns)", HoJ.CheckRules, self, rotationState,
                self._cooldowns, self._showCooldownsFrame, dpsCooldownRules, true, true )
        HoJ.ProfileFunction( "CheckRules(interrupts)", HoJ.CheckRules, self, rotationState,
                self._cooldowns, self._showCooldownsFrame, interruptRules, true, true )
        HoJ.ProfileFunction( "CheckRules(mitigations)", HoJ.CheckRules, self, rotationState,
                self._cooldowns, self._showCooldownsFrame, mitigationRules, true, true )
        if ( curSpecInfo._checkBlessingsFn ~= nil ) then
            curSpecInfo._checkBlessingsFn( self, curSpecInfo._activeBlessings )
        end
        
        HoJ.ProfileFunction( "UpdateIcons2:Cooldowns:", HoJ.UpdateIcons2, self, time, self._cooldowns, 
                curSpecInfo._cooldownSortPriorities, self._cooldownFrames, true )

        HoJ.ProfileEnd( "Cooldowns", startCooldownsMs )
    end

    -- update rotation
    if ( updateRotation ) then
        local startRotationMs = HoJ.ProfileBegin2( "Rotation" )

        HoJ:ClearIconList( HoJ._rotationSkills )

        HoJ:CheckDPSRotation( HoJ._rotationSkills, rotationRules )

        HoJ.ProfileFunction( "UpdateIcons2:Rotation:", HoJ.UpdateIcons2, self, time, HoJ._rotationSkills, 
                { }, HoJ._rotationFrames, false );

        HoJ.ProfileEnd( "Rotation", startRotationMs )
    end    
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_Reset_Subtlety
-- -----------------------------------------------------------------------------
function HoJ:Rogue_Reset_Subtlety()
    -- HoJ:Log( "Rogue_Reset_subtlety" )
    if ( self:IsTalentSelected( rogueSpellInfo.deeperStrategem ) ) then
        self._curSpecInfo._comboMax = 6
    else
        self._curSpecInfo._comboMax = 5
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_Reset_Assassination
-- -----------------------------------------------------------------------------
function HoJ:Rogue_Reset_Assassination()
    -- HoJ:Log( "Rogue_Reset_Assassination" )
    if ( self:IsTalentSelected( rogueSpellInfo.deeperStrategem ) ) then
        self._curSpecInfo._comboMax = 6
    else
        self._curSpecInfo._comboMax = 5
    end

    if ( self:IsTalentSelected( rogueSpellInfo.vigor ) ) then
        self._curSpecInfo._powerMax = 170
    else
        self._curSpecInfo._powerMax = 120
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_Reset_Outlaw
-- -----------------------------------------------------------------------------
function HoJ:Rogue_Reset_Outlaw()
    -- HoJ:Log( "Rogue_Reset_subtlety" )
    if ( self:IsTalentSelected( rogueSpellInfo.deeperStrategem ) ) then
        self._curSpecInfo._comboMax = 6
    else
        self._curSpecInfo._comboMax = 5
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_CheckComboPoints
-- -----------------------------------------------------------------------------
function HoJ:Rogue_CheckComboPoints( rotationState, rule, curComboPoints )
    -- HoJ:Log( "Rogue_CheckComboPoints" )

    local spellInfo = rule.spellInfo
    local comboGen = spellInfo._comboGen 
    local specInfo = self._curSpecInfo

    if ( comboGen ~= nil and comboGen > 0 ) then
        -- test as a generator of combo points
        local totalGen = 1
        local perAttacker = spellInfo._comboGenPerAttacker 
        if ( perAttacker ~= nil ) then
            totalGen = self._numMeleeAttackers * perAttacker
            if ( totalGen > spellInfo._comboGenMax ) then
                totalGen = spellInfo._comboGenMax
            end
        end

        local totalComboPoints = self._comboPoints + totalGen
        local overflow = totalComboPoints - specInfo._comboMax
        if ( overflow > 0 ) then
            -- only use the combo generator if we're re not already 
            -- at the finisher threshold, which is always one less than max
            local finisherThreshold = specInfo._comboMax - 1
            if ( self._comboPoints >= finisherThreshold ) then
                return false
            end
        end

        return true
    elseif ( rule._minComboPoints ~= nil ) then
        -- test as a consumer of combo points
        local hasEnoughComboPoints = ( curComboPoints >= rule._minComboPoints )
--[[	
        if ( ( hasEnoughComboPoints and rule._debugSuccess ) or ( not hasEnoughComboPoints and rule._debugFail ) ) then
            HoJ:Log( format( "curComboPoints = %i, rule._minComboPoints = %i, hasEnoughComboPoints = %s",
                curComboPoints, rule._minComboPoints, tostring( hasEnoughComboPoints ) ) )
        end
--]]
        return hasEnoughComboPoints
    end

    return true
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_CheckPower
-- -----------------------------------------------------------------------------
function HoJ:Rogue_CheckPower( rotationState, rule )
    local manaCost = rule.spellInfo._manaCost
    if ( manaCost ~= nil and self._mana < manaCost ) then
        return false
    end

    local powerGen = rule.spellInfo._powerGen
    if ( powerGen == nil ) then
        powerGen = 0
    end
    if ( powerGen == 0 ) then
        -- skill does not use or generate power
        return true
    end
    
    -- NOTE: Unlike the normal HoJ:CheckPower, we don't use the max power
    -- for current power so that we can check for overcapping correctly.
    -- This means that skills that need to use pooling should not check 
    -- power at all

    local power = rotationState._power
    if ( powerGen < 0 ) then
        -- the skill uses power
        local powerUsed = - rule.spellInfo._powerGen

        -- self:LogD( rule.spellInfo.id == 85256, -- TV
        --	format( "power = %i, powerUsed = %i", power, powerUsed ) ) 
        
        return ( power >= powerUsed ) 
    end

    -- the skill generates power
    local powerAfter = power + powerGen

    local powerOver = powerAfter - self._curSpecInfo._powerMax 
    if ( powerOver <= 0 ) then
        return true
    end
    if ( rule._maxPowerWaste ~= nil ) then
        return ( powerOver <= rule._maxPowerWaste )
    end

    return false
end


--------------------------------------------------------------------------------
-- SUBTLETY
--------------------------------------------------------------------------------

-- https://www.icy-veins.com/wow/subtlety-rogue-pve-dps-rotation-cooldowns-abilities
local subtletyRotationRules = {
    _name = "subtletyRotationRules",
    {
        -- marked for death if not in combat and stealthed
        spellInfo = rogueSpellInfo.markedForDeath,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.shadowBlades,
        _skipNextSkillCheck = true,
        tests = {
            { aType = HoJ.TYPE_CONSTANT, aConst = true, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_AND },
        }
    },
    {
        -- cheapshot from stealth or if subterfuge buff is active
        -- AND if target isn't a boss 
        -- AND we're not in combat yet
        -- AND target isn't already cheap shotted
        -- we allow combo point capping in this case for the stealth opener
        spellInfo = rogueSpellInfo.cheapShot,
        _rangeCheck = true,
        _powerCheck = true,
        _comboPtCheck = false,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.subterfugeBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.cheapShot, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 9, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- shadowstrike 
        -- if stealthed and not in combat
        -- OR subterfuge buff is active and we're in combat
        spellInfo = rogueSpellInfo.shadowstrike,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.subterfugeBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
        }
    },
    {
        -- backstab 
        -- if stealthed and not in combat yet
        -- OR we are in combat and have the subterfuge buff
        -- AND we haven't cast backstab in at least two global CD's (so we have )
        -- we allow overcapping combo points in this case for the in-stealth opener
        spellInfo = rogueSpellInfo.backstab,
        _rangeCheck = true,
        _powerCheck = true,
        _comboPtCheck = false,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.subterfugeBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
        }
    },
    {
        -- marked for death if <= 1 combo point 
        spellInfo = rogueSpellInfo.markedForDeath,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_LESS_EQUAL },
        }
    },
    {
        spellInfo = rogueSpellInfo.sliceAndDice,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = rogueSpellInfo.sliceAndDice, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
        }
    },
    {
        -- rupture if the target will live longer than 1/2 of the debuff
        spellInfo = rogueSpellInfo.rupture,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.rupture, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.rupture, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- vanish if
        -- we are in combat
        -- we are not already stealthed
        -- we aren't tanking the target
        -- and if the unit is alive (otherwise we'll see a vanish show up in rotation right as it dies)
        -- and we haven't recently cast cheap shot (which may mean the enemy couldn't acquire aggro yet after an opener)
        spellInfo = rogueSpellInfo.vanish,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_UNIT_IS_DEAD, aConst = "target", bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_IS_TANKING_TARGET, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TIME_SINCE_LAST_CAST, aConst = rogueSpellInfo.cheapShot, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
       
        _rangeCheck = true, -- symbols of death if > 50 power (pooling for casting shadow dance afterwards)
        spellInfo = rogueSpellInfo.symbolsOfDeath,
        _skipNextSkillCheck = true,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 50, op = HoJ.OP_GREATER_EQUAL },
        }
    },
    {
        -- shadow dance if the symbols of death CD is > 4 seconds away (either we cannot cast it soon OR we just cast it) and not in stealth
        -- or if we are near to capping shadow dance charges
        spellInfo = rogueSpellInfo.shadowDance,
        _rangeCheck = true,
        _skipNextSkillCheck = true,
        tests = {
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = rogueSpellInfo.symbolsOfDeathBuff, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.secretTechnique,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
    },
    {
        -- eviscerate with deeper strategem
        spellInfo = rogueSpellInfo.eviscerate,
        _debugName = "eviscerate_with_deeper_strategem",
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = rogueSpellInfo.eviscerate,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = rogueSpellInfo.shadowstrike,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = rogueSpellInfo.backstab,
        _rangeCheck = true,
        _powerCheck = false,
    },
    {
        spellInfo = rogueSpellInfo.shurikenToss,
        _rangeCheck = true,
        _powerCheck = false,
    },
}

local subtletyMultiTargetRotationRules = {
    _name = "subtletyMultitargetRotationRules",
    {
        -- slice and dice if < 6 targets
        spellInfo = rogueSpellInfo.sliceAndDice,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 6, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = rogueSpellInfo.sliceAndDice, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- rupture with < 5 targets and if the target's lifetime * 2 is > the buff time
        spellInfo = rogueSpellInfo.rupture,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.rupture, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.rupture, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- symbols of death if > 50 power
        spellInfo = rogueSpellInfo.symbolsOfDeath,
        _rangeCheck = true,
        _skipNextSkillCheck = true,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 50, op = HoJ.OP_GREATER_EQUAL },
        }
    },
    {
        -- shadow dance if the symbols of death buff is active and not in stealth
        spellInfo = rogueSpellInfo.shadowDance,
        _rangeCheck = true,
        _skipNextSkillCheck = true,
        tests = {
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = rogueSpellInfo.shadowDanceBuff, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_GREATER_EQUAL },
            -- { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.symbolsOfDeath, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.secretTechnique,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
    },
    {
        -- shadow strike if stealthed or shadow dance is active and we have less than 3 targets
        spellInfo = rogueSpellInfo.shadowstrike,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.shurikenStorm,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
        }
    },
    {
        -- black powder with 4+ deeper strategem
        spellInfo = rogueSpellInfo.blackPowder,
        _debugName = "blackPowder_with_deeper_strategem",
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- black powder with 4+ targets
        spellInfo = rogueSpellInfo.blackPowder,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- eviscerate with deeper strategem
        spellInfo = rogueSpellInfo.eviscerate,
        _debugName = "eviscerate_with_deeper_strategem",
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.eviscerate,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.shurikenStorm,
        _rangeCheck = true,
        _powerCheck = true,
    },
    {
        -- backstab if pooling and in range
        spellInfo = rogueSpellInfo.backstab,
        _rangeCheck = true,
        _powerCheck = false, -- for pooling
    },
    {
        -- shuriken toss if out of range
        spellInfo = rogueSpellInfo.shurikenToss,
        _rangeCheck = true,
        _powerCheck = false, -- for pooling
    },
}

local subtletyMitigationRules = {
    _name = "subtletyMitigationRules",
    {
        spellInfo = rogueSpellInfo.cloakOfShadows,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        -- vanish if about to die and not already stealthed
        spellInfo = rogueSpellInfo.vanish,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.evasion,
        -- _debugName = "evasion",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    -- TODO: feint when taking AoE damage
    {
        spellInfo = rogueSpellInfo.crimsonVial,
        _debugName = "crimsonVial",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 1, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 10.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = rogueSpellInfo.kidneyShot, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
        }
    },
}

local subtletyInterruptRules = {
    _name = "subtletyInterruptRules",
    {
        spellInfo = rogueSpellInfo.kick,
        _debugName = "kick",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.kick.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.kidneyShot,
        --_debugName = "kidneyShot",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.kidneyShot.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = rogueSpellInfo.kidneyShot, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

local subtletyDpsCooldownRules = {
    _name = "subtletyDpsCooldownRules",
    {
        spellInfo = rogueSpellInfo.shurikenTornado,
        _rangeCheck = true,
        _powerCheck = true,
    },
    {
        -- show shadow blades in the dps cooldowns if we're in multitarget, since it's discresionary
        spellInfo = rogueSpellInfo.shadowBlades,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
        }
    },
    {
        -- stealth if we're not in combat and not stealthed
        spellInfo = rogueSpellInfo.stealth,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- sap  if we're stealthed and not in combat and target is sappable
        spellInfo = rogueSpellInfo.sap,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_SPELL_CREATURE_TYPES, aConst = rogueSpellInfo.sap, bType = HoJ.TYPE_TARGET_CREATURE_TYPE, bConst = 1, op = HoJ.OP_TABLE_CONTAINS },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
}

local subtletyCooldownSortPriorities = {
    _name = "subtletyCooldownSortPriorities",
    [rogueSpellInfo.cloakOfShadows.id] = 13,
    [rogueSpellInfo.shurikenTornado.id] = 12,
    [rogueSpellInfo.stealth.id] = 11,
    [rogueSpellInfo.shadowBlades.id] = 10,
    [rogueSpellInfo.evasion.id] = 9,
    [rogueSpellInfo.crimsonVial.id] = 8,
    [rogueSpellInfo.kick.id] = 7,
    [rogueSpellInfo.kidneyShot.id] = 6,
    [rogueSpellInfo.blind.id] = 5,
    [rogueSpellInfo.feint.id] = 4,
    [rogueSpellInfo.sap.id] = 3,
    [rogueSpellInfo.cripplingPoison.id] = 2,
    [rogueSpellInfo.instantPoison.id] = 1,
}

local subtletyActiveBuffList = { 
    [rogueSpellInfo.subterfugeBuff.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.shadowDanceBuff.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.sliceAndDice.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.symbolsOfDeathBuff.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.cripplingPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.instantPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.woundPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.numbingPoison.id] = HoJ:ConstructActiveBuff(),
}

-- this is the list of buffs that we want to show on the buffs bar
-- in general we just show buffs that proc randomly here, or buffs with a stack count that needs tracking
local subtletyProcdBuffList = {
    { _spellInfo = rogueSpellInfo.subterfugeBuff },
    { _spellInfo = rogueSpellInfo.sliceAndDice },
    { _spellInfo = rogueSpellInfo.shadowDanceBuff },
    { _spellInfo = rogueSpellInfo.symbolsOfDeathBuff },
    { _spellInfo = rogueSpellInfo.cripplingPoison },
    { _spellInfo = rogueSpellInfo.instantPoison },
    { _spellInfo = rogueSpellInfo.woundPoison },
    { _spellInfo = rogueSpellInfo.numbingPoison },
}

local subtletyActiveBlessingList = {
    [rogueSpellInfo.instantPoison.id] = {
        spellInfo = rogueSpellInfo.instantPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.woundPoison.id] = {
        spellInfo = rogueSpellInfo.woundPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.numbingPoison.id] = {
        spellInfo = rogueSpellInfo.numbingPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.cripplingPoison.id] = {
        spellInfo = rogueSpellInfo.cripplingPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
}

local subtletyBuffSortPriorities = {
    _name = "subtletyBuffSortPriorities",
    [rogueSpellInfo.subterfugeBuff.id] = 8,
    [rogueSpellInfo.shadowDanceBuff.id] = 7,
    [rogueSpellInfo.symbolsOfDeathBuff.id] = 6,
    [rogueSpellInfo.sliceAndDice.id] = 5,
    [rogueSpellInfo.numbingPoison.id] = 4,
    [rogueSpellInfo.instantPoison.id] = 3,
    [rogueSpellInfo.cripplingPoison.id] = 2,
    [rogueSpellInfo.woundPoison.id] = 1,
}

--------------------------------------------------------------------------------
-- ASSASSINATION
--------------------------------------------------------------------------------

-- https://www.icy-veins.com/wow/assassination-rogue-pve-dps-rotation-cooldowns-abilities
local assassinationRotationRules = {
    _name = "assassinationRotationRules",
    {
        -- marked for death if not in combat and stealthed
        spellInfo = rogueSpellInfo.markedForDeath,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- cheapshot from stealth or if subterfuge buff is active
        -- AND if target isn't a boss 
        -- AND we're not in combat yet
        -- we allow combo point capping in this case for the stealth opener
        spellInfo = rogueSpellInfo.cheapShot,
        _rangeCheck = true,
        _powerCheck = true,
        _comboPtCheck = false,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.subterfugeBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.cheapShot, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 9, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- garrote when stealthed and not in combat
        -- or subterfuge buff is active and in combat
        -- and garrote debuff is expired or will expire in < 2 seconds
        -- we allow combo point capping in this case for the stealth opener
        spellInfo = rogueSpellInfo.garrote,
        _rangeCheck = true,
        _powerCheck = true,
        _comboPtCheck = false,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.subterfugeBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.garrote, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        -- ambush from stealth or when subterfuge buff is up
        -- AND we're not in combat yet
        -- we allow combo point capping in this case for the stealth opener
        spellInfo = rogueSpellInfo.ambush,
        _rangeCheck = true,
        _powerCheck = true,
        _comboPtCheck = false,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.subterfugeBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
        }
    },
    {
        -- marked for death if <= 1 combo point 
        spellInfo = rogueSpellInfo.markedForDeath,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_LESS_EQUAL },
        }
    },
    {
        spellInfo = rogueSpellInfo.sliceAndDice,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = rogueSpellInfo.sliceAndDice, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
        }
    },
    {
        -- rupture if the target will live longer than 1/2 of the debuff
        spellInfo = rogueSpellInfo.rupture,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.rupture, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.rupture, bConst2 = 0.5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- garrote when not stealthed and no subterfuge buff
        -- and garrote buff time remaining is < GCD remaining * 1.5
        spellInfo = rogueSpellInfo.garrote,
        _rangeCheck = true,
        _powerCheck = true,
        -- _debugFailFn = HoJ.Rogue_Garrote_Failed,
        -- _debugSuccessFn = HoJ.Rogue_Garotte_Succeeded,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.subterfugeBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.garrote, bType = HoJ.TYPE_GCD_REMAINING, bConst = 1.5, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- vanish if subterfuge or 5 combo points
        -- and 
        --   if we aren't tanking the target
        --  and if the unit is alive (otherwise we'll see a vanish show up in rotation right as it dies)
        -- and we haven't recently used cheap shot
        spellInfo = rogueSpellInfo.vanish,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.subterfuge, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_UNIT_IS_DEAD, aConst = "target", bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_IS_TANKING_TARGET, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TIME_SINCE_LAST_CAST, aConst = rogueSpellInfo.cheapShot, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 9, bType = HoJ.TYPE_RESULT, bConst = 11, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 10, bType = HoJ.TYPE_RESULT, bConst = 12, op = HoJ.OP_AND },
        }
    },
    {
        -- vendetta on cooldown if less than 80 energy
        -- NOTE: this comes after vanish even though the Icy Veins simple rotation shows it first 
        -- because it later specifies "It is best to use your Vanish before Vendetta to maximize 
        -- your Nightstalker or Subterfuge bonus"
        spellInfo = rogueSpellInfo.vendetta,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 80, op = HoJ.OP_LESS_EQUAL }
        }
    },
    {
        -- shiv on cooldown
        spellInfo = rogueSpellInfo.shiv,
        _rangeCheck = true,
        _powerCheck = true,
    },
    {
        -- envenom with deeper strategem
        spellInfo = rogueSpellInfo.envenom,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- envenom without deeper strategem
        spellInfo = rogueSpellInfo.envenom,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- eviscerate if no spec chosen
        spellInfo = rogueSpellInfo.eviscerate,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_SPEC_CHOSEN, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- fan of knives with hidden blades > 19 stacks
        spellInfo = rogueSpellInfo.fanOfKnives,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.hiddenBlades, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_STACK_COUNT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 19, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- ambush with blindside
        spellInfo = rogueSpellInfo.ambush,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.blindsideBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- sinister strike for initial spec only (i.e. haven't chosen a spec yet)
        spellInfo = rogueSpellInfo.sinisterStrike,
        _rangeCheck = true,
        _powerCheck = false,
        tests = {
            { aType = HoJ.TYPE_SPEC_CHOSEN, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- mutilate
        spellInfo = rogueSpellInfo.mutilate,
        _rangeCheck = true,
        _powerCheck = false, -- for power pooling
    },
    {
        -- poisoned knife if everything else is out of range
        spellInfo = rogueSpellInfo.poisonedKnife,
        _rangeCheck = true,
        _powerCheck = false, -- for power pooling
    },
}

local assassinationMultiTargetRotationRules = {
    _name = "assassinationMultiTargetRotationRules",
    {
        -- rupture if the target will live longer than 1/2 of the debuff
        spellInfo = rogueSpellInfo.rupture,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.rupture, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.rupture, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.sliceAndDice,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = rogueSpellInfo.sliceAndDice, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
        }
    },
    {
        -- vanish if subterfuge or 5 combo points
        -- and 
        --   if we aren't tanking the target
        --  and if the unit is alive (otherwise we'll see a vanish show up in rotation right as it dies)
        -- and if the unit is not already cheap-shotted
        spellInfo = rogueSpellInfo.vanish,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.subterfuge, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_UNIT_IS_DEAD, aConst = "target", bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_IS_TANKING_TARGET, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TIME_SINCE_LAST_CAST, aConst = rogueSpellInfo.cheapShot, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 9, bType = HoJ.TYPE_RESULT, bConst = 11, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 10, bType = HoJ.TYPE_RESULT, bConst = 12, op = HoJ.OP_AND },
        }
    },
    {
        -- vendetta on cooldown with < 80 energy
        -- NOTE: this comes after vanish even though the Icy Veins simple rotation shows it first 
        -- because it later specifies "It is best to use your Vanish before Vendetta to maximize 
        -- your Nightstalker or Subterfuge bonus"
        spellInfo = rogueSpellInfo.vendetta,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 80, op = HoJ.OP_LESS_EQUAL }
        }
    },
    {
        -- garrote if < 5 targets
        spellInfo = rogueSpellInfo.garrote,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.garrote, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- crimson tempest with deeper strategem
        spellInfo = rogueSpellInfo.crimsonTempest,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- crimson tempest without deeper strategem
        spellInfo = rogueSpellInfo.crimsonTempest,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- envenom with deeper strategem
        spellInfo = rogueSpellInfo.envenom,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- envenom without deeper strategem
        spellInfo = rogueSpellInfo.envenom,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- fan of knives with 4+ targets
        spellInfo = rogueSpellInfo.fanOfKnives,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_GREATER_EQUAL },
        }
    },
    {
        -- mutilate
        spellInfo = rogueSpellInfo.mutilate,
        _rangeCheck = true,
        _powerCheck = false, -- for power pooling
    },
    {
        -- poisoned knife if everything else is out of range
        spellInfo = rogueSpellInfo.poisonedKnife,
        _rangeCheck = true,
        _powerCheck = false, -- for power pooling
    },
}

local assassinationMitigationRules = {
    _name = "assassinationMitigationRules",
    {
        spellInfo = rogueSpellInfo.cloakOfShadows,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        -- vanish if about to die and not already stealthed
        spellInfo = rogueSpellInfo.vanish,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.evasion,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    -- TODO: feint when taking AoE damage
    {
        spellInfo = rogueSpellInfo.crimsonVial,
        _debugName = "crimsonVial",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 1, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 10.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = rogueSpellInfo.kidneyShot, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.blind,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
}

local assassinationInterruptRules = {
    _name = "assassinationInterruptRules",
    {
        spellInfo = rogueSpellInfo.kick,
        _debugName = "kick",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.kick.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.kidneyShot,
        --_debugName = "kidneyShot",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.kidneyShot.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = rogueSpellInfo.kidneyShot, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

local assassinationDpsCooldownRules = {
    _name = "assassinationDpsCooldownRules",
    {
        -- stealth if we're not in combat and not stealthed
        spellInfo = rogueSpellInfo.stealth,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- sap  if we're stealthed and not in combat and target is sappable
        spellInfo = rogueSpellInfo.sap,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_SPELL_CREATURE_TYPES, aConst = rogueSpellInfo.sap, bType = HoJ.TYPE_TARGET_CREATURE_TYPE, bConst = 1, op = HoJ.OP_TABLE_CONTAINS },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
}

local assassinationCooldownSortPriorities = {
    _name = "assassinationCooldownSortPriorities",
    [rogueSpellInfo.cloakOfShadows.id] = 12,
    [rogueSpellInfo.stealth.id] = 11,
    [rogueSpellInfo.evasion.id] = 10,
    [rogueSpellInfo.crimsonVial.id] = 9,
    [rogueSpellInfo.kick.id] = 8,
    [rogueSpellInfo.kidneyShot.id] = 7,
    [rogueSpellInfo.blind.id] = 6,
    [rogueSpellInfo.feint.id] = 5,
    [rogueSpellInfo.sap.id] = 4,
    [rogueSpellInfo.cripplingPoison.id] = 3,
    [rogueSpellInfo.deadlyPoison.id] = 2,
    -- this is here only for when we're in initial spec, which uses assasination but
    -- isn't exactly like assassination in that the available poison is instant instead
    -- of deadly
    [rogueSpellInfo.instantPoison.id] = 1, 
}

local assassinationActiveBuffList = { 
    [rogueSpellInfo.subterfugeBuff.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.blindsideBuff.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.sliceAndDice.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.cripplingPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.deadlyPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.woundPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.numbingPoison.id] = HoJ:ConstructActiveBuff(),
}

-- this is the list of buffs that we want to show on the buffs bar
-- in general we just show buffs that proc randomly here, or buffs with a stack count that needs tracking
local assassinationProcdBuffList = {
    { _spellInfo = rogueSpellInfo.subterfugeBuff },
    { _spellInfo = rogueSpellInfo.blindsideBuff },
    { _spellInfo = rogueSpellInfo.sliceAndDice },
    { _spellInfo = rogueSpellInfo.cripplingPoison },
    { _spellInfo = rogueSpellInfo.deadlyPoison },
    { _spellInfo = rogueSpellInfo.woundPoison },
    { _spellInfo = rogueSpellInfo.numbingPoison },
}

local assassinationActiveBlessingList = {
    -- instant poison is only included for the initial spec
    [rogueSpellInfo.instantPoison.id] = {
        spellInfo = rogueSpellInfo.instantPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.deadlyPoison.id] = {
        spellInfo = rogueSpellInfo.deadlyPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.woundPoison.id] = {
        spellInfo = rogueSpellInfo.woundPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.numbingPoison.id] = {
        spellInfo = rogueSpellInfo.numbingPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.cripplingPoison.id] = {
        spellInfo = rogueSpellInfo.cripplingPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
}

local assassinationBuffSortPriorities = {
    _name = "assassinationBuffSortPriorities",
    [rogueSpellInfo.subterfugeBuff.id] = 8,
    [rogueSpellInfo.blindsideBuff.id] = 7,
    [rogueSpellInfo.sliceAndDice.id] = 6,
    [rogueSpellInfo.numbingPoison.id] = 5,
    [rogueSpellInfo.deadlyPoison.id] = 4,
    [rogueSpellInfo.instantPoison.id] = 3, -- only here for initial spec
    [rogueSpellInfo.cripplingPoison.id] = 2,
    [rogueSpellInfo.woundPoison.id] = 1,
}

--------------------------------------------------------------------------------
-- OUTLAW
--------------------------------------------------------------------------------

-- https://www.icy-veins.com/wow/outlaw-rogue-pve-dps-rotation-cooldowns-abilities
local outlawRotationRules = {
    _name = "outlawRotationRules",
    {
        spellInfo = rogueSpellInfo.rollTheBones,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.broadside, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.trueBearing, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },

            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.ruthlessPrecision, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.skullAndCrossbones, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },

            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.buriedTreasure, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.grandMelee, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },

            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 9, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- cheapshot from stealth if target isn't a boss AND we're not in combat yet
        -- we allow overcapping combo points in this case for the in-stealth opener
        spellInfo = rogueSpellInfo.cheapShot,
        _rangeCheck = true,
        _powerCheck = true,
        _comboPtCheck = false,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- ambush from stealth
        -- AND we're not in combat yet
        -- we allow overcapping combo points in this case for the in-stealth opener
        spellInfo = rogueSpellInfo.ambush,
        _rangeCheck = true,
        _powerCheck = true,
        _comboPtCheck = false,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.adrenalineRush,
        _rangeCheck = true,
        _powerCheck = true,
    },
    {
        -- dreadblades if talented and combo points <= 1
        spellInfo = rogueSpellInfo.dreadblades,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_LESS_EQUAL },
        }
    },
    {
        -- betweenTheEyes with deeper strategem
        spellInfo = rogueSpellInfo.betweenTheEyes,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- betweenTheEyes without deeper strategem
        spellInfo = rogueSpellInfo.betweenTheEyes,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = rogueSpellInfo.sliceAndDice,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = rogueSpellInfo.sliceAndDice, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.sliceAndDice,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = rogueSpellInfo.sliceAndDice, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.ghostlyStrike,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_COMBO_POINTS_DEFICIT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER_EQUAL },
        }
    },
    --[[
    {
        -- killing spree if adrenaline rush isn't active
        spellInfo = rogueSpellInfo.killingSpree,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.adrenalineRush, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- blade rush if current power is < ( power max - blade rush power gen * 0.5 )
        spellInfo = rogueSpellInfo.bladeRush,
        _rangeCheck = true,
        -- _powerCheck = true,
        -- _debugFailFn = HoJ.Outlaw_BladeRush_Failed,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_POWER_MAX_ADJUSTED, bConst = rogueSpellInfo.bladeRush._powerGen * 0.2, op = HoJ.OP_LESS_EQUAL },
        }
    },
    --]]
    {
        -- dispatch with deeper strategem
        spellInfo = rogueSpellInfo.dispatch,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- dispatch without deeper strategem
        spellInfo = rogueSpellInfo.dispatch,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- vanish if
        -- we are in combat
        -- we are not already stealthed
        --  we aren't tanking the target
        --  and if the unit is alive (otherwise we'll see a vanish show up in rotation right as it dies)
        -- and the unit isn't under the effects of cheap shot (which may mean they haven't acquired aggro yet after an opener)
        spellInfo = rogueSpellInfo.vanish,
        -- _debugSuccessFn = HoJ.Outlaw_Vanish_Succeeded,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_UNIT_IS_DEAD, aConst = "target", bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_IS_TANKING_TARGET, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TIME_SINCE_LAST_CAST, aConst = rogueSpellInfo.cheapShot, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        -- cheapshot from stealth if talented for prey on the weak, and target isn't a boss
        spellInfo = rogueSpellInfo.cheapShot,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.preyOnTheWeak, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- ambush from stealth
        spellInfo = rogueSpellInfo.ambush,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- pistol shot if we have the opportunity buff and less than 5 combo points
        spellInfo = rogueSpellInfo.pistolShot,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.opportunity, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- sinister strike
        spellInfo = rogueSpellInfo.sinisterStrike2Outlaw,
        _rangeCheck = true,
        _powerCheck = false, -- for pooling
    },
    {
        -- pistol shot if everything else is out of range
        spellInfo = rogueSpellInfo.pistolShot,
        _rangeCheck = true,
        _powerCheck = false, -- for pooling
    },
}

local outlawMultiTargetRotationRules = {
    _name = "outlawMultiTargetRotationRules",
    {
        -- blade flurry with 2+ targets
        spellInfo = rogueSpellInfo.bladeFlurry,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
        }
    },
    {
        spellInfo = rogueSpellInfo.adrenalineRush,
        _rangeCheck = true,
        _powerCheck = true,
    },
    {
        spellInfo = rogueSpellInfo.rollTheBones,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.broadside, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.trueBearing, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },

            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.ruthlessPrecision, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.skullAndCrossbones, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },

            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.buriedTreasure, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.grandMelee, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },

            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 9, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- betweenTheEyes with deeper strategem
        spellInfo = rogueSpellInfo.betweenTheEyes,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- betweenTheEyes without deeper strategem
        spellInfo = rogueSpellInfo.betweenTheEyes,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = rogueSpellInfo.sliceAndDice,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = rogueSpellInfo.sliceAndDice, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.sliceAndDice,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = rogueSpellInfo.sliceAndDice, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    --[[
    {
        -- killing spree if blade furry is active
        spellInfo = rogueSpellInfo.killingSpree,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.bladeFlurry, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- blade rush if blade furry is active and power is >= blade rush power get * 0.2 from power cap
        spellInfo = rogueSpellInfo.bladeRush,
        _rangeCheck = true,
        -- _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.bladeFlurry, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_POWER_MAX_ADJUSTED, bConst = rogueSpellInfo.bladeRush._powerGen * 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    --]]
    {
        spellInfo = rogueSpellInfo.ghostlyStrike,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_COMBO_POINTS_DEFICIT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER_EQUAL },
        }
    },
    {
        -- dispatch with deeper strategem
        spellInfo = rogueSpellInfo.dispatch,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 5,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- dispatch without deeper strategem
        spellInfo = rogueSpellInfo.dispatch,
        _rangeCheck = true,
        _powerCheck = true,
        _minComboPoints = 4,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- vanish if
        -- we are in combat
        -- we are not already stealthed
        --  we aren't tanking the target
        --  and if the unit is alive (otherwise we'll see a vanish show up in rotation right as it dies)
        -- and the unit isn't under the effects of cheap shot (which may mean they haven't acquired aggro yet after an opener)
        spellInfo = rogueSpellInfo.vanish,
        -- _debugSuccessFn = HoJ.Outlaw_Vanish_Succeeded,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_UNIT_IS_DEAD, aConst = "target", bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_IS_TANKING_TARGET, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TIME_SINCE_LAST_CAST, aConst = rogueSpellInfo.cheapShot, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        -- cheapshot from stealth if talented for prey on the weak, and target isn't a boss
        spellInfo = rogueSpellInfo.cheapShot,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.preyOnTheWeak, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- ambush from stealth
        spellInfo = rogueSpellInfo.ambush,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- pistol shot if we have the opportunity buff and less than 3 combo points
        spellInfo = rogueSpellInfo.pistolShot,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.opportunity, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- sinister strike
        spellInfo = rogueSpellInfo.sinisterStrike,
        _rangeCheck = true,
        _powerCheck = false,
    },
    {
        -- pistol shot if everything else is out of range
        spellInfo = rogueSpellInfo.pistolShot,
        _rangeCheck = true,
        _powerCheck = false,
    },
}

local outlawMitigationRules = {
    _name = "outlawMitigationRules",
    {
        spellInfo = rogueSpellInfo.cloakOfShadows,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        -- vanish if about to die and not already stealthed
        spellInfo = rogueSpellInfo.vanish,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.evasion,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = rogueSpellInfo.crimsonVial,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 1, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 10.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = rogueSpellInfo.kidneyShot, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.gouge,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.blind,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = rogueSpellInfo.grapplingHook,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
}

local outlawInterruptRules = {
    _name = "outlawInterruptRules",
    {
        spellInfo = rogueSpellInfo.kick,
        _debugName = "kick",
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = rogueSpellInfo.gouge,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.kidneyShot,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local outlawDpsCooldownRules = {
    _name = "outlawDpsCooldownRules",
    {
        -- stealth if we're not in combat and not stealthed
        spellInfo = rogueSpellInfo.stealth,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- sap  if we're stealthed and not in combat and target is sappable
        spellInfo = rogueSpellInfo.sap,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_IN_COMBAT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_SPELL_CREATURE_TYPES, aConst = rogueSpellInfo.sap, bType = HoJ.TYPE_TARGET_CREATURE_TYPE, bConst = 1, op = HoJ.OP_TABLE_CONTAINS },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        -- blade rush if 
        -- blade flurry buff is up 
        -- 2 or more attackers
        -- we're >= blade rush power gen * 0.2 away from cap
        spellInfo = rogueSpellInfo.bladeRush,
        _rangeCheck = true,
        -- _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.bladeFlurry, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_POWER_MAX_ADJUSTED, bConst = rogueSpellInfo.bladeRush._powerGen * 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- blade rush if 
        -- only 1 attacker
        -- we're >= blade rush power gen * 0.2 away from cap
        spellInfo = rogueSpellInfo.bladeRush,
        _rangeCheck = true,
        -- _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_POWER_MAX_ADJUSTED, bConst = rogueSpellInfo.bladeRush._powerGen * 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- killing spree if
        -- blade flurry buff is up
        -- adrenaline rush is not up
        -- 2 or more attackers
        spellInfo = rogueSpellInfo.killingSpree,
        _rangeCheck = true,
        --_powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.bladeFlurry, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.adrenalineRush, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- killing spree if
        -- only 1 attacker
        -- adrenaline rush is not up
        spellInfo = rogueSpellInfo.killingSpree,
        _rangeCheck = true,
        --_powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.adrenalineRush, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },

}

local outlawCooldownSortPriorities = {
    _name = "outlawCooldownSortPriorities",
    [rogueSpellInfo.cloakOfShadows.id] = 16,
    [rogueSpellInfo.stealth.id] = 15,
    [rogueSpellInfo.evasion.id] = 14,
    [rogueSpellInfo.crimsonVial.id] = 13,
    [rogueSpellInfo.kick.id] = 12,
    [rogueSpellInfo.gouge.id] = 11,
    [rogueSpellInfo.kidneyShot.id] = 10,
    [rogueSpellInfo.blind.id] = 9,
    [rogueSpellInfo.grapplingHook.id] = 8,
    [rogueSpellInfo.feint.id] = 7,
    [rogueSpellInfo.bladeRush.id] = 6,
    [rogueSpellInfo.killingSpree.id] = 5,
    [rogueSpellInfo.adrenalineRush.id] = 4,
    [rogueSpellInfo.sap.id] = 3,
    [rogueSpellInfo.cripplingPoison.id] = 2,
    [rogueSpellInfo.instantPoison.id] = 1,
}

local outlawActiveBuffList = { 
    [rogueSpellInfo.bladeFlurry.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.opportunity.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.sliceAndDice.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.cripplingPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.instantPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.woundPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.numbingPoison.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.broadside.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.trueBearing.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.ruthlessPrecision.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.skullAndCrossbones.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.buriedTreasure.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.grandMelee.id] = HoJ:ConstructActiveBuff(),
    [rogueSpellInfo.adrenalineRush.id] = HoJ:ConstructActiveBuff(),
}

-- this is the list of buffs that we want to show on the buffs bar
-- in general we just show buffs that proc randomly here, or buffs 
-- with a stack count that needs tracking
local outlawProcdBuffList = {
    { _spellInfo = rogueSpellInfo.sliceAndDice },
    { _spellInfo = rogueSpellInfo.bladeFlurry },
    -- { _spellInfo = rogueSpellInfo.broadside },
    -- { _spellInfo = rogueSpellInfo.trueBearing },
    -- { _spellInfo = rogueSpellInfo.ruthlessPrecision },
    -- { _spellInfo = rogueSpellInfo.skullAndCrossbones },
    -- { _spellInfo = rogueSpellInfo.buriedTreasure },
    -- { _spellInfo = rogueSpellInfo.grandMelee },
    { _spellInfo = rogueSpellInfo.opportunity },
    { _spellInfo = rogueSpellInfo.cripplingPoison },
    { _spellInfo = rogueSpellInfo.instantPoison },
    { _spellInfo = rogueSpellInfo.woundPoison },
    { _spellInfo = rogueSpellInfo.numbingPoison },
    { _spellInfo = rogueSpellInfo.adrenalineRush },
}

local outlawActiveBlessingList = {
    [rogueSpellInfo.instantPoison.id] = {
        spellInfo = rogueSpellInfo.instantPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.woundPoison.id] = {
        spellInfo = rogueSpellInfo.woundPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.numbingPoison.id] = {
        spellInfo = rogueSpellInfo.numbingPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [rogueSpellInfo.cripplingPoison.id] = {
        spellInfo = rogueSpellInfo.cripplingPoison,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
}

local outlawBuffSortPriorities = {
    _name = "outlawBuffSortPriorities",
    [rogueSpellInfo.adrenalineRush.id] = 14,
    [rogueSpellInfo.bladeFlurry.id] = 13,
    [rogueSpellInfo.opportunity.id] = 12,
    [rogueSpellInfo.sliceAndDice.id] = 11,
    [rogueSpellInfo.numbingPoison.id] = 10,
    [rogueSpellInfo.instantPoison.id] = 9,
    [rogueSpellInfo.cripplingPoison.id] = 8,
    [rogueSpellInfo.woundPoison.id] = 7,
    [rogueSpellInfo.broadside.id] = 6,
    [rogueSpellInfo.trueBearing.id] = 5,
    [rogueSpellInfo.ruthlessPrecision.id] = 4,
    [rogueSpellInfo.skullAndCrossbones.id] = 3,
    [rogueSpellInfo.buriedTreasure.id] = 2,
    [rogueSpellInfo.grandMelee.id] = 1,

}

HoJ[HoJ.CLASS_NAME_ROGUE] = {
    _spellInfos = HoJ._rogueSpellInfos,
    _className = HoJ.CLASS_NAME_ROGUE,
    _defaultSpecId = HoJ.SPEC_ROGUE_ASSASSINATION,
    _powerColor = { r = 1.0, g = 1.0, b = 0.0, a = 1.0 },
    _usesComboPoints = true,
    [HoJ.SPEC_NAME_ROGUE_SUBTLETY] = {
        -- properties
        _name = HoJ._text["Subtlety"],
        _spec = HoJ.SPEC_ROGUE_SUBTLETY,
        _role = HoJ.ROLE_DPS,
        _enabled = true,
        _multiTargetRange = 10,
        _multiTargetThreshold = 2,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 10, -- testing shows around 10 pps
        _autoPowerGen = true,
        _powerType = HoJ.POWER_TYPE_ENERGY,
        _powerGainStep = 1,
        _powerMax = 100,
        _comboMax = 5,
        _usesPowerPooling = true,
        -- functions
        _updateFn = HoJ.Rogue_Update,
        _updateRenderFn = nil,
        _resetFn = HoJ.Rogue_Reset_Subtlety,
        _selectSingleOrMultiTargetFn = HoJ.Rogue_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = HoJ.Rogue_CheckPoisons,
        _checkComboPointsFn = HoJ.Rogue_CheckComboPoints,
        _checkPowerFn = HoJ.Rogue_CheckPower,
        -- tables
        _meleeAoESpells = nil,
        _activeBuffList = subtletyActiveBuffList,
        _procdBuffList = subtletyProcdBuffList,
        _activeBlessings = subtletyActiveBlessingList,
        _rotationRules = { _singleTarget = subtletyRotationRules, _multiTarget = subtletyMultiTargetRotationRules },
        _interruptRules = { _singleTarget = subtletyInterruptRules, _multiTarget = subtletyInterruptRules },
        _mitigationRules = { _singleTarget = subtletyMitigationRules, _multiTarget = subtletyMitigationRules },
        _dpsCooldownRules = { _singleTarget = subtletyDpsCooldownRules, _multiTarget = subtletyDpsCooldownRules },
        _cooldownSortPriorities = subtletyCooldownSortPriorities,
        _buffSortPriorities = subtletyBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
            [HoJ.RANGE_100_YARDS] = {
                _spellInfo = rogueSpellInfo.tricksOfTheTrade,
            },
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfos = {
                    rogueSpellInfo.backstab,
                    rogueSpellInfo.sinisterStrike
                }
            },
            [HoJ.RANGE_8_YARDS] = {
                _spellInfo = rogueSpellInfo.kick,
            },
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = rogueSpellInfo.blind,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = rogueSpellInfo.blind,
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = rogueSpellInfo.shadowstep,
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = rogueSpellInfo.shurikenToss
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = rogueSpellInfo.shurikenToss
            },
            [HoJ.RANGE_100_YARDS] = {
                _spellInfo = rogueSpellInfo.shadowstep
            },
        },
    },
    [HoJ.SPEC_NAME_ROGUE_ASSASSINATION] = {
        -- properties
        _name = HoJ._text["Assassination"],
        _spec = HoJ.SPEC_ROGUE_ASSASSINATION,
        _role = HoJ.ROLE_DPS,
        _enabled = true,
        _multiTargetRange = 10,
        _multiTargetThreshold = 1,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 10, -- testing shows around 10 pps
        _autoPowerGen = true,
        _powerType = HoJ.POWER_TYPE_ENERGY,
        _powerGainStep = 1,
        _powerMax = 120,
        _comboMax = 5,
        _usesPowerPooling = true,
        -- functions
        _updateFn = HoJ.Rogue_Update,
        _updateRenderFn = nil,
        _resetFn = HoJ.Rogue_Reset_Assassination,
        _selectSingleOrMultiTargetFn = HoJ.Rogue_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = HoJ.Rogue_CheckPoisons,
        _checkComboPointsFn = HoJ.Rogue_CheckComboPoints,
        _checkPowerFn = HoJ.Rogue_CheckPower,
        -- tables
        _meleeAoESpells = nil,
        _activeBuffList = assassinationActiveBuffList,
        _procdBuffList = assassinationProcdBuffList,
        _activeBlessings = assassinationActiveBlessingList,
        _rotationRules = { _singleTarget = assassinationRotationRules, _multiTarget = assassinationMultiTargetRotationRules },
        _interruptRules = { _singleTarget = assassinationInterruptRules, _multiTarget = assassinationInterruptRules },
        _mitigationRules = { _singleTarget = assassinationMitigationRules, _multiTarget = assassinationMitigationRules },
        _dpsCooldownRules = { _singleTarget = assassinationDpsCooldownRules, _multiTarget = assassinationDpsCooldownRules },
        _cooldownSortPriorities = assassinationCooldownSortPriorities,
        _buffSortPriorities = assassinationBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
            [HoJ.RANGE_100_YARDS] = {
                _spellInfo = rogueSpellInfo.tricksOfTheTrade,
            },
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = rogueSpellInfo.sinisterStrike
            },
            [HoJ.RANGE_8_YARDS] = {
                _spellInfo = rogueSpellInfo.kick,
            },
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = rogueSpellInfo.blind,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = rogueSpellInfo.blind,
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = rogueSpellInfo.shadowstep,
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = rogueSpellInfo.poisonedKnife
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = rogueSpellInfo.poisonedKnife
            },
            [HoJ.RANGE_100_YARDS] = {
                _spellInfo = rogueSpellInfo.poisonedKnife
            },
        },
    },
    [HoJ.SPEC_NAME_ROGUE_OUTLAW] = {
        -- properties
        _name = HoJ._text["Outlaw"],
        _spec = HoJ.SPEC_ROGUE_OUTLAW,
        _role = HoJ.ROLE_DPS,
        _enabled = true,
        _multiTargetRange = 10,
        _multiTargetThreshold = 1,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 10, -- testing shows around 10 pps
        _autoPowerGen = true,
        _powerType = HoJ.POWER_TYPE_ENERGY,
        _powerGainStep = 1,
        _powerMax = 120,
        _comboMax = 5,
        _usesPowerPooling = true,
        -- functions
        _updateFn = HoJ.Rogue_Update,
        _updateRenderFn = nil,
        _resetFn = HoJ.Rogue_Reset_Outlaw,
        _selectSingleOrMultiTargetFn = HoJ.Rogue_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = HoJ.Rogue_CheckPoisons,
        _checkComboPointsFn = HoJ.Rogue_CheckComboPoints,
        _checkPowerFn = HoJ.Rogue_CheckPower,
        -- tables
        _meleeAoESpells = nil,
        _activeBuffList = outlawActiveBuffList,
        _procdBuffList = outlawProcdBuffList,
        _activeBlessings = outlawActiveBlessingList,
        _rotationRules = { _singleTarget = outlawRotationRules, _multiTarget = outlawMultiTargetRotationRules },
        _interruptRules = { _singleTarget = outlawInterruptRules, _multiTarget = outlawInterruptRules },
        _mitigationRules = { _singleTarget = outlawMitigationRules, _multiTarget = outlawMitigationRules },
        _dpsCooldownRules = { _singleTarget = outlawDpsCooldownRules, _multiTarget = outlawDpsCooldownRules },
        _cooldownSortPriorities = outlawCooldownSortPriorities,
        _buffSortPriorities = outlawBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
            [HoJ.RANGE_100_YARDS] = {
                _spellInfo = rogueSpellInfo.tricksOfTheTrade,
            },
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = rogueSpellInfo.sinisterStrike,
            },
            [HoJ.RANGE_8_YARDS] = {
                _spellInfo = rogueSpellInfo.sinisterStrike,
            },
            [HoJ.RANGE_10_YARDS] = {
                _spellInfos = {
                    rogueSpellInfo.killingSpree,
                    rogueSpellInfo.sinisterStrike
                }
            },
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = rogueSpellInfo.blind,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfos = {
                    rogueSpellInfo.pistolShot,
                    rogueSpellInfo.blind,
                }
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfos = {
                    rogueSpellInfo.markedForDeath,
                    rogueSpellInfo.pistolShot,
                }
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfos = {
                    rogueSpellInfo.markedForDeath,
                    rogueSpellInfo.pistolShot,
                }
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfos = {
                    rogueSpellInfo.markedForDeath,
                    rogueSpellInfo.pistolShot,
                }
            },
            [HoJ.RANGE_100_YARDS] = {
                _spellInfos = {
                    rogueSpellInfo.markedForDeath,
                    rogueSpellInfo.pistolShot,
                }
            },
        },
    },
}

-- -----------------------------------------------------------------------------
-- HoJ:LoadRogue
-- -----------------------------------------------------------------------------
function HoJ:LoadRogue()
    if ( HoJ._classId ~= HoJ.CLASS_ID_ROGUE ) then
        HoJ:Error( "LoadRogue called but _classId is not rogue!" )
    end

    HoJ.SpellInfo = HoJ._rogueSpellInfos
end
