-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_SavedVariables.lua
-- DESC: Addon user interface.

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_SavedVariables.lua" )

local SAVED_VARIABLES_NAME = "HoJSavedVars"

-- this table is a list of all the members of HoJ that need to be copied or restored from 
-- saved variables.
-- This allows us to use one table for all of our addon properties, but save and restore
-- only specific parts of that table. It also means we can directly modify the addon
-- properties and have them synced to saved variables without writing special-case code
-- just to sync specific variables.
local saveTable_HoJ = {
    "MAJOR_VERSION",
    "MINOR_VERSION",
    "BUILD_VERSION",
    "_blockedNames",
    "_silentIgnore",
    "_stealthIgnore",
    "_showCooldownsFrame",
    "_showBuffsFrame",
    "_showTemplarsVerdictFrame",
    "_showHolyPowerFrame",
    "_showRotationFrame",
    "_showRotationSecondaryFrame",
    "_showDebugFrame",
    "_globalAlpha",
    "_rotationAlpha",
    "_rotationSecondaryAlpha",
    "_cooldownsAlpha",
    "_buffsAlpha",
    "_debugAlpha",
    "_versionBroadcastEnabled",
    "_versionChecksEnabled",
    "_uiScale",
    [HoJ.CLASS_NAME_PALADIN] = {
        [HoJ.SPEC_NAME_PALADIN_HOLY] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        },
        [HoJ.SPEC_NAME_PALADIN_PROTECTION] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        },
        [HoJ.SPEC_NAME_PALADIN_RETRIBUTION] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        }
    },
    [HoJ.CLASS_NAME_WARRIOR] = {
        [HoJ.SPEC_NAME_WARRIOR_ARMS] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        },
        [HoJ.SPEC_NAME_WARRIOR_FURY] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        },
        [HoJ.SPEC_NAME_WARRIOR_PROTECTION] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        }
    },
    [HoJ.CLASS_NAME_ROGUE] = {
        [HoJ.SPEC_NAME_ROGUE_SUBTLETY] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        },
        [HoJ.SPEC_NAME_ROGUE_ASSASSINATION] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        },
        [HoJ.SPEC_NAME_ROGUE_OUTLAW] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        }
    },
    [HoJ.CLASS_NAME_DEMONHUNTER] = {
        [HoJ.SPEC_NAME_DEMONHUNTER_HAVOC] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        },
        [HoJ.SPEC_NAME_DEMONHUNTER_VENGEANCE] = {
            HoJ.HEAL_NAME_MINOR,
            HoJ.HEAL_NAME_MAJOR,
            "_enabled",
        },
    },
}

-- specifies which items in a spellInfo table should be saved
local saveTable_spellInfo = {
    "showOnBar"
}

-- -----------------------------------------------------------------------------
-- HoJ:CopyTableUsingSaveTable
-- -----------------------------------------------------------------------------
function HoJ:CopyTableUsingSaveTable( saveTable, sourceTable, outTable, nilDeletes )
    if ( saveTable == nil ) then
        HoJ:Error( "saveTable is nil!" )
        return false
    end
    if ( sourceTable == nil ) then
        HoJ:Error( "sourceTable is nil!" )
        return false
    end
    if ( saveTable == nil ) then
        HoJ:Error( "saveTable is nil!" )
        return false
    end

    for key, value in pairs( saveTable ) do
        local valueType = type( value )
        if ( valueType == "table" ) then
            -- HoJ:DebugMsg( "key '" .. key .. "', type 'table'" )
            if ( outTable[key] == nil ) then
                outTable[key] = { }
            end
            HoJ:CopyTableUsingSaveTable( value, sourceTable[key], outTable[key], nilDeletes )
        else
            if ( nilDeletes ) then
                --HoJ:DebugMsg( "pre outTable[" .. key .. "] = " .. tostring( outTable[key] ) )
                --HoJ:DebugMsg( "key '" .. value .. "', type '" .. valueType .. "' = " .. tostring( sourceTable[key] ) )
                outTable[value] = sourceTable[value]
                --HoJ:DebugMsg( "outTable[" .. key .. "] = " .. tostring( outTable[key] ) )
            else
                if ( sourceTable[value] ~= nil ) then
                    outTable[value] = sourceTable[value]
                -- else
                --     HoJ:DebugMsg( "Not deleting outTable['" .. key .. "'] = " .. tostring( outTable[value] ) )
                end
            end
        end
    end
    return true
end

-- -----------------------------------------------------------------------------
-- HoJ:CopySpellInfosToTable
-- -----------------------------------------------------------------------------
function HoJ:CopySpellInfosToTable( saveTable, spellInfos, outTable, nilDeletes, name )
    if ( spellInfos == nil ) then
        -- not an error anymore if the class has never saved its vars
        -- HoJ:Error( "spellInfos is nil!" )
        return 
    end    
    if ( outTable == nil ) then
        HoJ:Error( "outTable '" .. name .. "' is nil!" )
        return 
    end

    for key, value in pairs( spellInfos ) do
        if ( outTable[key] == nil ) then
            outTable[key] = { }
        end
        HoJ:CopyTableUsingSaveTable( saveTable, value, outTable[key], nilDeletes )
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:CreateSavedVariables
-- Copies everything from HoJ that we want to save into a new table and returns
-- the new table
-- -----------------------------------------------------------------------------
function HoJ:CreateSavedVariables()
    savedVars = { }
    HoJ:CopyTableUsingSaveTable( saveTable_HoJ, HoJ, savedVars, true )

    savedVars[HoJ.CLASS_NAME_PALADIN]._spellInfos = { } 
    savedVars[HoJ.CLASS_NAME_WARRIOR]._spellInfos = { }
    savedVars[HoJ.CLASS_NAME_ROGUE]._spellInfos = { }
    
    HoJ:CopySpellInfosToTable( saveTable_spellInfo, HoJ[HoJ.CLASS_NAME_PALADIN]._spellInfos,     savedVars[HoJ.CLASS_NAME_PALADIN]._spellInfos, true, "1" )
    HoJ:CopySpellInfosToTable( saveTable_spellInfo, HoJ[HoJ.CLASS_NAME_WARRIOR]._spellInfos,     savedVars[HoJ.CLASS_NAME_WARRIOR]._spellInfos, true, "2" )
    HoJ:CopySpellInfosToTable( saveTable_spellInfo, HoJ[HoJ.CLASS_NAME_ROGUE]._spellInfos,       savedVars[HoJ.CLASS_NAME_ROGUE]._spellInfos, true, "3" )
    HoJ:CopySpellInfosToTable( saveTable_spellInfo, HoJ[HoJ.CLASS_NAME_DEMONHUNTER]._spellInfos, savedVars[HoJ.CLASS_NAME_DEMONHUNTER]._spellInfos, true, "4" )

    -- HoJ:PrintTable( savedVars, "savedVars", 0 )

    return savedVars
end

-- -----------------------------------------------------------------------------
-- HoJ:CreateSavedVariablesFromHoJ
-- -----------------------------------------------------------------------------
function HoJ:CreateSavedVariablesFromHoJ( msg )
    if ( msg ~= nil ) then
        HoJ:Msg( msg )
    end
    HoJ._savedVars = HoJ:CreateSavedVariables()
    _G[SAVED_VARIABLES_NAME] = HoJ._savedVars;
end

-- -----------------------------------------------------------------------------
-- HoJ:RemoveOldSpellInfos
-- -----------------------------------------------------------------------------
function HoJ:RemoveOldSpellInfos( savedSpellInfos, currentSpellInfos )
    if ( currentSpellInfos == nil ) then
        HoJ:Error( "Current spell infos are nil!" )
        return
    end

    if ( savedSpellInfos == nil ) then
        return
    end

    for key, value in pairs( savedSpellInfos ) do
        -- HoJ:Log( format( "Old spellInfo key '%s'", tostring( key ) ) )
        if ( key ~= nil ) then
            local curValue = currentSpellInfos[key]
            if ( curValue == nil ) then
                -- HoJ:Log( format( "Removing old spell info for '%s'", tostring( key ) ) )
                savedSpellInfos[key] = nil
            end
        end
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:LoadSavedVariablesIntoHoJ
-- copies saved variables back over the HoJ vars
-- -----------------------------------------------------------------------------
function HoJ:LoadSavedVariablesIntoHoJ() 
    -- for new classes added after other classes, we need to add their settings 
    -- individually without destroying the other classes' settings.
    if ( HoJ._savedVars[HoJ.CLASS_NAME_ROGUE] == nil ) then
        -- HoJ:Msg( "Added saved variables for Rogue class." )
        -- function HoJ:CopyTableUsingSaveTable( saveTable, sourceTable, outTable, nilDeletes )
        HoJ._savedVars[HoJ.CLASS_NAME_ROGUE] = { }
        HoJ:CopyTableUsingSaveTable( saveTable_HoJ[HoJ.CLASS_NAME_ROGUE], HoJ[HoJ.CLASS_NAME_ROGUE], HoJ._savedVars[HoJ.CLASS_NAME_ROGUE], true)
        HoJ._savedVars[HoJ.CLASS_NAME_ROGUE]._spellInfos = { }
    end

    -- copy only saved variables from saved vars to HoJ
    HoJ:CopyTableUsingSaveTable( saveTable_HoJ, HoJ._savedVars, HoJ, false )

    -- remove any old spell infos that existed pre-Shadowlands
    if ( HoJ._savedVars[HoJ.CLASS_NAME_PALADIN] ~= nil ) then
        HoJ:RemoveOldSpellInfos( HoJ._savedVars[HoJ.CLASS_NAME_PALADIN]._spellInfos, HoJ._paladinSpellInfos )
        HoJ:CopySpellInfosToTable( saveTable_spellInfo, HoJ._savedVars[HoJ.CLASS_NAME_PALADIN]._spellInfos,     HoJ[HoJ.CLASS_NAME_PALADIN]._spellInfos, false, "5" )
    end
    if ( HoJ._savedVars[HoJ.CLASS_NAME_WARRIOR] ~= nil ) then
        HoJ:RemoveOldSpellInfos( HoJ._savedVars[HoJ.CLASS_NAME_WARRIOR]._spellInfos, HoJ._warriorSpellInfos )
        HoJ:CopySpellInfosToTable( saveTable_spellInfo, HoJ._savedVars[HoJ.CLASS_NAME_WARRIOR]._spellInfos,     HoJ[HoJ.CLASS_NAME_WARRIOR]._spellInfos, false, "6" )
    end
    if ( HoJ._savedVars[HoJ.CLASS_NAME_ROGUE] ~= nil ) then
        HoJ:RemoveOldSpellInfos( HoJ._savedVars[HoJ.CLASS_NAME_ROGUE]._spellInfos, HoJ._rogueSpellInfos )
        HoJ:CopySpellInfosToTable( saveTable_spellInfo, HoJ._savedVars[HoJ.CLASS_NAME_ROGUE]._spellInfos,       HoJ[HoJ.CLASS_NAME_ROGUE]._spellInfos, false, "7" )
    end
    if ( HoJ._savedVars[HoJ.CLASS_NAME_DEMONHUNTER] ~= nil ) then
        HoJ:RemoveOldSpellInfos( HoJ._savedVars[HoJ.CLASS_NAME_DEMONHUNTER]._spellInfos, HoJ._demonHunterSpellInfos )
        HoJ:CopySpellInfosToTable( saveTable_spellInfo, HoJ._savedVars[HoJ.CLASS_NAME_DEMONHUNTER]._spellInfos, HoJ[HoJ.CLASS_NAME_DEMONHUNTER]._spellInfos, false, "8" )
    end

end

-- -----------------------------------------------------------------------------
-- HoJ:VersionToValue
-- -----------------------------------------------------------------------------
function HoJ:VersionToValue( major, minor, build )
    local value = 0
    if ( major ~= nil ) then
        value = major * 1000000
    end
    if ( minor ~= nil ) then
        value = value + minor * 1000
    end
    if ( build ~= nil ) then
        value = value + build
    end
    return value
end

-- -----------------------------------------------------------------------------
-- HoJ:LoadSavedVars
-- -----------------------------------------------------------------------------
function HoJ:LoadSavedVars()
    if ( _G[SAVED_VARIABLES_NAME] == nil ) then
        HoJ:CreateSavedVariablesFromHoJ( HoJ._text["Defaulting saved variables..."] )
    else
        local major = HoJ.MAJOR_VERSION
        local minor = HoJ.MINOR_VERSION
        local build = HoJ.BUILD_VERSION

        HoJ._savedVars = _G[SAVED_VARIABLES_NAME]

        local curVersionValue = self:VersionToValue( HoJ.MAJOR_VERSION, HoJ.MINOR_VERSION, HoJ.BUILD_VERSION )
        local savedVersionValue = self:VersionToValue( HoJ._savedVars["MAJOR_VERSION"], HoJ._savedVars["MINOR_VERSION"], HoJ._savedVars["BUILD_VERSION"] )

        -- HoJ:Log( format( "Current version value = %.0f", curVersionValue ) )
        -- HoJ:Log( format( "Saved version value = %.0f", savedVersionValue ) )

        if ( savedVersionValue < self:VersionToValue( 0, 97, 20 ) ) then
            HoJ:Msg( "Rogue specs were re-enabled due to old saved variables version." )
            local rogueSpecs = HoJ._savedVars[HoJ.CLASS_NAME_ROGUE]
            if ( rogueSpecs ~= nil ) then
                if ( rogueSpecs[HoJ.SPEC_NAME_ROGUE_ASSASSINATION] ~= nil ) then
                    rogueSpecs[HoJ.SPEC_NAME_ROGUE_ASSASSINATION]._enabled = true
                end
                if ( rogueSpecs[HoJ.SPEC_NAME_ROGUE_OUTLAW] ~= nil ) then
                    rogueSpecs[HoJ.SPEC_NAME_ROGUE_OUTLAW]._enabled = true
                end
                if ( rogueSpecs[HoJ.SPEC_NAME_ROGUE_SUBTLETY] ~= nil ) then
                    rogueSpecs[HoJ.SPEC_NAME_ROGUE_SUBTLETY]._enabled = true
                end
            end
        elseif ( savedVersionValue < self:VersionToValue( 0, 97, 0 ) ) then
            HoJ:CreateSavedVariablesFromHoJ( HoJ._text["Resetting old saved variables. You may need to re-configure some settings in the addon menu."] )
        elseif ( HoJ._savedVars["VersionMajor"] ~= nil ) then
            -- wipe out any existing variables from old versions
            -- HoJ:DebugMsg( "Saved Version = " .. HoJ._savedVars.VersionMajor .. "." .. HoJ._savedVars.VersionMinor .. "." .. HoJ._savedVars.VersionBuild )
            HoJ:CreateSavedVariablesFromHoJ( HoJ._text["Resetting old saved variables. You may need to re-configure some settings in the addon menu."] )
        else
            HoJ:Msg( HoJ._text["Restoring saved variables."] )
        end

        HoJ:LoadSavedVariablesIntoHoJ()

        -- restore the actual versions because loading the save overwrites it
        HoJ.MAJOR_VERSION = major
        HoJ.MINOR_VERSION = minor
        HoJ.BUILD_VERSION = build
    end
end
