-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Utils.lua
-- DESC: Addon utility functions.

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Utils.lua" )

-- ============================================================================

-- -----------------------------------------------------------------------------
-- HoJ:PrintTable
-- -----------------------------------------------------------------------------
function HoJ:PrintTable( t, name, depth )
    if ( t == nil ) then
        HoJ:DebugMsg( "PrintTable: table is nil!" )
        return
    end
    
    local prevIndentStr = string.rep( "  ", depth * 2 )
    HoJ:DebugMsg( prevIndentStr .. "[" .. name .. "] = {" )

    local indentStr = string.rep("  ", ( depth + 1 ) * 2)

    for key, value in pairs( t ) do
        local valueType = type( value )
        if ( valueType == "table" ) then
            HoJ:PrintTable( value, key, depth + 1 )
        elseif ( valueType == "string" ) then
            HoJ:DebugMsg( indentStr .. "[" .. key .. "] = '" .. value .. "'" )
        elseif ( valueType == "boolean" ) then
			HoJ:DebugMsg( indentStr .. "[" .. key .. "] = " .. tostring( value ) )
		elseif ( valueType == "function" ) then
            HoJ:DebugMsg( indentStr .. "[" .. key .. "] = " .. key )
        else
            HoJ:DebugMsg( indentStr .. "[" .. key .. "] = " .. value )
        end
    end

    HoJ:DebugMsg( prevIndentStr .. "}" )
end

--[[
function DeepCopyTable( sourceTable )
    local originalType = type( sourceTable )
    local destTable
    if originalType == 'table' then
        -- copy tables
        destTable = {}
        for originalKey, orginalValue in next, sourceTable, nil do
            destTable[DeepCopyTable( originalKey )] = DeepCopyTable( orginalValue )
        end
        setmetatable( destTable, DeepCopyTable( getmetatable( sourceTable ) ) )
    else
        -- copy everything else
        destTable = sourceTable
    end
    return destTable
end
]]

-- -----------------------------------------------------------------------------
-- HoJ:CopyTable
-- simple recursive copy that doesn't copy metatables
-- -----------------------------------------------------------------------------
function HoJ:CopyTable( source )
	local sourceType = type( source )
	local destTable
	if ( sourceType == "table" ) then
		destTable = { }
		for srcKey, srcValue in pairs( source ) do
			destTable[srcKey] = HoJ:CopyTable( srcValue )
		end
	else
		destTable = source
	end
	return destTable
end

-- -----------------------------------------------------------------------------
-- HoJ:MergeTable
-- simple recursive merge that doesn't copy metatables
-- sub-tables are merged recursively.
-- if a non-table key in dest matches a key in a source, then the dest key is
-- overwritten. 
-- -----------------------------------------------------------------------------
function HoJ:MergeTable( source, dest )
	local sourceType = type( source )
	local destType = type( dest )
	if ( sourceType ~= "table" or destType ~= "table" ) then
		HoJ:Error( "Attempted to merge non-tables! Merge failed!" )
		return false
	end

	for srcKey, srcValue in pairs( source ) do
		local srcValueType = type( srcValue )
		if ( srcValueType == "table" ) then
			if ( dest[srcKey] == nil ) then
				dest[srcKey] = { }
			end
			if ( not HoJ:MergeTable( srcValue, dest[srcKey] ) ) then
				return false
			end
		else
			dest[srcKey] = srcValue
		end
	end

	return true
end

-- -----------------------------------------------------------------------------
-- HoJ:ShallowCopyTable
-- does a shallow (non-recursive, one-level deep) copy of a table and returns
-- the copy.
-- -----------------------------------------------------------------------------
function HoJ:ShallowCopyTable( source )
	local copy = { }
	for key, value in pairs( source ) do
		copy[key] = value
	end
	return copy
end

-- -----------------------------------------------------------------------------
-- HoJ:ConstructForcedCooldown
-- -----------------------------------------------------------------------------
function HoJ:ConstructForcedCooldown( time, spellInfo )
	local cd = { _spellId = -1, _cooldownEndTime = -1.0 }
	if ( time == nil or spellInfo == nil ) then
		return cd
	end
	
	cd._spellId = spellInfo.id
	local startTime, duration = HoJ:GetSpellCooldown( spellInfo )
	cd._cooldownEndTime = time + duration
	
	return cd
end

-- -----------------------------------------------------------------------------
-- HoJ:ConstructCurrentRotationState
-- -----------------------------------------------------------------------------
function HoJ:ConstructCurrentRotationState( currentTime )
	
	local rs = { 
		_time = currentTime, 
		_predictAhead = 0.0,
		_power = self._power, 
		_gcdRemaining = self._gcdRemaining,
		-- if TYPE_SKILL_CHARGES would have returned 0 charges, it will
		-- isntead return _minSkillCharges 
		-- This is used when predicting the next skill, since if we're completely
		-- out of skills it's better to use a skill at minimum charges than skip
		-- it like we would do in normal rotation.
		_minSkillCharges = 0,
		_forcedCooldowns = { } 
	}

	return rs
end

-- -----------------------------------------------------------------------------
-- HoJ:PredictPower
-- -----------------------------------------------------------------------------
function HoJ:PredictPower( curPower, time, predictAhead )
	-- if class has auto power generation, we must also add the power 
	-- that will be gained between current and predicted time
	local power = curPower
	if ( self._curSpecInfo._autoPowerGen == true ) then
		if ( time >= self._nextPowerGainTime ) then
			local powerGained = predictAhead * self._predictedPowerPerSec
			power = power + powerGained
		end
	end
	return power
end

-- -----------------------------------------------------------------------------
-- HoJ:ConstructPredictedRotationState

-- -----------------------------------------------------------------------------
function HoJ:ConstructPredictedRotationState( currentTime, predictAhead, spellInfo )
	
	if ( predictAhead == nil ) then
		HoJ:Error( "ConstructPredictedRotationState() called with invalid parameters!" )
	end

	local rs = self:ConstructCurrentRotationState( currentTime )

	rs._time = rs._time + predictAhead
	rs._predictAhead = rs._predictAhead + predictAhead
	rs._gcdRemaining = rs._gcdRemaining - predictAhead
	if ( rs._gcdRemaining < 0.0 ) then
		rs._gcdRemaining = 0.0
	end

	-- when predicting, treat all skills as if they have at least 1 charge
	rs._minSkillCharges = 1

	if ( spellInfo ~= nil ) then
		rs._power = rs._power + spellInfo._powerGen
		rs._forcedCooldowns[spellInfo.id] = HoJ:ConstructForcedCooldown( rs._time, spellInfo )
	else
		rs._power = self:PredictPower( rs._power, rs._time, rs._predictAhead )
	end

	return rs
end
--[[
-- -----------------------------------------------------------------------------
-- HoJ:ConstructPredictedRotationStateFromBase
-- -----------------------------------------------------------------------------
function HoJ:ConstructPredictedRotationStateFromBase( rotationState, predictAhead )
	local rs = HoJ:ShallowCopyTable( rotationState )

	rs._time = rs._time + predictAhead -- rotationState._time will already include the base predictAhead
	rs._predictAhead = rs._predictAhead + predictAhead
	rs._gcdRemaining = rs._gcdRemaining - predictAhead
	rs._minSkillCharges = 1
	if ( rs._gcdRemaining < 0.0 ) then
		rs._gcdRemaining = 0.0
	end

	rs._power = self:PredictPower( rs._power, rs._time, rs._predictAhead )

	return rs
end
--]]

-- -----------------------------------------------------------------------------
-- HoJ:InitClass
-- -----------------------------------------------------------------------------
function HoJ:InitClass( className )
	self._className = className
	self._classId = HoJ._classNameToClassId[className]
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateSpec
-- -----------------------------------------------------------------------------
function HoJ:UpdateSpec()
	local specId = GetSpecialization()
	-- HoJ:Log( "specId = " .. specId )
	if ( specId == nil or specId <= 0 or specId >= 5 ) then
		-- HoJ:Log( "No spec has been chosen!" )
		self._specChosen = false
	else
		self._specChosen = true
	end

	self._classSpecInfo = HoJ[self._className]
	if ( self._classSpecInfo ~= nil ) then
		if ( specId == nil ) then
			specId = self._classSpecInfo._defaultSpecId
			-- this should never really happen, but is probably survivable
			-- I thnk I've seen GetSpecialization return nil after entering world?
			HoJ:Error( "No specialization selected. Defaulting to " .. specName .. "." )
		end

		local classSpecIdTable = HoJ._specIdToSpecName[self._className]
		if ( classSpecIdTable ~= nil ) then
			-- we should always have a spec id table for a supported class
			local specName = classSpecIdTable[specId]
			if ( specName ~= nil ) then
				-- HoJ:Log( "specName = " .. specName )
				self._curSpecInfo = self._classSpecInfo[specName]
				self._role = self._curSpecInfo._role
				self._isActive = true

				return self._curSpecInfo
			else
				HoJ:Error( "classSpecIdTable has nil entry for spec Id " .. tostring( specId ) )
			end
		else
			HoJ:Error( "classSpecIdTable is nil for " .. self._className .. "!" )
		end
	end
	self._curSpecInfo = nil
	self._role = nil
	self._isActive = false
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellManaCost
-- -----------------------------------------------------------------------------
function HoJ:GetSpellManaCost( spellId )
	local a = GetSpellPowerCost( spellId )
	for k, t in pairs(a) do
		if ( t.type == 0 ) then
			-- HoJ:Log( "Spell " .. spellId .. " mana cost is " .. t.cost )
			return t.cost
		end
	end
	--HoJ:Log( "Spell " .. spellId .. " does not use mana!" )
	return 0
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellName
-- -----------------------------------------------------------------------------
function HoJ:GetSpellName( spellID )
	name = GetSpellInfo( spellID );
	if ( name == nil ) then
	end
	return name;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellRank
-- -----------------------------------------------------------------------------
function HoJ:GetSpellRank( spellID )
	_, rank = GetSpellInfo( spellID )
	return rank
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellIcon
-- -----------------------------------------------------------------------------
function HoJ:GetSpellIcon( spellID )
	_, _, icon = GetSpellInfo( spellID )
	return icon
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellDebugName
-- -----------------------------------------------------------------------------
function HoJ:GetSpellDebugName( spell )
	if ( spell == nil ) then
		return "nil";
	end
	if ( spell._debugName ~= nil ) then
		return spell._debugName;
	else
		return tostring( spell.name )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellIndex
-- this tries to find the index of a spell in the spellbook, which is not the
-- same as the spell ID of a spell. For some reason the spell index is needed
-- to find if a target is in range of a spell, rather than the spell ID.
-- -----------------------------------------------------------------------------
function HoJ:GetSpellIndex( spellInfo )
	-- tabs are the side tabs which are currently "General", "Retribution", "Holy", and "Protection"
	-- there is no way to know how many spells are on a single tab, but it won't be over 48 for paladins in 8.1
	local numSpellTabs = GetNumSpellTabs()
	--HoJ:DebugMsg( "Found " .. numSpellTabs .. " spellbook tabs" );
	local max = numSpellTabs * 96
	for i = 1, max do
		HoJ:ErrorIf( spellInfo._bookType == nil, format( "spellInfo '%s' has nil _bookType!", spellInfo.name ) )
			
		local spellName, spellSubName = GetSpellBookItemName( i, spellInfo._bookType )
		if ( spellName == nil ) then
			-- HoJ:DebugMsg( "No spellbook info found for index " .. i .. " when looking up " .. spellInfo.name );
			return -1;
	   	else
			-- use spellName and spellSubName here
			-- HoJ:DebugMsg( "Spell #" .. i .. ": '" .. spellName .. "'" )
			if ( spellInfo.name == spellName ) then
				-- HoJ:DebugMsg( "Found spellbook index " .. i .. " for spell id " .. spellInfo.name );
				return i
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_InitSpellInfo
-- -----------------------------------------------------------------------------
function HoJ_InitSpellInfo()
	-- HoJ:DebugMsg( "InitSpellInfo" );

	local spellIdsToSpellInfos = { }
	HoJ._spellIdsToSpellInfos = spellIdsToSpellInfos

	for key, value in pairs( HoJ.SpellInfo ) do
		if ( value == nil or value.id == nil ) then
			HoJ:DevError( format( "Skipped InitSpellInfo for key '%s'", tostring( key ) ) )
		else
			spellIdsToSpellInfos[value.id] = key

			-- HoJ:Log( format( "InitSpellInfo( %s, %s, %s)", tostring( value.id), tostring( value.name ), tostring( value.id ) ) )
			local name, rank, icon = GetSpellInfo( value.id )
			-- some spell infos use an alternate icon id -- for instance a buff that allows another
			-- spell to be cast (Empyrean Power), we want to show the spell that can be cast on the
			-- rotation frame, not the buff.
			if ( value.iconId ~= nil ) then
				-- HoJ:DebugMsg( "Getting iconId " .. value.iconId .. " for " .. value.name );
				_, _, icon = GetSpellInfo( value.iconId )
			end
			-- For some reason, if we query Crusader Strike while prot, we get a weird texture id
			-- so when we switch roles we go ahead and re-init all of the spell infos.
			-- HoJ:DebugMsg( "  id = " .. value.id );
			-- HoJ:DebugMsg( "  old icon = " .. value.icon .. ", new icon = " .. icon );
			value.name = name
			value.rank = rank
			value.icon = icon
			value.spellIndex = FindSpellBookSlotBySpellID( value.id )

			-- everything that's not set to _bookType == BOOKTYPE_PET should just be
			-- BOOKTYPE_SPELL
			-- This is necessary for range checks by spell index to work, which is necessary
			-- for spells that can have different ids based on spec or player level.
			if ( value._bookType == nil ) then
				value._bookType = BOOKTYPE_SPELL
			end

			if ( value._breaksStealth == nil ) then
				value._breaksStealth = true
			end

			if ( value.spellIndex == nil or value.spellIndex < 0 ) then
				value.spellIndex = HoJ:GetSpellIndex( value )
				--[[
				if ( value.spellIndex ~= nil and value.spellIndex >= 0 ) then
					HoJ:Log( format( "Found spellIndex %i for %s the hard way!", value.spellIndex, name ) )
				else
					HoJ:Error( format( "Could not find spellIndex for %s!", name ) )
				end
				--]]
			end

			value._manaCost = HoJ:GetSpellManaCost( value.id )
			value._lastCastTime = -1.0
		end
	end

	HoJ._spellsChanged = false

	HoJ:NotifyDevMode()
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIcon
-- -----------------------------------------------------------------------------
function HoJ:AddIcon( list, spellInfo, text )
	-- HoJ:DebugMsg( "HoJ:AddIcon( " .. spellInfo.name .. " )" )
	text = text or ""

	if ( list == nil ) then
		HoJ:Error( "HoJ:AddIcon - Icon list is NIL" )
	end

	if ( not spellInfo.showOnBar ) then
		HoJ:RemoveIcon( list, spellInfo )
		return
	end

	-- HoJ:FTLogD( 15, spellInfo.id == 188370, "AddIcon( " .. spellInfo.name .. " )" )
	local si = spellInfo
	si.text = text
	-- table.insert( list, si )
	list[#list + 1] = si
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIconExclusive
-- -----------------------------------------------------------------------------
function HoJ:AddIconExclusive( list, spellInfo, text )
	-- HoJ:DebugMsg( "HoJ:AddIcon( " .. spellInfo.name .. " )" )
	text = text or ""

	if ( list == nil ) then
		self:DebugMsg( "HoJ:AddIcon - Icon list is NIL" )
	end

	if ( not spellInfo.showOnBar ) then
		self:RemoveIcon( list, spellInfo )
		return
	end

	for i = 1, #list do
		if ( list[i].name == spellInfo.name ) then
			return
		end
	end

	-- self:DebugMsg( "AddIconExclusive( " .. spellInfo.name .. " )" )
	local si = spellInfo
	si.text = text
	-- table.insert( list, si )
	list[#list + 1] = si
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIconDebug
-- just adds debug text where the cooldown text would be
function HoJ:AddIconDebug( list, spellInfo, debugText )
	HoJ:AddIcon( list, spellInfo, "" )
end

-- -----------------------------------------------------------------------------
-- HoJ:RemoveIcon
-- -----------------------------------------------------------------------------
function HoJ:RemoveIcon( list, spellInfo )
	if ( list == nil ) then
		HoJ:DebugMsg( "HoJ:RemoveIcon - Icon list is NIL" );
	end
	-- HoJ:DebugMsg( "RemoveIcon( " .. spellInfo.name .. " )" );
	for i, v in pairs( list ) do
		if ( v == spellInfo ) then
			table.remove( list, i );
			break;
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ClearIconList
-- -----------------------------------------------------------------------------
function HoJ:ClearIconList( list )
	for i, v in pairs( list ) do
		list[ i ] = nil;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellCooldown
-- -----------------------------------------------------------------------------
function HoJ:GetSpellCooldown( spellInfo )
	local id = spellInfo._cdId
	if ( id == nil ) then
		id = spellInfo.id
	end
	return GetSpellCooldown( id )
end

-- -----------------------------------------------------------------------------
-- HoJ:SetIconFrame
-- -----------------------------------------------------------------------------
function HoJ:SetIconFrame( curTime, v, spellInfo )
	--local debug = spellInfo.id == 188370
	--self:LogD( debug, "SetIconFrame for " .. spellInfo.name .. ", frame " .. v.frame:GetName() .. ", icon = " .. spellInfo.icon )
	if ( v == nil ) then
		return -- there are more skills than frames
	end

	v.frame:EnableMouse( true )
	--v.frame:Enable();
	v.frame:Show()
	-- HoJ:DebugMsg( "Enable mouse for " .. v.frame:GetName() );

	v.spellInfo = spellInfo

	v.texture:SetTexture( spellInfo.icon )
	v.texture:Show()

	local hideText = true
	if ( v.text ~= nil ) then
		if ( spellInfo.text ~= nil and spellInfo.text ~= "" ) then
			local fontFile = GameFontHighlightLarge:GetFont();
			v.text:SetFont( fontFile, 12, "THICKOUTLINE" );
			v.text:SetText( spellInfo.text );
			v.text:Show();
			if ( spellInfo.spellType == HoJ.SPELL_SKILL or spellInfo.spellType == HoJ.SPELL_TALENT ) then
				if ( v.texture ~= nil ) then
					v.texture:SetAlpha( HoJ._skillCDAlpha );
				end
			end
		end
	end
	if ( hideText ) then
		v.text:Hide();
		if ( v.texture ~= nil ) then
			v.texture:SetAlpha( HoJ._skillNormalAlpha );
		end
	end

	-- if the frame has a cooldown, set it
	if ( v._cooldownFrame ~= nil ) then
		local hideCooldown = true
		local start, duration, remaining
		start, duration = HoJ:GetSpellCooldown( spellInfo )
		if ( start > 0 and duration > 0 ) then
			v._cooldownFrame:Show()
			v._cooldownFrame:SetCooldown( start, duration )
			hideCooldown = false
		end
		if ( hideCooldown ) then
			v._cooldownFrame:Hide()
		end
	end

	local buffTimeRemaining, buffStart, buffDuration, buffStacks
	local isBuff = spellInfo.spellType == HoJ.SPELL_BUFF 
					or spellInfo.spellType == HoJ.SPELL_TALENT_BUFF
					or spellInfo.spellType == HoJ.SPELL_SKILL_BUFF
	local showTime = false
	if ( isBuff ) then
		buffTimeRemaining, buffStart, buffDuration, buffStacks = HoJ:GetBuffTimeRemaining( curTime, spellInfo.id )
		showTime = ( spellInfo._showTimeThreshold == nil or buffTimeRemaining <= spellInfo._showTimeThreshold )
	end

	if ( v._timeLeftFontString ~= nil and showTime ) then
		local hide = true
		if ( isBuff and buffTimeRemaining > 0 and buffDuration > 0 ) then
			local fontFile = GameFontHighlightLarge:GetFont();
			v._timeLeftFontString:SetFont( fontFile, 12, "THICKOUTLINE" );
			v._timeLeftFontString:SetText( string.format( "%.0f", buffTimeRemaining ) );
			v._timeLeftFontString:Show();
			hide = false
		end
		if ( hide ) then
			v._timeLeftFontString:Hide()
		end
	end

	if ( v._stacksFontString ~= nil ) then
		local hide = true
		if ( isBuff and spellInfo._showStacks and buffStacks > 0 ) then
			if ( buffStacks == nil ) then
				HoJ:Error( "buffStacks is nil for " .. spellInfo.name .. " with _showStacks = true!" )
			else
				local fontFile = GameFontHighlightLarge:GetFont();
				v._stacksFontString:SetFont( "Fonts\\ARIALN.ttf", 10, "OUTLINE" );
				v._stacksFontString:SetText( string.format( "%.0f", buffStacks ) );
				v._stacksFontString:Show();
				hide = false
			end
		end
		if ( hide ) then
			v._stacksFontString:Hide()
		end
	end

	if ( v._chargesFontString ~= nil ) then
		local hide = true
		local fontFile = GameFontHighlightSmallOutline:GetFont()
		v._chargesFontString:SetFont( fontFile, 12, "THICKOUTLINE" )
		local currentCharges, maxCharges = GetSpellCharges( spellInfo.id )
		if ( spellInfo.id == 202539 ) then
			HoJ:DebugMsg( "currentCharges = " .. currentCharges .. ", maxCharges = " .. maxCharges )
		end
		if ( currentCharges ~= nil and maxCharges > 1 ) then
			v._chargesFontString:SetText( tostring( currentCharges ) )
			v._chargesFontString:Show()
			hide = false
		end
		if ( hide ) then
			v._chargesFontString:Hide()
		end
	end
end

-- -----------------------------------------------------------------------------
-- SkillCompare2
-- -----------------------------------------------------------------------------
function SkillCompare2( priorityList, aSpellInfo, bSpellInfo )
	local aPriority = priorityList[aSpellInfo.id];
	local bPriority = priorityList[bSpellInfo.id];

	-- NOTE: a nil warning here means the priority list is missing an entry for this a spell id
	local hadError = false
	if ( aPriority == nil ) then
		HoJ:Error(aSpellInfo.name .. " is missing from priorityList " .. priorityList._name )
		HoJ:DebugBreak()
		hadError = true
	end
	if ( bPriority == nil ) then
		HoJ:Error(bSpellInfo.name .. " is missing from priorityList "  .. priorityList._name )
		HoJ:DebugBreak()
		hadError = true
	end
	if ( hadError ) then
		return true
	end
	return ( aPriority > bPriority )
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateIcons2
-- -----------------------------------------------------------------------------
local SortedKeyLists2 = {
	{ 1 },
	{ 1, 2 },
	{ 1, 2, 3 },
	{ 1, 2, 3, 4 },
	{ 1, 2, 3, 4, 5 },
	{ 1, 2, 3, 4, 5, 6 },
	{ 1, 2, 3, 4, 5, 6, 7 },
	{ 1, 2, 3, 4, 5, 6, 7, 8 },
	{ 1, 2, 3, 4, 5, 6, 7, 8, 9 },
	{ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 },
}

function HoJ:UpdateIcons2( time, list, priorityList, frameList, sortList )
	-- clear all icon frames
	local setIconFrame = HoJ.SetIconFrame
	local frameNum = 0

	local size = #list
	if ( size > 1 and sortList ) then
		if ( size > #SortedKeyLists2 ) then
			HoJ:Error( "Too many icons in list." );
			size = #SortedKeyLists2
		end

		local sortIconsStartMs = HoJ.ProfileBegin2( "SortIcons" )

		-- use a table lookup to initialize the sorted keys list to the correct size
		local sortedKeys = SortedKeyLists2[size]
		table.sort( sortedKeys, function( a, b ) return SkillCompare2( priorityList, list[a], list[b] ) end )

		HoJ.ProfileEnd( "SortIcons", sortIconsStartMs )

		for i = 1, size do
			local v = list[sortedKeys[i]]
			setIconFrame( self, time, frameList[i], list[ sortedKeys[i] ] )
			frameNum = i
		end
	else
		--local sortIconFramesStartMs = HoJ.ProfileBegin2( "SetIconFramesSorted" )

		-- using #list is slightly faster in this case than ipairs()
		for i = 1, size do
			setIconFrame( self, time, frameList[i], list[i] )
			frameNum = i
		end

		-- HoJ.ProfileEnd( "SetIconFramesSorted", sortIconFramesStartMs )
	end

	--local enableMouseStartMs = HoJ.ProfileBegin2( "EnableMouse" )
	for i = (frameNum + 1), #frameList do
		local frame = frameList[ i ].frame
		frame:EnableMouse( false )
		frame:Hide()
	end
	--HoJ.ProfileEnd( "EnableMouse", enableMouseStartMs )
end

-- -----------------------------------------------------------------------------
-- HoJ:UnitBuffForName
-- -----------------------------------------------------------------------------
function HoJ:UnitBuffForName( unit, buffName )
	local i = 1;
	local name, icon, count = UnitAura( unit, i );
	while name do
		if ( name == buffName ) then
			return name, icon, count;
		end
		i = i + 1;
		name, icon, count = UnitBuff( unit, i );
	end
	return name, icon, count;
end

-- -----------------------------------------------------------------------------
-- HoJ:ConstructActiveBuff
-- -----------------------------------------------------------------------------
function HoJ:ConstructActiveBuff() 
	return { start = 0, duration = 0, _expirationTime = -1.0, stacks = 0 }
end

-- -----------------------------------------------------------------------------
-- HoJ:ConstructAoeBuff
-- -----------------------------------------------------------------------------
function HoJ:ConstructAoeBuff( castBySpellInfo, buffSpellInfo )
	return {
        _castBy = castBySpellInfo,
        _spellInfo = buffSpellInfo,
        _start = 0.0,
        _duration = 0.0,
        _expiration = 0.0,
    }
end

-- -----------------------------------------------------------------------------
-- HoJ:LoopOverActiveBuffs
-- returns number of buffs found
-- -----------------------------------------------------------------------------
function HoJ:LoopOverActiveBuffs( time, unit, activeBuffs, filter )
	local i = 1
	local numBuffs = 0
	local name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i, filter )
	while name do
		if ( spellId ~= nil ) then
			-- local debug = spellId == 188370

			numBuffs = numBuffs + 1
			-- HoJ:FTLogD( 15, debug, filter .. ": Buff " .. name .. " is active, spellId = " .. spellId )
			local activeBuff = activeBuffs[spellId]
			if ( activeBuff ~= nil ) then
				-- HoJ:LogD( debug, name .. ": duration = " .. tostring( duration ) .. ", expiration = " .. tostring( expirationTime ) )

				if ( duration ~= nil and expirationTime ~= nil ) then
					if ( duration == 0.0 and expirationTime == 0.0 ) then
						--HoJ:LogD( debug, "Check for active AoE buff " .. name )
						local aoeBuff = self._curSpecInfo._aoeBuffTable[spellId]
						if ( aoeBuff ~= nil ) then
							-- HoJ:LogD( debug, "Found AoE buff " .. name )
							if ( aoeBuff._expiration > time ) then
								activeBuff.start = aoeBuff._start
								activeBuff.duration = aoeBuff._duration
								activeBuff._expirationTime = aoeBuff._expiration
								-- HoJ:LogD( debug, format( "Added active AoE buff %s, start = %.2f, duration = %.2f, expire = %.2f",
								--	name, activeBuff.start, activeBuff.duration, activeBuff._expirationTime ) )
							end
						end
					else
						-- HoJ:FTLogD( 15, debug, "duration = " .. tostring( duration ) .. ", expiration = " .. tostring( expirationTime ) )
						activeBuff.start = expirationTime - duration
						activeBuff.duration = duration
						activeBuff._expirationTime = expirationTime
					end
				end

				if ( count ~= nil ) then
					activeBuff.stacks = count
				end
			else
				-- HoJ:FTLogD( 15, debug, filter .. ": Buff " .. name .. " (id " .. spellId .. ") not in activeBuffs list" )
			end
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i, filter )
	end	
	return numBuffs
end

-- -----------------------------------------------------------------------------
-- HoJ:FindActiveBuffs
-- table of flags for each buff, indexed by spell id
--	{
--		start,			-- time buff started
--		duration,		-- duration of the buff
--		_expirationTime,-- time when buff will expire
--		stacks,			-- number of stacks of the buff
--		spellId,		-- spell Id of the buff
--	}
-- -----------------------------------------------------------------------------
function HoJ:FindActiveBuffs( time, unit, activeBuffs )
    -- set all active skills active flag to false
    if ( activeBuffs == nil ) then
        return 
    end

	for key, buff in pairs( activeBuffs ) do
		buff.isActive = false
		buff.start = 0.0
		buff.duration = 0.0
		buff._expirationTime = -1.0
		buff.stacks = 0
	end

	-- we can't get back helpful and harmful buffs at the same time so we must loop twice
	HoJ:LoopOverActiveBuffs( time, unit, activeBuffs, "HELPFUL" )
	HoJ._hasDebuff = HoJ:LoopOverActiveBuffs( time, unit, activeBuffs, "HARMFUL" ) ~= 0

	-- insert aoe's that don't have buffs
	local aoeBuffTable = self._curSpecInfo._aoeBuffTable
	if ( aoeBuffTable ~= nil ) then
		for key, aoeBuff in pairs( aoeBuffTable ) do
			local si = aoeBuff._spellInfo
			if ( si.spellType == HoJ.SPELL_SKILL_BUFF ) then
				-- this spell will never show up as an actual aura, so just add it manually
				local activeBuff = activeBuffs[si.id]
				if ( activeBuff ~= nil ) then
					local tr = aoeBuff._expiration - time
					if ( tr > 0.0 ) then
						-- HoJ:Log( format( "%s expires in %.2f seconds", si.name, tr ) )
						-- HoJ:Log( si.name .. " added to active buffs" )
						activeBuff.start = aoeBuff._start
						activeBuff.duration = aoeBuff._duration
						activeBuff._expirationTime = aoeBuff._expiration
						activeBuff.stacks = 1
					else
						activeBuff.start = 0.0
						activeBuff.duration = 0.0
						activeBuff._expirationTime = -1.0
						activeBuff.stacks = 0
					end
				end
			end
		end
	end

	--HoJ:DebugMsg( "_hasDebuff = " .. tostring( HoJ._hasDebuff ) )
end

-- -----------------------------------------------------------------------------
-- HoJ:IsBuffActive
-- TODO: change to spellInfo from spellId
-- -----------------------------------------------------------------------------
function HoJ:IsBuffActive( time, activeBuffList, spellId )
	if ( activeBuffList == nil ) then
        return false
	end

	-- local debug = spellId == 188370

	local buff = activeBuffList[spellId]
	if ( buff == nil ) then
		-- if this happens then the buff is missing from the activeBuffList for the current spec
		HoJ:DevError( format( "Buff for spellId %i == nil!", spellId ) )
		return false
	end

	local tr = ( buff._expirationTime - time )
	-- HoJ:LogD( debug, "tr = " .. tr .. ", _expire = " .. buff._expirationTime .. ", t = " .. time )

	return tr > 0.0
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuffStackCount
-- TODO: change spellId to spellInfo
-- -----------------------------------------------------------------------------
function HoJ:GetBuffStackCount( time, activeBuffList, spellId )
	if ( activeBuffList == nil ) then
		return 0
	end

	local buff = activeBuffList[spellId]
	if ( buff == nil ) then
		return 0
	end

	-- make sure the buff won't be expired at the passed time, and if it will be
	-- return 0
	local isExpired = ( buff._expirationTime - time ) <= 0.0
	if ( isExpired ) then
		return 0
	end

	return buff.stacks;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuff
-- -----------------------------------------------------------------------------
function HoJ:GetBuff( activeBuffList, spellId )
	return activeBuffList[spellId]
end

-- -----------------------------------------------------------------------------
-- HoJ:GetUnitBuffTimeRemaining
-- returns:
--	float 	buffTimeRemaining
--	int 	stacks
-- -----------------------------------------------------------------------------
function HoJ:GetUnitBuffTimeRemaining( time, unit, inSpellId, filter )
	for i = 1, 40 do
		local name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i, filter )
		if ( name == nil ) then
			return 0.0, 0
		end
		-- self:DebugMsg( "Index " .. i .. " " .. unit .. " buff " .. name .. " with id = " .. spellId )
		if ( spellId == inSpellId ) then
			--self:DebugMsg( "Index " .. i .. " " .. unit .. " buff " .. name .. " with id = " .. spellId )
			return expirationTime - time, count
		end
	end
	return 0.0, 0
end

-- local debugCooldowns = false

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingAfterGCD
-- For predicting what use next, we don't really care about the GCD period, we
-- just want to know if a spell will be usable once the GCD ends.
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingAfterGCD( rotationState, spellInfo )
	local start, duration, enabled = HoJ:GetSpellCooldown( spellInfo );
	if ( start <= 0.0 or duration <= 0.0 ) then
		-- not on cooldown at all
		return 0.0
	end

	local expirationTime = start + duration
	local gcdExpireTime = rotationState._time + rotationState._gcdRemaining

	local timeRemainingAfterGCD = expirationTime - gcdExpireTime
	if ( timeRemainingAfterGCD < 0.1 ) then -- 0.1 to account for any drift between GCD and CD
		timeRemainingAfterGCD = 0.0
	end
	-- HoJ:LogD( debugCooldowns, format( "%s CD - GCD = %.2f", HoJ:GetSpellDebugName( spellInfo ), timeRemainingAfterGCD ) )
	return timeRemainingAfterGCD
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingIgnoringGCD
-- normally when we check for a spell cooldown, all spells return the GCD cooldown,
-- i.e., even if the spell itself is not on cooldown, if the GCD is, then it will
-- appear that the spell is on cooldown.
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingIgnoringGCD( rotationState, spellInfo )
	local start, duration, enabled = HoJ:GetSpellCooldown( spellInfo );

	if ( start <= 0 or duration <= 0 ) then
		return 0.0
	end

	local endTime = start + duration
	local remaining = endTime - rotationState._time

	if ( rotationState._gcdRemaining > 0.0 ) then
		-- we're in GCD...
		if ( rotationState._gcdRemaining >= remaining ) then
			-- this is just a GCD cooldown, not the actual spell cooldown
			return 0.0
		end
	end
	return remaining
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuffTimeRemaining
-- returns:
--   time remaining in seconds
--   start time
--   duration in seconds
--   number of stacks

-- TODO: change to spellInfo instead of spellId
-- -----------------------------------------------------------------------------
function HoJ:GetBuffTimeRemaining( time, spellId )
	local buff = self._curSpecInfo._activeBuffList[spellId]
	if ( buff == nil ) then
		return 0.0, 0.0, 0.0, 0
	end

	local tr = buff._expirationTime - time
	if ( tr < 0.0 ) then
		tr = 0.0
	end
	return tr, buff.start, buff.duration, buff.stacks
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckProcdBuffs
-- -----------------------------------------------------------------------------
function HoJ:CheckProcdBuffs( rotationState, activeBuffList, procdBuffList )
	if ( not HoJ._showBuffsFrame ) then
		return;
	end

	if ( activeBuffList == nil ) then
		HoJ:Error( "activeBuffList is nil!" )
	end
	if ( procdBuffList == nil ) then
		HoJ:Error( "procdBuffList is nil!" )
	end

    for i, v in ipairs( procdBuffList ) do
		if ( v._spellInfo ~= nil ) then
			-- local debug = v._spellInfo.id == 188370
			-- HoJ:LogD( debug, "Check procd buff '" .. v._spellInfo.name .. "'" )
			if ( self:IsBuffActive( rotationState._time, activeBuffList, v._spellInfo.id ) ) then
				-- HoJ:LogD( debug, "Found procd buff '" .. v._spellInfo.name .. "'" )
				if ( not v._spellInfo._showTimeThreshold ) then
					self:AddIcon( self._buffs, v._spellInfo )
				else 
					local buff = activeBuffList[v._spellInfo.id]
					if ( buff ~= nil ) then
						local timeRemaining = buff._expirationTime - HoJ._curTime
						if ( timeRemaining <= v._spellInfo._showTimeThreshold ) then
							self:AddIcon( self._buffs, v._spellInfo )
						end
					end
				end
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:SpellInRange_SpellIndex
-- Wraps IsSpellInRange with error checking
--------------------------------------------------------------------------------
function HoJ:SpellInRange_SpellIndex( spellIndex, bookType, unit, debugName, debug )
	if ( spellIndex == nil or spellIndex < 0 ) then
		HoJ:DebugMsg( "Invalid spellIndex " .. spellIndex .. " for " .. debugName )
		return false
	end

	local inRange = IsSpellInRange( spellIndex, bookType, unit )

	if ( debug ) then
		local spellName = GetSpellBookItemName( spellIndex, bookType )
		HoJ:Log( format( "HoJ:SpellInRange_SpellIndex( %s, %s ) = %s for %s", 
			tostring( spellIndex ), tostring( bookType ), 
			tostring( inRange ),
			tostring( spellName) ) )
	end

	return inRange == 1
end

-- -----------------------------------------------------------------------------
-- HoJ:SpellInRange_SpellName
-- Wraps IsSpellInRange with error checking
--------------------------------------------------------------------------------
function HoJ:SpellInRange_SpellName( spellName, unit, debugName, debug )
	if ( spellName == nil ) then
		HoJ:Error( "Invalid spell name for " .. debugName )
		return false
	end

	local inRange = IsSpellInRange( spellName, unit )
	HoJ:LogD( debug, "IsSpellInRange( " .. spellName .. " ) = " .. tostring( inRange ) )
	return inRange == 1
end

-- -----------------------------------------------------------------------------
-- HoJ:InRange
-- -----------------------------------------------------------------------------
function HoJ:InRange( spellInfo, unit, debugName, debug )
	local inRange = false;

	HoJ:LogD( debug ~= nil and debug == true, "HoJ:InRange( " .. spellInfo.name .. " )" )

	if ( spellInfo.spellIndex ~= nil and spellInfo.spellIndex >= 0 ) then
		inRange = HoJ:SpellInRange_SpellIndex( spellInfo.spellIndex, spellInfo._bookType, unit, debugName, debug )
	else
		inRange = HoJ:SpellInRange_SpellName( spellInfo.name, unit, debugName, debug );
	end
	
	HoJ:LogD( debug, "InRange( " .. spellInfo.name .. " ) = " .. tostring( inRange) );
	return inRange;
end

-- -----------------------------------------------------------------------------
-- HoJ:WillOverflowHoPo
-- CLASSFIX - HoPo
-- -----------------------------------------------------------------------------
function HoJ:WillOverflowHoPo( spellInfo, curHoPo, maxHoPo )
	local overflow = ( curHoPo + spellInfo._powerGen ) - maxHoPo;
	-- HoJ:DebugMsg( spellInfo.name .. ": overflow = " .. overflow );
	return overflow;
end

-- -----------------------------------------------------------------------------
-- HoJ:IsOffCooldown
-- -----------------------------------------------------------------------------
function HoJ:IsOffCooldown( rotationState, spellInfo, debug )
	local id = spellInfo._cdId
	if ( id == nil ) then
		id = spellInfo.id
	end
	if ( HoJ._level < spellInfo.minLevel ) then
		return false, 100000.0
	end
	if ( spellInfo.spellType == HoJ.SPELL_TALENT and not HoJ:IsTalentSelected( spellInfo ) ) then
		return false, 1000000.0
	end
	local cdRemaining = HoJ:GetCDRemainingAfterGCD( rotationState, spellInfo );
	if ( debug ) then
		HoJ:Log( "spell " .. HoJ:GetSpellDebugName( spellInfo ) .. ": cdRemaining = " .. cdRemaining .. ", GCD Remaining = " .. rotationState._gcdRemaining );
	end
	if ( cdRemaining == nil ) then
		HoJ:DebugMsg( "cdRemaining == nil!" )
		return false, 1000000.0
	end
	return cdRemaining <= 0.0, cdRemaining
end

-- -----------------------------------------------------------------------------
-- ArrayContains
-- -----------------------------------------------------------------------------
function ArrayContains( t, value )
    for i, curValue in ipairs( t ) do
        if ( value == curValue ) then
            return true
        end
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:IsTalentSelected
-- -----------------------------------------------------------------------------
function HoJ:IsTalentSelected( spellInfo )
    if ( spellInfo._talentSpecs == nil ) then
        return false
    end

    if ( not ArrayContains( spellInfo._talentSpecs, self._curSpecInfo._spec ) ) then
		if ( HoJ._debugSpellName ~= nil and spellInfo.name == HoJ._debugSpellName ) then
			HoJ:DebugMsg( "Talent " .. spellInfo.name .. " does not match spec " .. HoJ._curSpecInfo._spec .. "!" )
		end
		return false
	end

	if ( spellInfo.talentRow == nil or spellInfo.talentCol == nil ) then
		if ( spellInfo.spellType ~= HoJ.SPELL_TALENT ) then
			HoJ:DebugMsg( "IsTalentSelected() called on non-talent " .. spellInfo.name );
		else
			HoJ:DebugMsg( "Talent " .. spellInfo.name .. " missing row and column" );
		end
		return true;
	end

	local _, _, _, selected, available = GetTalentInfo( spellInfo.talentRow, spellInfo.talentCol, 1 );
	if ( HoJ._debugSpellName ~= nil and spellInfo.name == HoJ._debugSpellName ) then
		HoJ:DebugMsg( "Talent " .. spellInfo.name .. ", available = " .. tostring( available ) .. ", selected = " .. tostring( selected ) )
	end
	return ( available and selected );
end

-- -----------------------------------------------------------------------------
-- HoJ:GetGroupType
-- -----------------------------------------------------------------------------
function HoJ:GetGroupType()
	if ( IsInRaid() ) then
		return HoJ.GROUP_RAID
	elseif ( IsInGroup() ) then
		return HoJ.GROUP_PARTY
	end
	return HoJ.GROUP_NONE
end

-- -----------------------------------------------------------------------------
-- HoJ:MaxMemberForGroupType
-- The player doesn't count as a partyN unit id, so if there are 3 party members,
-- only "party1" and "party2" are valid unit ids.
-- If the player is in a raid, however, this is not true. If there are 3 raid
-- members, then "raid1", "raid2", and "raid3" are all valid and one of them will
-- represent the "player".
-- To account for this, we subtract one from the number of group members when
-- the player is in a party, but not when in a raid.
-- -----------------------------------------------------------------------------
function HoJ:MaxMemberForGroupType( groupType )
	if ( groupType == HoJ.GROUP_PARTY ) then
		return GetNumGroupMembers() - 1
	end
	return GetNumGroupMembers()
end

-- -----------------------------------------------------------------------------
-- HoJ:PrefixForGroupType
-- -----------------------------------------------------------------------------
function HoJ:PrefixForGroupType( groupType )
	if ( IsInRaid() ) then
		return "raid"
	elseif ( IsInGroup() ) then
		return "party"
	end
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:UnitIDForName
-- -----------------------------------------------------------------------------
function HoJ:UnitIDForName( name )
	if ( name == HoJ._playerName ) then
		return "player"
	end

	local prefix = HoJ:PrefixForGroupType( HoJ._groupType )
	if ( prefix == nil ) then
		return nil
	end

	local max = HoJ:MaxMemberForGroupType( HoJ:GetGroupType() )
	for i = 1, max do
		local unitID = prefix .. i
		local memberName = UnitName( unitID )
		if ( memberName == name ) then
			return unitID
		end
	end
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetBlessing
-- -----------------------------------------------------------------------------
function HoJ:ResetBlessing( blessing )
	blessing.targetName = nil
	blessing.start = 0.0
	blessing.duration = 0.0
	blessing.updateIndex = -1
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetBlessings
-- -----------------------------------------------------------------------------
function HoJ:ResetBlessings( blessings )
    if ( blessings == nil ) then
        return 
    end
	for spellId, blessing in pairs( blessings) do
		--HoJ:DebugMsg( "Resetting " .. blessing.spellInfo.name .. "..." )
		HoJ:ResetBlessing( blessing )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateBlessingFromUnitID
-- returns true if the unitID matches the name, in which case we can definitively
-- know that the unit is the one that was cast on.
-- -----------------------------------------------------------------------------
function HoJ:UpdateBlessingFromUnitID( time, blessing, unitID )
	--HoJ:DebugMsg( "Checking for " .. blessing.spellInfo.name .. " on " .. unitID )
	if ( unitID == nil ) then
		--HoJ:FrameThrottleMsg( 75, "Error: unitID == nil!" )
		return false
	end

	local unitName = UnitName( unitID )
	if ( unitName == nil ) then
		--HoJ:FrameThrottleMsg( 75, "Error: UnitName() == nil for unit id '" .. unitID .. "'!" )
		return false
	end

	if ( blessing.targetName ~= unitName ) then
		return false
	end

	--HoJ:DebugMsg( "Blessing target is " .. unitID .. "( " .. unitName .. " )" )

	local found = false
	local i = 1
	local name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	while name do
		if ( spellId == blessing.spellInfo.id and unitCaster == "player" ) then
			-- this is the spell we cast
			--HoJ:DebugMsg( "Updating " .. blessing.spellInfo.name .. " from " .. unitName )
			
			-- assume buffs are expired if < 1 second, because there is a bug where Wisdom sometimes sticks at
			-- 0 seconds remaining and causes the active buff to flash on. Even if we do this here, we also need
			-- to be sure that we won't alsojust re-add the blessing in FindBlessingOnUnitID because it wasn't
			-- found on a known target.
			local tr = expirationTime - time
			if ( tr > 0.0 ) then
				found = true
			end
			if ( HoJ._debugBlessings ) then
				HoJ:DebugMsg( "UpdateBlessing: " .. blessing.spellInfo.name .. " duration = " .. duration .. ", expire = " .. expirationTime .. ", remain = " .. timeRemaining )
			end

			local start = expirationTime - duration
			blessing.start = start
			blessing.duration = duration
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	end
	-- not found on the unit it was cast on, so it's expired or been canceled
	if ( not found ) then
		--HoJ:DebugMsg( blessing.spellInfo.name .. " was not found. Resetting." )
		HoJ:ResetBlessing( blessing )
	end

	return true
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessingOnUnitID
-- -----------------------------------------------------------------------------
function HoJ:FindBlessingOnUnitID( time, blessing, unitID )
	--HoJ:DebugMsg( "Checking for " .. blessing.spellInfo.name .. " on " .. unitID )
	local i = 1
	local name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	while name do
		if ( spellId == blessing.spellInfo.id and unitCaster == "player" ) then
			-- this is the spell we cast
			--HoJ:DebugMsg( "Found " .. blessing.spellInfo.name .. " on " .. unitID )
			local tr = expirationTime - time
			if ( HoJ._debugBlessings ) then
				HoJ:DebugMsg( "FindBlessing: " .. blessing.spellInfo.name .. " duration = " .. duration .. ", expire = " .. expirationTime .. ", remain = " .. timeRemaining )
			end			
			if ( tr <= 0.0 ) then
				-- blessing sometimes remain on the unit beyond their expiration time
				return false
			end
			local start = expirationTime - duration
			blessing.start = start
			blessing.duration = duration
			return true
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	end
	return false
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessing
-- -----------------------------------------------------------------------------
function HoJ:UnitIDForUpdateIndex( index, groupType )
	if ( index < 0 ) then
		return "player"
	elseif ( index == 0 ) then
		return "target"
	end

	local max = HoJ:MaxMemberForGroupType( groupType )
	local prefix = HoJ:PrefixForGroupType( groupType )
	if ( prefix == nil ) then
		return nil
	end

	local i = index
	if ( i > max ) then
		i = max
	end

	local unitID = prefix .. i
	return unitID
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessing
-- -----------------------------------------------------------------------------
function HoJ:FindBlessing( time, blessing )
	--HoJ:DebugMsg( "FindBlessing " .. blessing.spellInfo.name )

	local unitID = HoJ:UnitIDForUpdateIndex( blessing.updateIndex, self._groupType, self._numGroupMembers )
	if ( unitID == nil ) then
		return
	end

	if ( HoJ:FindBlessingOnUnitID( time, blessing, unitID ) ) then
		-- store the name so we can use the direct lookup
		blessing.targetName = UnitName( unitID )
		return
	end

	-- start over from the beginning
	blessing.updateIndex = blessing.updateIndex + 1
	if ( blessing.updateIndex > self._numGroupMembers ) then
		blessing.updateIndex = -1
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateBlessings
-- For our purposes here a blessing is a buff that can only be on a single target
-- at a time and must be tracked for that target.

-- We can't rely on a blessing being set on a target immediately after cast due
-- to network lag.
-- We can't rely on a blessing being set on the player immediately after they
-- enter the worl due to network lag.
-- We also cannot rely on being able to check the buff on the a player the buff
-- has been cast on if that player is not in the player's party.
-- For this reason, if we don't find a blessing on a target we have to treat it
-- differently based on whether or not the buff target is in the player's party.

-- To deal with all of the above, when a blessing is cast, we record the name
-- of the target it was cast on (not the unitID it was cast on).

-- Each time we check for blessings, if the blessing has a recorded targetName,
-- we cycle through unit IDs ("player",
-- "target" or and "partyN"/"raidN" for our current group) to try to find the
-- one that matches the recorded name.

-- If we don't find the a unit ID with a matching name, we just check for
-- buff expiration using the stored state.

-- If we do find a unit ID with a matching name, we search the unit's auras for
-- the blessing. If we don't find it, we know it's expired or canceled and we
-- can confidently show the buff icon to indicate it has expired.
--
-- If the buff has no recorded name, then we search through auras on the "player",
-- "target", and "partyN" or "raidN" unitIDs to find a matching buff cast by
-- the player. If we find it, we then store the name of that unit. This handles
-- initializing the buff tracking at any time after the player enters the world.
-- If any player buff is on another player that can be inspected via a unit ID
-- then it will start being tracked. This search through auras on all unit IDs
-- could be very expensive to do all at once for a large raid, so this test is
-- amortized and will check one unit ID per update frame (currently @ 15 fps).
-- This means it could take up to 40 / 15 seconds to find a buff after logging
-- in.

-- table of flags for each buf indexed by spell id
--	{
--		targetName,	-- name of the target the spell was cast on
--		start,		-- time buff started
--		duration,	-- duration of the buff
--		updateIndex	-- index of the current party member to check auras for
--	}
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function HoJ:UpdateBlessings( time, blessings )
	-- HoJ:DebugMsg( "UpdateBlessings: " .. tostring( blessings ) )
	if ( blessings == nil ) then
		return
	end

	local groupType = HoJ:GetGroupType()
	local numGroupMembers = GetNumGroupMembers()
	if ( groupType ~= self._groupType ) then
		-- group changed so restart the buff checks
		--HoJ:DebugMsg( "Resetting update due to group change!" )
		for spellId, blessing in pairs( blessings ) do
			blessing.updateIndex = -1
		end
	elseif ( numGroupMembers ~= self._numGroupMembers ) then
		-- number of group members changed, so restart the buff checks if we're beyond the number of group members
		-- HoJ:DebugMsg( "Resetting update due to group size change!" )
		local max = HoJ:MaxMemberForGroupType( self._groupType )
		for spellId, blessing in pairs( blessings ) do
			if ( blessing.updateIndex >= max ) then
				blessing.updateIndex = -1
			end
		end
	end

	self._groupType = groupType
	self._numGroupMembers = numGroupMembers

	for spellId, blessing in pairs( blessings ) do
		if ( blessing.targetName ~= nil ) then
			-- HoJ:DebugMsg( "Blessing " .. blessing.spellInfo.name .. " targets " .. blessing.targetName )
			if ( not HoJ:UpdateBlessingFromUnitID( time, blessing, "player" ) ) then
				if ( not HoJ:UpdateBlessingFromUnitID( time, blessing, "target" ) ) then
					if ( HoJ._groupType ~= HoJ.GROUP_NONE ) then

						local unitID = HoJ:UnitIDForName( blessing.targetName )
						if ( unitID ~= nil ) then
							if ( HoJ:UpdateBlessingFromUnitID( time, blessing, unitID ) ) then
								found = true
								break
							end
						end
					end
				end
			end
			-- check for expiration
			local expireTime = blessing.start + blessing.duration
			local timeRemaining = expireTime - time
			if ( timeRemaining <= 0.0 ) then
				HoJ:ResetBlessing( blessing )
			end
		else
			HoJ:FindBlessing( time, blessing )
		end
	end
	HoJ._nextUpdateBlessingFrame = self._updateFrameNum + self._updateBlessingsInterval
end

-- -----------------------------------------------------------------------------
-- HoJ:GetPowerAccountingForPooling
-- -----------------------------------------------------------------------------
function HoJ:GetPowerAccountingForPooling( rotationState )
	if ( self._curSpecInfo._usesPowerPooling ) then
		-- TODO: this is a hack -- it would be better to remove the power check from skill 
		-- rules that should wait on pooling
		-- report as if we have max power. This makes the rotation simply choose the next
		-- best skill and wait on the energy to pool
		return self._curSpecInfo._powerMax
	end

	return rotationState._power
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckPower
-- -----------------------------------------------------------------------------
function HoJ:CheckPower( rotationState, rule )
	local manaCost = rule.spellInfo._manaCost
	if ( manaCost ~= nil and self._mana < manaCost ) then
		return false
	end

	local powerGen = rule.spellInfo._powerGen
	if ( powerGen == nil ) then
		powerGen = 0
	end
	if ( powerGen == 0 ) then
		-- skill does not use or generate power
		return true
	end
	
	local power = self:GetPowerAccountingForPooling( rotationState )
	if ( powerGen < 0 ) then
		-- the skill uses power
		local powerUsed = - rule.spellInfo._powerGen

		-- self:LogD( rule.spellInfo.id == 85256, -- TV
		--	format( "power = %i, powerUsed = %i", power, powerUsed ) ) 
		
		return ( power >= powerUsed ) 
	end

	-- the skill generates power
	local powerAfter = power + powerGen

	local powerOver = powerAfter - self._curSpecInfo._powerMax 
	if ( powerOver <= 0 ) then
		return true
	end
	if ( rule._maxPowerWaste ~= nil ) then
		return ( powerOver <= rule._maxPowerWaste )
	end

	return false
end

-- -----------------------------------------------------------------------------
-- Data-driven Skill Selection
-- -----------------------------------------------------------------------------
HoJ.OP_NONE = -999

-- boolean comparison operators
HoJ.OP_LESS_EQUAL = -2
HoJ.OP_LESS = -1
HoJ.OP_EQUAL = 0
HoJ.OP_GREATER = 1
HoJ.OP_GREATER_EQUAL = 2
HoJ.OP_NOT_EQUAL = 3

-- logic operators
HoJ.OP_AND = 1000
HoJ.OP_OR = 1001

-- range operator
HoJ.OP_IN_RANGE_INCLUSIVE = 1002 -- requires a const array { min, max }, true if >= min and <= max, i.e. [ min, max ]
HoJ.OP_IN_RANGE_MIN_EX_MAX_IN = 1003 -- requires a const array { min, max }, true if > min and <= max, i.e. ( min, max ]
HoJ.OP_IN_RANGE_MIN_IN_MAX_EX = 1004 -- requires a const array { min, max }, true if >= min and < max, i.e. [ min, max )
HoJ.OP_IN_RANGE_MIN_EX_MAX_EX = 1005 -- requires a const array { min, max }, true if > min and < max, i.e. ( min, max )

-- table
HoJ.OP_TABLE_CONTAINS = 2000 -- if the value is in the table

-- TODO: need a way to group operations

-- constant types
HoJ.TYPE_CONSTANT = 0 	-- use the constant specified by aValue or bValue

-- result types
HoJ.TYPE_RESULT = 1     -- variale is the result of a previous evaluation

-- variable types
HoJ.TYPE_POWER 					        = 1000	-- get player's current power level (holy, rage, etc.)
HoJ.TYPE_PLAYER_HEALTH_RATIO			= 1001	-- get the target's health ratio
HoJ.TYPE_TARGET_HEALTH_RATIO			= 1002	-- get the target's health ratio
HoJ.TYPE_PLAYER_BUFF_ACTIVE				= 1003	-- (constant = spellId) true if the buff specified by the const parm is active
HoJ.TYPE_TARGET_RANGE					= 1004	-- get the target's range with const as the range type
HoJ.TYPE_TALENT_SELECTED				= 1005	-- true if the talent specified by the const parm (spellInfo) is selected
HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING	= 1006	-- time remaining for a debuff applied by the player
HoJ.TYPE_PLAYER_BUFF_REMAINING			= 1007	-- time remaining for a buff on the player
HoJ.TYPE_SKILL_CHARGES                  = 1008	-- number of current charges a skill has (for instance, crusader strike has up to 2 charges)
HoJ.TYPE_NUM_MELEE_ATTACKERS			= 1009	-- number of enemies attacking the player with attacks <= 8 yards
HoJ.TYPE_PLAYER_BUFF_STACK_COUNT		= 1010	-- (constant = spellInfo) number of stacks a buff (on player only) has
HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO		= 1011	-- minor heals will show in rotation if health ratio < this
HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO		= 1012	-- major heals will show in rotation if health ratio < this
HoJ.TYPE_TARGET_IS_BOSS					= 1014	-- maps to HoJ._targetIsBoss
HoJ.TYPE_SKILL_OFF_COOLDOWN				= 1015	-- returns true if the skill is off cooldown
HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED		= 1016	-- returns true if the target can be interrupted
HoJ.TYPE_COMBO_POINTS					= 1017	-- current combo points
HoJ.TYPE_PLAYER_IS_STEALTHED			= 1018	-- true if player is stealthed
HoJ.TYPE_INCOMING_DPS_RATIO				= 1019	-- returns damage per second ratio
HoJ.TYPE_SECONDS_SINCE_DAMAGED_BY_SPELL	= 1020	-- seconds since the player last took spell damage
HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE		= 1021	-- seconds until target is predicted to die, based on combat session dps
HoJ.TYPE_SPELL_DURATION					= 1022	-- duration a skill or spell's effect lasts, in seconds
HoJ.TYPE_TARGET_IS_TOUGH				= 1023	-- returns HoJ._targetIsTough
HoJ.TYPE_COMBAT_ATTACKS_HIT				= 1024	-- returns number of attacks that have hit during current combat session
HoJ.TYPE_SKILL_POWER_COST				= 1025	-- cost of the specified skill
HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF		= 1026	-- returns true if the target has the specified debuff
HoJ.TYPE_SKILL_COOLDOWN_REMAINING		= 1027	-- returns time remaining on the spell's cooldown, not including GCD
HoJ.TYPE_UNIT_IS_SLOWED					= 1028	-- returns true if the unit is under some slowing effect, determined by checking speed
HoJ.TYPE_NET_INCOMING_DPS				= 1029	-- returns net incoming damage per second
HoJ.TYPE_NET_INCOMING_DPS_RATIO			= 1030	-- returns net incoming damage per second ratio
HoJ.TYPE_NET_INCOMING_DPS_PCT			= 1031	-- returns net incoming damage per second ratio as a percentage
HoJ.TYPE_PLAYER_TIME_TO_LIVE			= 1032	-- returns predicted time before player will die
HoJ.TYPE_PLAYER_TIME_TO_NEAR_DEATH		= 1033	-- returns predicted time before player will be at 10% health
HoJ.TYPE_NET_INCOMING_DPS_CHR			= 1034	-- returns ( net incoming dps / current health ) ratio
HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL	= 1035	-- returns true if a known major damage spell is being cast by the target
HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS	= 1036	-- returns the expected duration of a buff applied for combo points
HoJ.TYPE_POWER_AFTER_SKILL_USED			= 1037	-- returns the amount of power that would be remaining after using the specified spellInfo
HoJ.TYPE_POWER_AFTER_CD_MINUS_SKILL_POWER	= 1038	-- returns the power at the time constant1 skill comes off cooldown, if constant2 skill is used
HoJ.TYPE_SKILL_POWER_GEN				= 1039	-- power generated by the specified skill
HoJ.TYPE_POWER_MAX						= 1040	-- max power
HoJ.TYPE_SPELL_CREATURE_TYPES			= 1041	-- returns creature types spell is usable on
HoJ.TYPE_TARGET_CREATURE_TYPE			= 1042	-- returns the creature type ("Humanoid", "Demon", etc.) of the target
HoJ.TYPE_IS_SHIELD_EQUIPPED				= 1043	-- returns true if the player has a shiled equipped in the offhand
HoJ.TYPE_PLAYER_LEVEL					= 1044	-- returns the current level of the player
HoJ.TYPE_SPELL_MIN_LEVEL				= 1045	-- returns the minLevel that a spell becomes usable / learned
HoJ.TYPE_UNIT_IS_DEAD					= 1046	-- returns true f the unit specified by constant1 is dead
HoJ.TYPE_CAN_USE_EXECUTE				= 1047	-- true if spell, constant, is off cooldown and target health ratio is < constant2
HoJ.TYPE_TARGET_IS_FRIEND				= 1048	-- true if the current target is friendly
HoJ.TYPE_TARGET_IS_PLAYER				= 1049	-- true if the target is a player
HoJ.TYPE_PLAYER_IN_COMBAT				= 1050	-- true if the player is in combat
HoJ.TYPE_IS_TANKING_TARGET				= 1051	-- returns true if tanking the target
HoJ.TYPE_HAS_AGGRO						= 1052	-- returns true if the unit has aggro
HoJ.TYPE_SPEC_CHOSEN					= 1053	-- returns false if the player hasn't chosen a spec yet
HoJ.TYPE_POWER_MAX_ADJUSTED				= 1054  -- returns the spec's max power minus constant 1
HoJ.TYPE_TIME_SINCE_LAST_CAST			= 1055	-- time since the spell was last cast
HoJ.TYPE_COMBO_POINTS_DEFICIT			= 1056	-- max combo points - current combo points
HoJ.TYPE_GCD_REMAINING					= 1057	-- time remaining on Global Cooldown
HoJ.TYPE_GCD_DURATION					= 1058	-- total duration of the player's GCD

local _varTable = {
	[HoJ.TYPE_RESULT] = function( rotationState, constant, constant2, results )
		if ( results == nil ) then
			HoJ:Error( "Type is a result but results list == nil!" )
			return 0
		elseif ( constant == nil ) then
			HoJ:Error( "Type is a result but constant == nil! constant  must specify a result index!" )
			return 0
		else
			return results[constant]
		end
	end,
	[HoJ.TYPE_POWER] = function( rotationState, constant, constant2, results )
		local power = HoJ:GetPowerAccountingForPooling( rotationState )
		--HoJ:LogD( type( constant ) == "table" and constant.id == 85256, "power for TV = " .. tostring( power ) )
		return power
	end,
	[HoJ.TYPE_PLAYER_HEALTH_RATIO] = function( rotationState, constant, constant2, results ) 
		return HoJ._playerHealthRatio 
	end,
	[HoJ.TYPE_TARGET_HEALTH_RATIO] = function( rotationState, constant, constant2, results ) 
		return HoJ._targetHealthRatio 
	end,
	[HoJ.TYPE_PLAYER_BUFF_ACTIVE] = function( rotationState, constant, constant2, results ) 
		return HoJ:IsBuffActive( rotationState._time, HoJ._curSpecInfo._activeBuffList, constant.id )
	end,
	[HoJ.TYPE_TARGET_RANGE] = function( rotationState, constant, constant2, results ) 
		return HoJ._targetInRange[constant] 
	end,
	[HoJ.TYPE_TALENT_SELECTED] = function( rotationState, constant, constant2, results ) 
		return HoJ:IsTalentSelected( constant ) 
	end,
	[HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING] = function( rotationState, constant, constant2, results ) 
		return HoJ:GetUnitBuffTimeRemaining( rotationState._time, "target", constant.id, "PLAYER|HARMFUL" )
	end,
	[HoJ.TYPE_PLAYER_BUFF_REMAINING] = function( rotationState, constant, constant2, results ) 
		return HoJ:GetBuffTimeRemaining( rotationState._time, constant.id )
	end,
	[HoJ.TYPE_SKILL_CHARGES] = function( rotationState, constant, constant2, results )
		if ( constant == nil ) then
			HoJ:Error( "TYPE_SKILL_CHARGES constant must be valid spellInfo!" )
			return 0
		end
		if ( constant.id == nil or constant.id == 0 ) then
			HoJ:Error( "TYPE_SKILL_CHARGES constant.id = " .. tostring( constant.id ) )
			return 0
		end

		local numCharges, maxCharges = GetSpellCharges( constant.id )
		if ( numCharges == nil ) then
			HoJ:Error( "GetSpellCharges(" .. tostring( constant.id ) .. ") == nil! Check spellInfo for " .. constant.name .. " in rule!" )
			return 0
		end

		if ( numCharges == 0 and rotationState._minSkillCharges > 0 ) then
			-- HoJ:LogD( debugCooldowns, format( "TYPE_SKILL_CHARGES = %i (min)", rotationState._minSkillCharges ) )
			return rotationState._minSkillCharges
		end

		-- HoJ:LogD( debugCooldowns, format( "TYPE_SKILL_CHARGES = %i", numCharges ) )
		return numCharges 
	end,
	[HoJ.TYPE_NUM_MELEE_ATTACKERS] = function( rotationState, constant, constant2, results ) 
		return HoJ._numMeleeAttackers 
	end,
	[HoJ.TYPE_PLAYER_BUFF_STACK_COUNT] = function( rotationState, constant, constant2, results ) 
		return HoJ:GetBuffStackCount( rotationState._time, HoJ._curSpecInfo._activeBuffList, constant.id ) 
	end,
	[HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO] = function( rotationState, constant, constant2, results ) 
		return HoJ._curSpecInfo._minorHealHealthRatio 
	end,
	[HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO] = function( rotationState, constant, constant2, results ) 
		return HoJ._curSpecInfo._majorHealHealthRatio 
	end,
	[HoJ.TYPE_TARGET_IS_BOSS] = function( rotationState, constant, constant2, results ) 
		--HoJ:DebugMsg( "targetIsBoss = " .. tostring( HoJ._targetIsBoss ) )
		return HoJ._targetIsBoss
	end,
	[HoJ.TYPE_SKILL_OFF_COOLDOWN] = function( rotationState, constant, constant2, results ) 
		return HoJ:IsOffCooldown( rotationState, constant )
	end,
	[HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED] = function( rotationState, constant, constant2, results ) 
		return HoJ._targetCanBeInterrupted
	end,
	[HoJ.TYPE_COMBO_POINTS] = function( rotationState, constant, constant2, results )
		return HoJ._comboPoints
	end,
	[HoJ.TYPE_PLAYER_IS_STEALTHED] = function( rotationState, constant, constant2, results )
		return HoJ._isStealthed
	end,
	[HoJ.TYPE_INCOMING_DPS_RATIO] = function( rotationState, constant, constant2, results )
		local dps = HoJ._incomingDpsRatio * constant
		-- HoJ:DebugMsg( "dps = " .. dps )
		return dps
	end,
	[HoJ.TYPE_SECONDS_SINCE_DAMAGED_BY_SPELL] = function( rotationState, constant, constant2, results )
		local t = rotationState._time - HoJ._lastSpellDamageTime
		-- HoJ:DebugMsg( "Seconds since damaged by spell = " .. t )
		return t
	end,
	[HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE] = function( rotationState, constant, constant2, results )
		local tttl = ( HoJ._combatDpsStats._predictedTargetDeathTime - rotationState._time ) * constant
		if ( tttl < 0.0 ) then 
			tttl = 0.0
		end
		-- HoJ:DebugMsg( format( "Target TTL * %.2f = %.2f", constant, tttl ) )
		return tttl
	end,
	[HoJ.TYPE_SPELL_DURATION] = function( rotationState, constant, constant2, results )
		if ( constant == nil or constant._duration == nil ) then
			return 0
		end
		-- HoJ:DebugMsg( constant.name .. "._duration = " .. constant._duration )
		return constant._duration
	end,
	[HoJ.TYPE_TARGET_IS_TOUGH] = function( rotationState, constant, constant2, results )
		return HoJ._targetIsTough
	end,
	[HoJ.TYPE_COMBAT_ATTACKS_HIT] = function( rotationState, constant, constant2, results )
		-- HoJ:DebugMsg( "Combat attacks hit = " .. HoJ._combatDpsStats._attacksHit )
		return HoJ._combatDpsStats._attacksHit
	end,
	[HoJ.TYPE_SKILL_POWER_COST] = function( rotationState, constant, constant2, results )
		return -constant._powerGen
	end,
	[HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF] = function( rotationState, constant, constant2, results ) 
		return HoJ:GetUnitBuffTimeRemaining( rotationState._time, "target", constant.id, "PLAYER|HARMFUL" ) > 0.0
	end,
	[HoJ.TYPE_SKILL_COOLDOWN_REMAINING] = function( rotationState, constant, constant2, results ) 
		if ( HoJ._level < constant.minLevel ) then
			return 1000000.0
		end
		if ( constant.spellType == HoJ.SPELL_TALENT and not HoJ:IsTalentSelected( constant ) ) then
			return 1000000.0
		end
		return HoJ:GetCDRemainingAfterGCD( rotationState, constant )
	end,
	[HoJ.TYPE_UNIT_IS_SLOWED] = function( rotationState, constant, constant2, results ) 
		if ( not HoJ._hasDebuff ) then
			return false
		end
		if ( HoJ._controlLost ) then
			return true
		end
		local speed = GetUnitSpeed( constant )
		return speed > 0 and ( ( speed / BASE_MOVEMENT_SPEED ) < 1.0 )
	end,
	[HoJ.TYPE_NET_INCOMING_DPS] = function( rotationState, constant, constant2, results )
		local dps = HoJ._netIncomingDps * constant
		-- HoJ:DebugMsg( "netIncomingDps = " .. dps )
		return dps
	end,
	[HoJ.TYPE_NET_INCOMING_DPS_RATIO] = function( rotationState, constant, constant2, results )
		local dps = ( HoJ._netIncomingDps * constant ) / HoJ._playerMaxHealth
		-- HoJ:DebugMsg( "netIncomingDpsRatio = " .. dps )
		return dps
	end,
	[HoJ.TYPE_NET_INCOMING_DPS_PCT] = function( rotationState, constant, constant2, results )
		local dps = ( ( HoJ._netIncomingDps * constant ) / HoJ._playerMaxHealth ) * 100.0
		-- HoJ:DebugMsg( "netIncomingDpsRatio = " .. dps )
		return dps
	end,
	[HoJ.TYPE_PLAYER_TIME_TO_LIVE] = function( rotationState, constant, constant2, results )
		return HoJ._playerTimeToLive
	end,
	[HoJ.TYPE_PLAYER_TIME_TO_NEAR_DEATH] = function( rotationState, constant, constant2, results )
		return HoJ._playerTimeToNearDeath
	end,
	[HoJ.TYPE_NET_INCOMING_DPS_CHR] = function( rotationState, constant, constant2, results )
		local dpschr = ( HoJ._netIncomingDps * constant ) / HoJ._playerHealth
		return dpschr
	end,
	[HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL] = function( rotationState, constant, constant2, results )
		return ( HoJ._incomingMajorDamage ~= nil )
	end,
	[HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS] = function( rotationState, constant, constant2, results )
		if ( constant._durationForComboPoints == nil or HoJ._comboPoints == 0 ) then
			return 0
		end
		local duration = constant._durationForComboPoints[HoJ._comboPoints]
		-- allow scaling by constant2
		if ( constant2 ~= nil and constant2 > 0.0 ) then
			duration = duration * constant2
		end
		-- HoJ:DebugMsg( format( "%s DURATION_FOR_COMBO_POINTS = %.2f", constant.name, duration ) )
		return duration
	end,
	[HoJ.TYPE_POWER_AFTER_SKILL_USED] = function( rotationState, constant, constant2, results )
		return rs._power + constant._powerGen
	end,
	[HoJ.TYPE_POWER_AFTER_CD_MINUS_SKILL_POWER] = function( rotationState, constant, constant2, results )
		local cdRemaining = HoJ:GetCDRemainingIgnoringGCD( rotationState, constant )
		local powerUsed = constant2._powerGen
		local futurePower = ( rotationState._power + powerUsed ) + ( cdRemaining * HoJ._actualPowerPerSec )
		return futurePower
	end,
	[HoJ.TYPE_SKILL_POWER_GEN] = function( rotationState, constant, constant2, results )
		return constant._powerGen
	end,
	[HoJ.TYPE_POWER_MAX] = function( rotationState, constant, constant2, results )
		return HoJ._curSpecInfo._powerMax
	end,
	[HoJ.TYPE_SPELL_CREATURE_TYPES] = function( rotationState, constant, constant2, results )
		return constant._creatureTypes
	end,
	[HoJ.TYPE_TARGET_CREATURE_TYPE] = function( rotationState, constant, constant2, results )
		return HoJ._targetCreatureType
	end,
	[HoJ.TYPE_IS_SHIELD_EQUIPPED] = function( rotationState, constant, constant2, results )
		return C_PaperDollInfo.OffhandHasShield()
	end,
	[HoJ.TYPE_PLAYER_LEVEL] = function( rotationState, constant, constant2, results )
		return HoJ._level
	end,
	[HoJ.TYPE_SPELL_MIN_LEVEL] = function( rotationState, constant, constant2, results )
		return constant.minLevel
	end,
	[HoJ.TYPE_UNIT_IS_DEAD] = function( rotationState, constant, constant2, results )
		return UnitIsDead( constant ) or UnitHealth( constant ) <= 0
	end,
	[HoJ.TYPE_CAN_USE_EXECUTE] = function( rotationState, constant, constant2, results )
		return HoJ:IsOffCooldown( rotationState, constant ) and HoJ._targetExists and HoJ._targetHealthRatio <= constant2
	end,
	[HoJ.TYPE_TARGET_IS_FRIEND] = function( rotationState, constant, constant2, results )
		return HoJ._targetIsFriend
	end,
	[HoJ.TYPE_TARGET_IS_PLAYER] = function( rotationState, constant, constant2, results )
		return HoJ._targetIsPlayer
	end,
	[HoJ.TYPE_PLAYER_IN_COMBAT] = function( rotationState, constant, constant2, results )
		return HoJ._inCombat
	end,
	[HoJ.TYPE_IS_TANKING_TARGET] = function( rotationState, constant, constant2, results )
		return HoJ._threat == HoJ.THREAT_AGGRO_LOST or HoJ._threat == HoJ.THREAT_AGGRO_HELD
	end,
	[HoJ.TYPE_HAS_AGGRO] = function( rotationState, constant, constant2, results )
		return HoJ._threat == HoJ.THREAT_AGGRO_PULLED or HoJ._threat == HoJ.THREAT_AGGRO_HELD
	end,
	[HoJ.TYPE_SPEC_CHOSEN] = function( rotationState, constant, constant2, results )
		return HoJ._specChosen
	end,
	[HoJ.TYPE_POWER_MAX_ADJUSTED] = function( rotationState, constant, constant2, results )
		return HoJ._curSpecInfo._powerMax + constant
	end,
	[HoJ.TYPE_TIME_SINCE_LAST_CAST] = function( rotationState, constant, constant2, results )
		return HoJ._curTime - constant._lastCastTime
	end,
	[HoJ.TYPE_COMBO_POINTS_DEFICIT] = function( rotationState, constant, constant2, results )
		return HoJ._curSpecInfo._comboMax - HoJ._comboPoints
	end,
	[HoJ.TYPE_GCD_REMAINING] = function( rotationState, constant, constant2, results )
		return HoJ._gcdRemaining * constant
	end,
	[HoJ.TYPE_GCD_DURATION] = function( rotationState, constant, constant2, results )
		return HoJ._gcdDuration * constant
	end,
}

local _opTable = {
	[HoJ.OP_LESS_EQUAL] = function( a, b )
		return a <= b
	end,
	[HoJ.OP_LESS] = function( a, b )
		return a < b
	end,
	[HoJ.OP_EQUAL] = function( a, b )
		--HoJ:DebugMsg( "OP_EQUAL: ( " .. tostring( a ) .. " == " .. tostring( b ) .. " ) = " .. tostring( a == b ) )
		return a == b
	end,
	[HoJ.OP_GREATER] = function( a, b )
		return a > b
	end,
	[HoJ.OP_GREATER_EQUAL] = function( a, b )
		return a >= b
	end,
	[HoJ.OP_NOT_EQUAL] = function( a, b )
		return a ~= b
	end,
	[HoJ.OP_AND] = function( a, b )
		return a and b
	end,
	[HoJ.OP_OR] = function( a, b )
		return a or b
	end,
	[HoJ.OP_IN_RANGE_INCLUSIVE] = function( a, b )
		return a >= b[1] and a <= b[2]
	end,
	[HoJ.OP_IN_RANGE_MIN_EX_MAX_IN] = function( a, b )
		return a > b[1] and a <= b[2]
	end,
	[HoJ.OP_IN_RANGE_MIN_IN_MAX_EX] = function( a, b )
		return a >= b[1] and a < b[2]
	end,
	[HoJ.OP_IN_RANGE_MIN_EX_MAX_EX] = function( a, b )
		return a > b[1] and a < b[2]
	end,
	[HoJ.OP_TABLE_CONTAINS] = function( a, b )
		local aType = type( a )
		if ( aType ~= "table" ) then
			HoJ:Error( "Type A for OP_TABLE_CONTAINS is '" .. aType .. "' but must be 'table'!" )
			return false
		end
		return a[b] ~= nil
	end,
}

local _opNameTable = {
	[HoJ.OP_LESS_EQUAL] = "<=",
	[HoJ.OP_LESS] = "<",
	[HoJ.OP_EQUAL] = "==",
	[HoJ.OP_GREATER] = ">",
	[HoJ.OP_GREATER_EQUAL] = ">=",
	[HoJ.OP_NOT_EQUAL] = "!=",
	[HoJ.OP_AND] = "&&",
	[HoJ.OP_OR] = "||",
	[HoJ.OP_IN_RANGE_INCLUSIVE] = "[]",
	[HoJ.OP_IN_RANGE_MIN_EX_MAX_IN] = "[]",
	[HoJ.OP_IN_RANGE_MIN_IN_MAX_EX] = "()",
	[HoJ.OP_IN_RANGE_MIN_EX_MAX_EX] = "()",
	[HoJ.OP_TABLE_CONTAINS] = " contains ",
}

function valueToString( value )
	local typeName = type( value )
	if ( typeName == "number" ) then
		return format( "%.2f", value )
	elseif ( typeName == "string" ) then
		return value
	else
		return tostring( value )
	end
end

function valueFn( name, value, constant, results )
	return valueToString( value )
end

function valueScaledFn( name, value, constant, results )
	return "{ " .. valueToString( value ) .. " * " .. format( "%.2f }", constant )
end

function scaledRatioFn( name, value, constant, results )
	return format( "( ( %s * %.2f ) / ? ) { ", name, constant ) .. valueToString( value ) .. " }" 
end

function arrayValueFn( name, value, constant, results )
	return format( "%s[%i]{ ", name, constant ) .. valueToString( value ) .. " }"
end

function resultValueFn( name, value, constant, results )
	return format( "result[%i]:{ ", constant ) .. valueToString( value ) .. " }"
end

function funcValueFn( name, value, constant, results )
	return name .. "( " .. tostring( constant ) .. " ){ " .. valueToString( value ) .. " }"
end

function funcSpellInfoFn( name, value, constant, results )
	return name .. "( " .. constant.name .. " ){ " .. valueToString( value ) .. " }"
end

function spellMemberFn( name, value, constant, results )
	return constant.name .. "." .. name .. "{ " .. value .. " }"
end

function tableValueFn( name, value, constant, results )
	local tableStr = "{ "
	for i, v in pairs( value ) do
		tableStr = tableStr .. tostring( v ) .. ", "
	end
	tableStr = tableStr .. " }"

	return name .. tableStr
end

local _varNameTable = {
	[HoJ.TYPE_CONSTANT] = {
		_name = "const",
		_fn = valueFn
	},
	[HoJ.TYPE_RESULT] = {
		_name = "result",
		_fn = resultValueFn
	},
	[HoJ.TYPE_POWER] = {
		_name = "_power",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_HEALTH_RATIO] = {
		_name = "_playerHealthRatio",
		_fn = valueFn
	},
	[HoJ.TYPE_TARGET_HEALTH_RATIO] = {
		_name = "_targetHealthRatio",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_BUFF_ACTIVE] = {
		_name = "IsBuffActive(",
		_fn = funcValueFn
	},
	[HoJ.TYPE_TARGET_RANGE] = {
		_name = "InRange",
		_fn = arrayValueFn
	},
	[HoJ.TYPE_TALENT_SELECTED] = {
		_name = "IsTalentSelected",
		_fn = funcSpellInfoFn
	},
	[HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING] = {
		_name = "GetUnitBuffTimeRemaining",
		_fn = funcValueFn
	},
	[HoJ.TYPE_PLAYER_BUFF_REMAINING] = {
		_name = "GetBuffTimeRemaining",
		_fn = valueFn
	},
	[HoJ.TYPE_SKILL_CHARGES] = {
		_name = "GetSpellCharges",
		_fn = funcValueFn
	},
	[HoJ.TYPE_NUM_MELEE_ATTACKERS] = {
		_name = "_numAttackers",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_BUFF_STACK_COUNT] = {
		_name = "GetBuffStackCount",
		_fn = funcValueFn
	},
	[HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO] = {
		_name = "_minorHealHealthRatio",
		_fn = valueFn
	},
	[HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO] = {
		_name = "_majorHealHealthRatio",
		_fn = valueFn
	},
	[HoJ.TYPE_TARGET_IS_BOSS] = {
		_name = "_targetIsBoss",
		_fn = valueFn
	},
	[HoJ.TYPE_SKILL_OFF_COOLDOWN] = {
		_name = "IsSkillOffCooldown",
		_fn = funcValueFn
	},
	[HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED] = {
		_name = "_targetCanBeInterrupted",
		_fn = valueFn
	},
	[HoJ.TYPE_COMBO_POINTS] = {
		_name = "_comboPoints",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_IS_STEALTHED] = {
		_name = "_isStealthed",
		_fn = valueFn
	},
	[HoJ.TYPE_INCOMING_DPS_RATIO] = {
		_name = "_incomingDpsRatio",
		_fn = valueScaledFn
	},
	[HoJ.TYPE_SECONDS_SINCE_DAMAGED_BY_SPELL] = {
		_name = "SecondsSinceDamageBySpell",
		_fn = funcValueFn
	},
	[HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE] = {
		_name = "_targetTimeToLive",
		_fn = valueScaledFn
	},
	[HoJ.TYPE_SPELL_DURATION] = {
		_name = "_duration",
		_fn = spellMemberFn
	},
	[HoJ.TYPE_TARGET_IS_TOUGH] = {
		_name = "_targetIsTough",
		_fn = valueFn
	},
	[HoJ.TYPE_COMBAT_ATTACKS_HIT] = {
		_name = "_attacksHit",
		_fn = valueFn
	},
	[HoJ.TYPE_SKILL_POWER_COST] = {
		_name = "_powerGen",
		_fn = spellMemberFn
	},
	[HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF] = {
		_name = "GetUnitBuffRemaining",
		_fn = funcValueFn
	},
	[HoJ.TYPE_SKILL_COOLDOWN_REMAINING] = {
		_name = "GetCDRemainingAfterGCD",
		_fn = funcSpellInfoFn
	},
	[HoJ.TYPE_UNIT_IS_SLOWED] = {
		_name = "UnitIsSlowed",
		_fn = funcValueFn
	},
	[HoJ.TYPE_NET_INCOMING_DPS] = {
		_name = "_netIncomingDPS",
		_fn = scaledRatioFn
	},
	[HoJ.TYPE_NET_INCOMING_DPS_RATIO] = {
		_name = "_netIncomingDPSRatio",
		_fn = scaledRatioFn
	},
	[HoJ.TYPE_NET_INCOMING_DPS_PCT] = {
		_name = "_netIncomingDPSPct",
		_fn = scaledRatioFn
	},
	[HoJ.TYPE_PLAYER_TIME_TO_LIVE] = {
		_name = "_playerTimeToLive",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_TIME_TO_NEAR_DEATH] = {
		_name = "_playerTimeToNearDeath",
		_fn = valueFn
	},
	[HoJ.TYPE_NET_INCOMING_DPS_CHR] = {
		_name = "_netIncomingDpsChr",
		_fn = scaledRatioFn
	},
	[HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL] = {
		_name = "_incomingMajorDamageSpell",
		_fn = valueFn
	},
	[HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS] = {
		_name = "_durationForComboPoints",
		_fn = arrayValueFn
	},
	[HoJ.TYPE_POWER_AFTER_SKILL_USED] = {
		_name = "PowerAfterSkillUsed",
		_fn = funcValueFn
	},
	[HoJ.TYPE_POWER_AFTER_CD_MINUS_SKILL_POWER] = {
		_name = "PowerAfterCdMinusSkillPower",
		_fn = funcValueFn
	},
	[HoJ.TYPE_SKILL_POWER_GEN] = {
		_name = "_powerGen",
		_fn = spellMemberFn
	},
	[HoJ.TYPE_SPELL_CREATURE_TYPES] = {
		_name = "spellInfo._creatureTypes",
		_fn = tableValueFn
	},
	[HoJ.TYPE_TARGET_CREATURE_TYPE] = {
		_name = "_targetCreatureType",
		_fn = valueFn
	},
	[HoJ.TYPE_IS_SHIELD_EQUIPPED] = {
		_name = "_isShieldEquipped",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_LEVEL] = {
		_name = "_level",
		_fn = valueFn
	},
	[HoJ.TYPE_SPELL_MIN_LEVEL] = {
		_name = "_spellMinLevel",
		_fn = valueFn
	},
	[HoJ.TYPE_UNIT_IS_DEAD] = {
		_name = "_unitIsDead",
		_fn = valueFn
	},
	[HoJ.TYPE_CAN_USE_EXECUTE] = {
		_name = "_canUseExecute",
		_fn = valueFn
	},
	[HoJ.TYPE_TARGET_IS_FRIEND] = {
		_name = "_targetIsFriend",
		_fn = valueFn
	},
	[HoJ.TYPE_TARGET_IS_PLAYER] = {
		_name = "_targetIsPlayer",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_IN_COMBAT] = {
		_name = "_playerInCombat",
		_fn = valueFn
	},
	[HoJ.TYPE_IS_TANKING_TARGET] = {
		_name = "_isTankingTarget",
		_fn = valueFn
	},
	[HoJ.TYPE_HAS_AGGRO] = {
		_name = "_hasAggro",
		_fn = valueFn
	},
	[HoJ.TYPE_SPEC_CHOSEN] = {
		_name = "_specChosen",
		_fn = valueFn
	},
	[HoJ.TYPE_POWER_MAX_ADJUSTED] = {
		_name = "_powerMaxAdjusted",
		_fn = valueFn
	},
	[HoJ.TYPE_TIME_SINCE_LAST_CAST] = {
		_name = "_timeSinceLastCast",
		_fn = valueFn
	},
	[HoJ.TYPE_COMBO_POINTS_DEFICIT] = {
		_name = "_comboPointsDeficit",
		_fn = valueFn
	},
	[HoJ.TYPE_GCD_REMAINING] = {
		_name = "_gcdRemaining",
		_fn = valueFn
	},
	[HoJ.TYPE_GCD_DURATION] = {
		_name = "_gcdDuration",
		_fn = valueFn
	},
}

-- -----------------------------------------------------------------------------
-- HoJ:BooleanOp
-- compares two values using the specified operation constant
-- returns true or false
-- -----------------------------------------------------------------------------
function HoJ:BooleanOp( op, a, b )
	local func = _opTable[op]
	if ( func ~= nil ) then
		return func( a, b )
	end
	self:Error( "Unknown op: " .. op )
	return false
end

-- -----------------------------------------------------------------------------
-- HoJ:ConstructRuleState
-- -----------------------------------------------------------------------------
function HoJ:ConstructRuleState( ruleSetName, ruleIndex, testIndex, test )
	return { 
		_ruleSetName = ruleSetName or "",
		_ruleIndex = ruleIndex or -1,
		_testIndex = testIndex or -1,
		_test = test or nil
	}
end

-- -----------------------------------------------------------------------------
-- HoJ:GetValue
-- returns a value based on the type and/or constant
-- -----------------------------------------------------------------------------
function HoJ:GetValue( rotationState, ruleState, parmName, valueType, constant, constant2, results )
	if ( valueType == HoJ.TYPE_CONSTANT ) then
		return constant
	else
		local func = _varTable[valueType]
		if ( func ~= nil ) then
			return func( rotationState, constant, constant2, results )
		end
	end

	self:DevError( format( "Rule set '%s', rule %i, test %i, parm %s has unknown value type: %s",
			ruleState._ruleSetName,
			ruleState._ruleIndex,
			ruleState._testIndex,
			parmName,
			tostring( valueType ) ) )
	return 0
end

-- -----------------------------------------------------------------------------
-- HoJ:Compare
-- compares two values specified by a type constant and or a value using the
-- specified operation and returns a bool
-- -----------------------------------------------------------------------------
function HoJ:Compare( rotationState, 
		ruleState,
		results )
	local test = ruleState._test
	local a = self:GetValue( rotationState, ruleState, "a", test.aType, test.aConst, test.aConst2, results )
	local b = self:GetValue( rotationState, ruleState, "b", test.bType, test.bConst, test.bConst2, results )
	if ( a == nil ) then
		HoJ:Error( "Rule set " .. tostring( ruleState._ruleSetName ) .. ", rule " .. ruleState._ruleIndex .. ", test " .. ruleState._testIndex .. " value A is nil!" )
		-- HoJ:DebugBreak()
	end
	if ( b == nil ) then
		HoJ:Error( "Rule set " .. tostring( ruleState._ruleSetName ) .. ", rule " .. ruleState._ruleIndex .. " test " .. ruleState._testIndex .. " value B is nil!" )
		-- HoJ:DebugBreak()
	end
	return self:BooleanOp( test.op, a, b ), a, b
end

-- -----------------------------------------------------------------------------
-- HoJ:GetRuleDebugName
function HoJ:GetRuleDebugName( rule )
	if ( rule._debugName ~= nil ) then
		return rule._debugName
	else
		return HoJ:GetSpellDebugName( rule.spellInfo )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:DoRotSkillTest
function HoJ:DoRotSkillTest( rotationState, ruleState, results, debug )
	local test = ruleState._test
	local ruleSetName = ruleState._ruleStateName
	local ruleIndex = ruleState._ruleIndex
	local testIndex = ruleState._testIndex

	if ( test == nil ) then
		return true 	-- no test, so always pass
    end
    if ( test.op == nil ) then
        HoJ:DebugMsg( "test has nil operator!" )
        return
	end
	if ( test.aType == nil ) then
		HoJ:DebugMsg( "test.aType = nil in " .. HoJ:GetRuleDebugName( test ) )
		HoJ:DebugMsg( "test.bType = " .. tostring( test.bType ) .. " in " .. HoJ:GetRuleDebugName( test ) )
	end
	if ( test.bType == nil ) then
		HoJ:DebugMsg( "test.aType = " .. tostring( test.aType ) .. " in " .. HoJ:GetRuleDebugName( test ) )
		HoJ:DebugMsg( "test.bType = nil in " .. HoJ:GetRuleDebugName( test ) )
	end
	
	local result, valueA, valueB = self:Compare( rotationState, ruleState, results )

	if ( debug ) then
		local varATbl = _varNameTable[test.aType]
		local varBTbl = _varNameTable[test.bType]
		if ( varATbl._fn == nil ) then
			HoJ:Error( format( "varATbl._fn for rule set '%s', rule index #%i, test #%i is nil!", ruleSetName, ruleIndex, testIndex ) )
		elseif ( varBTbl._fn == nil ) then
			HoJ:Error( format( "varBTbl._fn for rule set '%s', rule index #%i, test #%i is nil!", ruleSetName, ruleIndex, testIndex ) )
		else
			HoJ:DebugMsg( varATbl._fn( varATbl._name, valueA, test.aConst, results ) .. 
				" " .. _opNameTable[test.op] .. " " ..
				varBTbl._fn( varBTbl._name, valueB, test.bConst, results ) ..
				" = " .. tostring( result ) )
		end
	end

	return result
end

-- -----------------------------------------------------------------------------
-- HoJ:ExecuteRuleTests
-- runs the encoded tests and returns true or false
-- -----------------------------------------------------------------------------
function HoJ:ExecuteRuleTests( rotationState, ruleSetName, ruleIndex, rule )
	HoJ:LogD( rule._debugSuccess, "ExecuteRuleTests" )
    -- if disabled, never return true
	if ( C_SpellBook.IsSpellDisabled( rule.spellInfo.id ) ) then
		return false
	end

	-- if there are no explicit tests, then return true by default
	if ( rule.tests == nil or next( rule.tests ) == nil ) then
		HoJ:LogD( rule._debugSuccess, format( "rule %s succeeded because tests == nil",
			tostring( HoJ:GetRuleDebugName( rule ) ) ) )
		return true
	end

	HoJ:LogD( rule._debugSuccess, "ExecuteRuleTests2" )

	local results = { }
	local numRules = 0
	for i, test in ipairs( rule.tests ) do
		if ( test.op == nil ) then
			HoJ:DebugMsg( "Test " .. i .. " for rule " .. HoJ:GetRuleDebugName( rule ) .. 
					" in ruleset '" .. tostring( ruleSetName ) .. "' has a nil operator!" )
			results[i] = false
		else
			local ruleState = self:ConstructRuleState( ruleSetName, ruleIndex, i, test )
			local result = self:DoRotSkillTest( rotationState, ruleState, results, false )
			if ( result == nil ) then
				HoJ:Error( "Rule " .. i .. " in " .. ruleSetName .. " returned nil!")
			end
            results[i] = result
		end
		numRules = numRules + 1
	end

	HoJ:LogD( rule._debugSuccess, "ExecuteRuleTests3" )

	-- return the last result
	local lastResult = results[#results]
	if ( lastResult == nil ) then
		self:Error( "lastResult = nil in rule set " .. ruleSetName .. "! numRules = " .. numRules )
	elseif ( ( rule._debugSuccess and lastResult == true ) or ( rule._debugFail and lastResult == false ) ) then
----[[		
		local ruleName = HoJ:GetRuleDebugName( rule, ruleIndex )
		HoJ:Log( "Skill '" .. ruleName .. "' (minLevel " .. rule.spellInfo.minLevel .. ") results:" )

		-- execute tests again with debug info
		local retestResults = { }
		for i, test in ipairs( rule.tests ) do
			if ( test.op == nil ) then
				HoJ:Log( "Test " .. i .. " for " .. ruleName .. " has a nil operator!" )
				retestResults[i] = false
			else
				local ruleState = self:ConstructRuleState( ruleSetName, ruleIndex, i, test )
				local result = self:DoRotSkillTest( rotationState, ruleState, retestResults, true )
				retestResults[i] = result
			end
		end

		for i, r in ipairs(retestResults) do
			HoJ:Log( format("  [%i] = %s", i, tostring( r ) ) )
		end
--]]
	end

	HoJ:LogD( rule._debugSuccess, "ExecuteRuleTests4" )
	
	if ( rule._debugSuccessFn and lastResult == true ) then
		rule._debugSuccessFn( self, rule )
	end

	return lastResult
end

-- -----------------------------------------------------------------------------
-- HoJ:DumpRotationRules
-- -----------------------------------------------------------------------------
function HoJ:DumpRotationRules( rulesName, rules )
    if ( rules == nil ) then
        HoJ:DebugMsg( "Rules " .. rulesName .. " is nil!" )
        return
    end

    for i, v in ipairs( rules ) do
        if ( v.spellInfo == nil ) then
            HoJ:DebugMsg( "Rule " .. i .. " spellInfo is nil!" )
        else
            HoJ:DebugMsg( "Rule " .. i .. " spellInfo = " .. v.spellInfo.name )
        end
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:CalcFuturePower
-- -----------------------------------------------------------------------------
function HoJ:CalcFuturePower( curTime, duration, nextPowerGainTime, powerGainInterval, curPower, powerStep )

	if ( powerStep <= 0 ) then
		return curPower
	end

	local timeUntilGain = nextPowerGainTime - curTime
	if ( timeUntilGain > duration ) then
		return curPower
	end

	local numIntervals = math.floor( ( duration - timeUntilGain ) / powerGainInterval ) + 1
	local predictedPower = curPower + ( powerStep * numIntervals )
	return predictedPower

end

-- -----------------------------------------------------------------------------
-- HoJ:IsRuleSkillOffCooldown
function HoJ:IsRuleSkillOffCooldown( rotationState, rule )
	if ( HoJ:IsOffCooldown( rotationState, rule.spellInfo, rule._debugSuccess == true or rule._debugFail == true ) ) then
		if ( rule._debugSuccess ~= nil and rule._debugSuccess == true ) then
			HoJ:Log( "Rule " .. HoJ:GetRuleDebugName( rule ) .. " spell is OFF cooldown." )
		end
		return true
	end

	if ( rule._debugFail ~= nil and rule._debugFail == true ) then
		HoJ:Log( "Rule " .. HoJ:GetRuleDebugName( rule ) .. " spell is ON cooldown." )
	end
	return false
end

-- -----------------------------------------------------------------------------
-- HoJ:SelectSkill
-- returns first-best and second-best choices
-- -----------------------------------------------------------------------------
function HoJ:SelectSkill( rotationState, rulesList )
	--HoJ:DebugMsg( "SelectSkill" )
	if ( rulesList == nil ) then
		-- HoJ:Error( "rulesList is nil!" )
		return
	end

	local isTalentSelectedFn = self.IsTalentSelected
	local isOffCooldownFn = self.IsRuleSkillOffCooldown
	local executeRuleTestsFn = self.ExecuteRuleTests
	local checkPowerFn = self._curSpecInfo._checkPowerFn
	local checkComboPointsFn = self._curSpecInfo._checkComboPointsFn
	local playerLevel = self._level

	local emptyFn = function( rule, msg )
	end

	local printFn = function( rule, msg )
		HoJ:DebugMsg( HoJ:GetRuleDebugName( rule ) .. ": " .. msg )
	end

	local SPELL_TALENT = HoJ.SPELL_TALENT -- local to avoid hash lookups in loop
	local SPELL_TALENT_BUFF = HoJ.SPELL_TALENT_BUFF -- local to avoid hash lookups in loop

	local firstRule = nil
	local secondRule = nil
	local power = self:GetPowerAccountingForPooling( rotationState )
	local curComboPoints = HoJ._comboPoints
	local debugLevel = HoJ._debugRulesLevel

	for i, rule in ipairs( rulesList ) do
		local check = true
		if ( firstRule ~= nil ) then
			if ( firstRule.spellInfo == spellInfo ) then
				check = false
			end
			if ( check ) then
				check = ( rule._skipNextSkillCheck == nil or rule._skipNextSkillCheck == false )
				-- if ( not check ) then
				-- 	HoJ:Log( format( "Skipping %s for second skill check!", rule._debugName ) )
				-- end
			end
		end
		if ( check ) then
			if ( rule.replacedByTalent == nil or not isTalentSelectedFn( self, rule.replacedByTalent ) ) then
				if ( rule.spellInfo == nil ) then
					HoJ:DevError( "Rule " .. i .. ":" .. HoJ:GetRuleDebugName(rule) .. " in " .. rulesList._name .. " has nil spellInfo!" )
				elseif ( rule.spellInfo.showOnBar and playerLevel >= rule.spellInfo.minLevel ) then
					if ( ( rule.spellInfo.spellType ~= SPELL_TALENT and rule.spellInfo.spellType ~= SPELL_TALENT_BUFF )
						or isTalentSelectedFn( self, rule.spellInfo ) ) then
						if ( rule._rangeCheck == nil or rule._rangeCheck == false or HoJ._targetInRange[rule.spellInfo.range] == true ) then
							--[[
							if ( rule._debugSuccess or rule._debugFail ) then
								if ( rule._rangeCheck == nil or rule._rangeCheck == false ) then
									HoJ:Log( "_rangeCheck = " .. tostring( rule._rangeCheck ) )
								else
									HoJ:Log( "_targetInRange[" .. rule.spellInfo.range .. "] = " .. tostring( HoJ._targetInRange[rule.spellInfo.range] ) )
								end
							end
							--]]
							if ( rule._powerCheck == nil or rule._powerCheck == false 
									or checkPowerFn( self, rotationState, rule, power ) == true ) then
								if ( ( rule._comboPtCheck ~= nil and rule._comboPtCheck == false ) or
										checkComboPointsFn == nil or checkComboPointsFn( self, rotationState, rule, curComboPoints ) ) then
									if ( isOffCooldownFn( self, rotationState, rule ) or rule._skipCooldownCheck ) then -- do the _skipCooldownCheck last because it's almost never used
										if ( executeRuleTestsFn( self, rotationState, rulesList._name, i, rule ) ) then
											if ( firstRule == nil ) then
												firstRule = rule
											else 
												secondRule = rule
												break
											end
										else
											if ( rule._debugFail ) then
												HoJ:Log( format( "Rule %s failed custom tests.", HoJ:GetRuleDebugName( rule ) ) )	
											end
										end
									elseif ( rule._debugFail ) then
										HoJ:Log( format( "Rule %s failed cooldown test.", HoJ:GetRuleDebugName( rule ) ) )	
									end
								elseif ( rule._debugFail ) then
									HoJ:Log( format( "Rule %s failed min combo points check, curComboPoints = %i", HoJ:GetRuleDebugName( rule ), curComboPoints ) )	
								end
							elseif ( rule._debugFail ) then
								HoJ:Log( format( "Rule %s failed power check", HoJ:GetRuleDebugName( rule ) ) )	
							end
						elseif ( rule._debugFail ) then
							HoJ:Log( format( "Rule %s failed range check", HoJ:GetRuleDebugName( rule ) ) )	
						end
					elseif ( rule._debugFail ) then
						HoJ:Log( format( "Rule %s failed talent selected test.", HoJ:GetRuleDebugName( rule ) ) )	
					end
				elseif ( rule._debugFail ) then
					HoJ:Log( format( "Rule %s failed minLevel or showOnBar tests.", HoJ:GetRuleDebugName( rule ) ) )	
				end
			end
			if ( rule._debugFailFn and firstRule ~= rule and secondRule ~= rule ) then
				rule._debugFailFn( self, rule )
			end
		end
	end

	 local firstSpell = nil
	 if ( firstRule ~= nil ) then
		firstSpell = firstRule.spellInfo
	 end
	 local secondSpell = nil
	 if ( secondRule ~= nil ) then
		secondSpell = secondRule.spellInfo
	 end
	return firstSpell, secondSpell
end

function HoJ:FindSkillsOffCooldownAtTime( rotationState, rotationSelectionList )

	local nextSkill = nil
	local nextSkillCdRemaining = -1.0
	local nextNextSkill = nil
	
	local isOffCooldownFn = self.IsOffCooldown
	local checkPowerFn = self.CheckPower
	local executeRuleTestsFn = self.ExecuteRuleTests

	-- HoJ:Log( format( "FindSkillsOffCooldownAtTime: rs._time = %.2f", rotationState._time ) )

	for i, rule in ipairs( rotationSelectionList ) do
		if ( rule._skipNextSkillCheck == nil or rule._skipNextSkillCheck == false ) then
			if ( rule._rangeCheck == nil or rule._rangeCheck == false or HoJ._targetInRange[rule.spellInfo.range] == true ) then
				-- execute a power check whether or not the rule says to. We want to know definitively 
				-- the player will have enough power to use this skill.
				if ( rule._powerCheck == nil or rule._powerCheck == false or checkPowerFn( self, rotationState, rule ) == true ) then
					-- HoJ:LogD( rule.spellInfo.id == 85256, "TV power check: rule._powerCheck = " .. tostring( rule._powerCheck ) )
					-- if ( rule.spellInfo.id == 35395 ) then -- CS
					-- 	debugCooldowns = true
					-- else 
					-- 	debugCooldowns = false
					-- end
					local offCooldown, cdRemaining = isOffCooldownFn( self, rotationState, rule.spellInfo, false )
					if ( offCooldown ) then
						-- HoJ:LogD( debugCooldowns, "Executing rule " .. HoJ:GetRuleDebugName( rule ) )
						if ( executeRuleTestsFn( self, rotationState, rotationSelectionList._name, i, rule ) ) then
							if ( nextSkill == nil ) then
								nextSkill = rule.spellInfo
								nextSkillCdRemaining = cdRemaining
								-- HoJ:Log( "Found skill " .. rule.spellInfo.name )
							elseif ( nextNextSkill == nil ) then
								-- if the next skills CD is actually shorter, swap it
								if ( cdRemaining < nextSkillCdRemaining ) then
									nextNextSkill = nextSkill
									nextSkill = rule.spellInfo
								else
									nextNextSkill = rule.spellInfo
								end
								break
							end
						end
					end
				end
			end
		end
	end

	-- debugCooldowns = false

	return nextSkill, nextNextSkill
end

function HoJ:NextSkillOffCooldown( rotationState, rotationSelectionList, resetBaseTime )
	-- HoJ:Log( "NextSkillOffCooldown: base time = " .. rotationState._time )
	local nextSkill = nil
	local nextNextSkill = nil

	local maxChecks = 10
	local timeBetweenChecks = 1.0
	local totalTime = maxChecks * timeBetweenChecks
	
	-- once we start checking for a next skill to use, we lock the base time so 
	-- that our cooldown checks happen at stable offsets. If we do not do this,
	-- then every time this function executes the time offsets "slide forward"
	-- by the update interval. This can make a skill that wasn't usable until 
	-- a later time offset suddenly become usable one time offset sooner. That
	-- can make prediction rotate between multiple skills as they each become
	-- usable in a new offset.
	if ( resetBaseTime or self._nextSkillBaseTime < 0.0 ) then
		self._nextSkillBaseTime = rotationState._time
	end
	
	if ( ( rotationState._time - self._nextSkillBaseTime ) > totalTime ) then
		self._nextSkillBaseTime = rotationState._time
		-- HoJ:Error( "_nextSkillBaseTime never reset! Resetting!" )
	end

	local baseTime = self._nextSkillBaseTime
		
	local findSkillsOffCooldownAtTimeFn = self.FindSkillsOffCooldownAtTime
	local constructFn = self.ConstructPredictedRotationState

	for i = 1, maxChecks do
		local timeOffset = i * timeBetweenChecks
		local rs = constructFn( self, baseTime, timeOffset, nil )
		-- HoJ:Log( format( "  timeOffset = %.2f, t = %.2f, power = %i", timeOffset, rs._time, rs._power ) )
		nextSkill, nextNextSkill = findSkillsOffCooldownAtTimeFn( self, rs, rotationSelectionList )
		if ( nextSkill ~= nil ) then
			-- HoJ:Log( format( "Found next skill %s at time %.2f", nextSkill.name, timeOffset ) )
			return nextSkill, nextNextSkill
		end
	end

	-- if nothing was found, reset the base time so it doesn't get infinitely further away
	self._nextSkillBaseTime = rotationState._time

	return nil, nil
end

-- -----------------------------------------------------------------------------
-- HoJ:IsBossUnit
-- returns:
-- - (bool) true if enemy boss
-- - (bool) true if friendly boss
-- -----------------------------------------------------------------------------
function HoJ:IsBossUnit( unit, targetClassification )
	if ( targetClassification == "worldboss" ) then
		return true, false
	end

	local isFriendlyBoss = UnitIsFriend( "player", unit ) and UnitLevel( unit ) == -1
	local isInstanceBoss = false

	for i = 1, 5 do
	    local bossGUID = UnitGUID( "boss" .. i )

	    if UnitGUID(unit) == bossGUID then
	        isInstanceBoss = true
	        break
	    end
	end
	return isInstanceBoss, isFriendlyBoss
end

local INTERACT_INSPECT = 1 -- 28 yards -- this appears to == 30 yards
local INTERACT_TRADE = 2 -- 11.11 yards according to WoWWiki, but exactly equals duel distance now, at ~9.9 yards
local INTERACT_DUEL = 3 -- 9.9 yards according to WoWWiki
local INTERACT_FOLLOW = 4 -- 28 yards -- this is < 30 yards, ~28 yards

local rangeToInteract = {
	[HoJ.RANGE_NONE] = {
		_type = 0,
		_name = "NONE"
	},
	[HoJ.RANGE_SELF] = {
		_type = 0,
		_name = "SELF"
	},
	[HoJ.RANGE_MELEE] = {
		_type = INTERACT_DUEL,
		_name = "MELEE"
	},
	[HoJ.RANGE_8_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "8 yards"
	},
	[HoJ.RANGE_10_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "10 yards"
	},
	[HoJ.RANGE_12_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "12 yards"
	},
	[HoJ.RANGE_15_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "15 yards"
	},
	[HoJ.RANGE_20_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "20 yards"
	},
	[HoJ.RANGE_25_YARDS] = {
		_type = INTERACT_FOLLOW,
		_name = "25 yards"
	},
	[HoJ.RANGE_30_YARDS] = {
		_type = INTERACT_INSPECT,
		_name = "30 yards"
	},
	[HoJ.RANGE_40_YARDS] = {
		_type = INTERACT_INSPECT,
		_name = "40 yards"
	},
	[HoJ.RANGE_100_YARDS] = {
		_type = INTERACT_INSPECT,
		_name = "100 yards"
	},
}

-- -----------------------------------------------------------------------------
-- HoJ:InRangeOfSpells
-- -----------------------------------------------------------------------------
function HoJ:InRangeOfSpells( rangeEnum, t, target, debugName, debug )
	-- spell info may be a table array if multiple spells can be used
	local inRange = false
	local subTable = t[rangeEnum]
	if ( subTable == nil ) then
		HoJ:LogD( debug, debugName .. ": no table for " .. rangeEnum )
		local canInteract = false
		local tbl = rangeToInteract[rangeEnum]
		if ( tbl._type < 1 ) then
			inRange = true
		else
			inRange = CheckInteractDistance( target, tbl._type )
		end
		HoJ:LogD( debug, "CheckInteractDistance(" .. tbl._type .. ") = " .. tostring( inRange ) .. " for " .. tbl._name )
	else 
		local spellInfos = subTable._spellInfos
		if ( spellInfos ~= nil ) then
			for i, spell in ipairs( spellInfos ) do
				HoJ:LogD( debug, debugName .. " checking spellInfo " .. spell.name )
				if ( HoJ:InRange( spell, target, debugName, debug ) ) then
					inRange = true
					break
				end
			end
		else
			HoJ:LogD( debug, "No spellInfos for " .. debugName )
			if ( subTable._spellInfo ~= nil ) then
				inRange = HoJ:InRange( subTable._spellInfo, target, debugName, debug )
			else 
				HoJ:Error( "HoJ:InRangeOfSpells - _spellInfo for " .. debugName .. " is nil!" )
				inRange = true
			end
		end
	end
	HoJ:LogD( debug, "InRangeOfSpells(" .. debugName .. ") = " .. tostring( inRange ) )
	return inRange
end

function HoJ:ConstructTargetRanges()
	-- HoJ:Log( "ConstructTargetRanges" )

	local ranges = { }

	ranges[HoJ.RANGE_NONE] = true;
	ranges[HoJ.RANGE_SELF] = true;
    ranges[HoJ.RANGE_MELEE] = false;
    ranges[HoJ.RANGE_8_YARDS] = false;
	ranges[HoJ.RANGE_10_YARDS] = false;
	ranges[HoJ.RANGE_12_YARDS] = false;
	ranges[HoJ.RANGE_15_YARDS] = false;
    ranges[HoJ.RANGE_20_YARDS] = false;
	ranges[HoJ.RANGE_25_YARDS] = false;
	ranges[HoJ.RANGE_30_YARDS] = false;
	ranges[HoJ.RANGE_40_YARDS] = false;
	ranges[HoJ.RANGE_100_YARDS] = false;
	ranges[HoJ.RANGE_100_YARDS] = false;
	return ranges
end

-- -----------------------------------------------------------------------------
-- HoJ:FindRanges
-- -----------------------------------------------------------------------------
function HoJ:FindRanges( target, specInfo )
	-- HoJ:Log( "FindRanges" )
	-- target range finding
	HoJ._targetInRange[HoJ.RANGE_NONE] = true;
	HoJ._targetInRange[HoJ.RANGE_SELF] = true;
    HoJ._targetInRange[HoJ.RANGE_MELEE] = false;
    HoJ._targetInRange[HoJ.RANGE_8_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_10_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_12_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_15_YARDS] = false;
    HoJ._targetInRange[HoJ.RANGE_20_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_25_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_30_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_40_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_100_YARDS] = false;
	
	local debug = HoJ._showRange

	local t = nil
	if ( ( not UnitExists( target ) ) or UnitIsDead( target ) ) then
		return
	elseif ( UnitCanAttack( "player", target ) ) then
		if ( debug ) then
			HoJ:LogD( debug, "Using _enemyRangeSpells" )
		end
		t = specInfo._enemyRangeSpells
	elseif ( UnitCanAssist( "player", target ) ) then
		if ( debug ) then
			HoJ:LogD( debug, "Using _friendlyRangeSpells" )
		end
		t = specInfo._friendlyRangeSpells
	else
		-- HoJ:DebugMsg( "Cannot find range to target!" )
		return
	end

    HoJ._targetInRange[HoJ.RANGE_MELEE] 		= HoJ:InRangeOfSpells( HoJ.RANGE_MELEE,    t, target, "Melee", debug )
    HoJ._targetInRange[HoJ.RANGE_8_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_8_YARDS,  t, target, "8", debug )
	HoJ._targetInRange[HoJ.RANGE_10_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_10_YARDS, t, target, "10", debug )
	HoJ._targetInRange[HoJ.RANGE_12_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_12_YARDS, t, target, "12", debug )
	HoJ._targetInRange[HoJ.RANGE_15_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_15_YARDS, t, target, "15", debug )
    HoJ._targetInRange[HoJ.RANGE_20_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_20_YARDS, t, target, "20", debug )
    HoJ._targetInRange[HoJ.RANGE_25_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_25_YARDS, t, target, "25", debug )
	HoJ._targetInRange[HoJ.RANGE_30_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_30_YARDS, t, target, "30", debug )
	HoJ._targetInRange[HoJ.RANGE_40_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_40_YARDS, t, target, "40", debug )
	HoJ._targetInRange[HoJ.RANGE_100_YARDS] 	= HoJ:InRangeOfSpells( HoJ.RANGE_100_YARDS, t, target, "100", debug )

    if ( debug ) then
        local foundRange = false
        for i, v in ipairs( HoJ._targetInRange ) do
            if ( i >= HoJ.RANGE_MELEE ) then
                if ( v == true ) then                  
                    HoJ:FrameThrottleMsg( 5, "Target range = " .. HoJ._rangeNames[i] )
                    foundRange = true
                    break
                end
            end
        end
        if ( not foundRange ) then
            HoJ:FrameThrottleMsg( 5, "Target range = max" )
        end
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:UnitCreatureType
-- -----------------------------------------------------------------------------
local creatureTypeTable = {
	-- missing values for non-english languages
	["Aberration"] = "Aberration",

	["Beast"] = "Beast",
	["Wildtier"] = "Beast",
	["Bestia"] = "Beast",
	["Bestia"] = "Beast",
	["Bête"] = "Beast",
	["Bestia"] = "Beast",
	["Fera"] = "Beast",
	["Животное"] = "Beast",
	["야수"] = "Beast",
	["野兽"] = "Beast",
	["野獸"] = "Beast",

	["Critter"] = "Critter",
	["Kleintier"] = "Critter",
	["Alma"] = "Critter",
	["Alma"] = "Critter",
	["Bestiole"] = "Critter",
	["Animale"] = "Critter",
	["Bicho"] = "Critter",
	["Существо"] = "Critter",
	["동물"] = "Critter",
	["小动物"] = "Critter",
	["小動物"] = "Critter",

	["Demon"] = "Demon",
	["Dämon"] = "Demon",
	["Demonio"] = "Demon",
	["Demonio"] = "Demon",
	["Démon"] = "Demon",
	["Demone"] = "Demon",
	["Demônio"] = "Demon",
	["Демон"] = "Demon",
	["악마"] = "Demon",
	["恶魔"] = "Demon",
	["惡魔"] = "Demon",

	["Dragonkin"] = "Dragonkin",
	["Drachkin"] = "Dragonkin",
	["Dragon"] = "Dragonkin",
	["Dragón"] = "Dragonkin",
	["Draconien"] = "Dragonkin",
	["Dragoide"] = "Dragonkin",
	["Dracônico"] = "Dragonkin",
	["Дракон"] = "Dragonkin",
	["용족"] = "Dragonkin",
	["龙类"] = "Dragonkin",
	["龍類"] = "Dragonkin",

	["Elemental"] = "Elemental",
	["Elementar"] = "Elemental",
	["Elemental"] = "Elemental",
	["Elemental"] = "Elemental",
	["Élémentaire"] = "Elemental",
	["Elementale"] = "Elemental",
	["Elemental"] = "Elemental",
	["Элементаль"] = "Elemental",
	["정령"] = "Elemental",
	["元素生物"] = "Elemental",
	["元素生物"] = "Elemental",

	["Gas Cloud"] = "Gas Cloud",
	["Gaswolke"] = "Gas Cloud",
	["Nube de Gas"] = "Gas Cloud",
	["Nube de Gas"] = "Gas Cloud",
	["Nuage de gaz"] = "Gas Cloud",
	["Nube di Gas"] = "Gas Cloud",
	["Gasoso"] = "Gas Cloud",
	["Газовое облако"] = "Gas Cloud",
	["가스"] = "Gas Cloud",
	["气体云雾"] = "Gas Cloud",
	["氣體雲"] = "Gas Cloud",

	["Giant"] = "Giant",
	["Riese"] = "Giant",
	["Gigante"] = "Giant",
	["Gigante"] = "Giant",
	["Géant"] = "Giant",
	["Gigante"] = "Giant",
	["Gigante"] = "Giant",
	["Великан"] = "Giant",
	["거인"] = "Giant",
	["巨人"] = "Giant",
	["巨人"] = "Giant",

	["Humanoid"] = "Humanoid",
	["Humanoid"] = "Humanoid",
	["Humanoide"] = "Humanoid",
	["Humanoide"] = "Humanoid",
	["Humanoïde"] = "Humanoid",
	["Umanoide"] = "Humanoid",
	["Humanoide"] = "Humanoid",
	["Гуманоид"] = "Humanoid",
	["인간형"] = "Humanoid",
	["人型生物"] = "Humanoid",
	["人型生物"] = "Humanoid",

	["Mechanical"] = "Mechanical",
	["Mechanisch"] = "Mechanical",
	["Mecánico"] = "Mechanical",
	["Mecánico"] = "Mechanical",
	["Machine"] = "Mechanical",
	["Meccanico"] = "Mechanical",
	["Mecânico"] = "Mechanical",
	["Механизм"] = "Mechanical",
	["기계"] = "Mechanical",
	["机械"] = "Mechanical",
	["機械"] = "Mechanical",

	["Non-combat Pet"] = "Non-combat Pet",
	["Haustier"] = "Non-combat Pet",
	["Mascota no combatiente"] = "Non-combat Pet",
	["Mascota mansa"] = "Non-combat Pet",
	["Familier pacifique"] = "Non-combat Pet",
	["Animale Non combattente"] = "Non-combat Pet",
	["Mascote não-combatente"] = "Non-combat Pet",
	["Спутник"] = "Non-combat Pet",
	["애완동물"] = "Non-combat Pet",
	["非战斗宠物"] = "Non-combat Pet",
	["非戰鬥寵物"] = "Non-combat Pet",

	["Not specified"] = "Not specified",
	["Nicht spezifiziert"] = "Not specified",
	["No especificado"] = "Not specified",
	["Sin especificar"] = "Not specified",
	["Non spécifié"] = "Not specified",
	["Non Specificato"] = "Not specified",
	["Não especificado"] = "Not specified",
	["Не указано"] = "Not specified",
	["기타"] = "Not specified",
	["未指定"] = "Not specified",
	["不明"] = "Not specified",

	["Totem"] = "Totem",
	["Totem"] = "Totem",
	["Tótem"] = "Totem",
	["Totém"] = "Totem",
	["Totem"] = "Totem",
	["Totem"] = "Totem",
	["Totem"] = "Totem",
	["Тотем"] = "Totem",
	["토템"] = "Totem",
	["图腾"] = "Totem",
	["圖騰"] = "Totem",

	["Undead"] = "Undead",
	["Untoter"] = "Undead",
	["No-muerto"] = "Undead",
	["No-muerto"] = "Undead",
	["Mort-vivant"] = "Undead",
	["Non Morto"] = "Undead",
	["Renegado"] = "Undead",
	["Нежить"] = "Undead",
	["언데드"] = "Undead",
	["亡灵"] = "Undead",
	["不死族"] = "Undead",

	-- missing values for some languages
	["Wild Pet"] = "Wild Pet",
	["Ungezähmtes Tier"] = "Wild Pet",
	["Mascotte sauvage"] = "Wild Pet",
}

-- -----------------------------------------------------------------------------
-- HoJ:UnitCreatureType
-- -----------------------------------------------------------------------------
function HoJ:UnitCreatureType( unitId )
	local ct = UnitCreatureType( unitId )
	local tct = creatureTypeTable[ct]
	if ( tct == nil ) then
		tct = "Unknown"
	end
	return tct
end

-- -----------------------------------------------------------------------------
-- HoJ:GetThreat
-- -----------------------------------------------------------------------------
function HoJ:GetThreat( unit, mob )
	local t = UnitThreatSituation( unit, mob )
	if ( t == nil ) then
		return HoJ.THREAT_NONE
	else
		return t
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CinematicInProgress
-- returns true if InCinematic() == true OR is we received a CINEMATIC_START
-- event and haven't received a corresponding CINEMATIC_STOP event yet.
-- -----------------------------------------------------------------------------
function HoJ:CinematicInProgress()
	return HoJ._inCinematic or InCinematic() or IsInCinematicScene()
end