-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Warrior.lua
-- DESC: Warrior-specific functions.

-- BUGS:
-- - another warrior's battleshout buff doesn't make buff bar go away
-- - sometimes Rend shows up as usable (full rage bar on rotation icon) but isn't. 
--   is this due to rage Prediction?
-- TODO:
-- - remove power cost checks from arms to suppor rage pooling
-- X rallying cry as mitigation
-- X dragonRoar as mitigation
-- X victory rush almost never shows right now because health must be below 60%
--   X with impending victory this seemed ok because you can use it more often
--   X maybe need a health ratio on healing spells to make it part of the rules
-- X recklessness option to show in rotation
-- X recklessness (add as "dpsCooldowns", also convert avenging wrath)
-- X battleshout (as a blessing)
--
-- NOTES:
-- * if target is near to dying and the warrior needs to heal, Impending Victory could be 
--   shown as a cooldown as a reminder to use it before the target dies (since it will 
--   reset when the target dies)

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Warrior.lua" )

-- execute 5308
-- sudden death 280721
-- bloodthirst 23881
-- enrage 184361
-- raging blow 85288
-- whirlwind 190411
local warriorSpellInfo = HoJ._warriorSpellInfos

-- -----------------------------------------------------------------------------
-- HoJ:Warrior_SelectSingleOrMultitarget
-- returns
--   rotationRules
--   interruptRules
--   mitigationRules
--   dpsCooldownRules
-- -----------------------------------------------------------------------------
function HoJ:Warrior_SelectSingleOrMultitarget( curSpecInfo, numAttackers )
    if ( self._numMeleeAttackers > curSpecInfo._multiTargetThreshold ) then
        return curSpecInfo._rotationRules._multiTarget, 
                curSpecInfo._interruptRules._multiTarget,
                curSpecInfo._mitigationRules._multiTarget,
                curSpecInfo._dpsCooldownRules._multiTarget
    else
        return curSpecInfo._rotationRules._singleTarget, 
                curSpecInfo._interruptRules._singleTarget,
                curSpecInfo._mitigationRules._singleTarget,
                curSpecInfo._dpsCooldownRules._singleTarget
    end
end


-- -----------------------------------------------------------------------------
-- HoJ:Warrior_CheckBlessings
-- CLASSFIX - paladin specific
-- -----------------------------------------------------------------------------
function HoJ:Warrior_CheckBlessings( activeBlessingsList )
	if ( not HoJ._showBuffsFrame ) then
		return
	end

	-- this has to be called every time we check blessings to update non-party member blessings
	if ( self._updateFrameNum >= self._nextUpdateBlessingFrame ) then
		HoJ:UpdateBlessings( HoJ._curTime, activeBlessingsList )
	end

	--HoJ:DebugMsg( "Warrior_CheckBlessings )

	-- dont show any blessing buffs for a second after entering world
	if ( self._updateFrameNum < self._enteredWorldUpdateFrameNum + HoJ.UPDATES_PER_SEC ) then
		return
	end


	if ( HoJ._level >= HoJ.SpellInfo.battleShout.minLevel ) then
		-- handle party sync disabling this buff
		if ( not C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.battleShout.id ) ) then
			local battleShout = activeBlessingsList[HoJ.SpellInfo.battleShout.id]
			if ( battleShout.start <= 0.0 ) then
				--HoJ:DebugMsg( "Adding Blessing of Kings icon" )
				HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.battleShout )
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Warrior_Update
-- -----------------------------------------------------------------------------
function HoJ:Warrior_Update( time, curSpecInfo, updateCooldowns, updateBuffs, updateRotation )
    -- HoJ:DebugMsg( "Updating spec " .. specInfo._name );

    local rotationRules, interruptRules, mitigationRules, dpsCooldownRules = curSpecInfo._selectSingleOrMultiTargetFn( 
        HoJ, curSpecInfo,
        self._numMeleeAttackers )
        
    local rotationState = HoJ:ConstructCurrentRotationState( HoJ._curTime )

    -- update buffs
    if ( updateBuffs ) then
        local startBuffsMs = HoJ.ProfileBegin2( "Buffs" )
        
        HoJ:ClearIconList( HoJ._buffs );

        HoJ:CheckProcdBuffs( rotationState, curSpecInfo._activeBuffList, curSpecInfo._procdBuffList );
        HoJ:Warrior_CheckBlessings( curSpecInfo._activeBlessings );

        HoJ.ProfileFunction( "UpdateIcons2:Buffs2:", HoJ.UpdateIcons2, self, time, HoJ._buffs, 
                curSpecInfo._buffSortPriorities, HoJ._buffFrames, true )

        HoJ.ProfileEnd( "Buffs", startBuffsMs )
    end

    -- update cooldowns
    if ( updateCooldowns ) then
        local startCooldownsMs = HoJ.ProfileBegin2( "Cooldowns" )

        HoJ.ProfileFunction( "ClearIconList", HoJ.ClearIconList, self, HoJ._cooldowns )

        HoJ.ProfileFunction( "CheckRules(dpsCooldowns)", HoJ.CheckRules, self, rotationState,
                self._cooldowns, self._showCooldownsFrame, dpsCooldownRules, true, true )
        HoJ.ProfileFunction( "CheckRules(interrupts)", HoJ.CheckRules, self, rotationState,
                self._cooldowns, self._showCooldownsFrame, interruptRules, true, true )
        HoJ.ProfileFunction( "CheckRules(mitigations)", HoJ.CheckRules, self, rotationState,
                self._cooldowns, self._showCooldownsFrame, mitigationRules, true, true )

        HoJ.ProfileFunction( "UpdateIcons2:Cooldowns:", HoJ.UpdateIcons2, self, time, self._cooldowns, 
                curSpecInfo._cooldownSortPriorities, self._cooldownFrames, true )

        HoJ.ProfileEnd( "Cooldowns", startCooldownsMs )
    end

    -- update rotation
    if ( updateRotation ) then
        local startRotationMs = HoJ.ProfileBegin2( "Rotation" )

        HoJ:ClearIconList( HoJ._rotationSkills )

        HoJ:CheckDPSRotation( HoJ._rotationSkills, rotationRules )

        HoJ.ProfileFunction( "UpdateIcons2:Rotation:", HoJ.UpdateIcons2, self, time, HoJ._rotationSkills, 
                { } , HoJ._rotationFrames, false );

        HoJ.ProfileEnd( "Rotation", startRotationMs )
    end    
end

local armsMeleeAoESpells = { 

}
local furyMeleeAoESpells = {

}

local protMeleeAoESpells = {

}

local lowHealthMitigationList = { }
local moderateMitigationList = { }
local emergencyMitigationList = { }


--------------------------------------------------------------------------------
-- ARMS
--------------------------------------------------------------------------------

-- https://www.icy-veins.com/wow/arms-warrior-pve-dps-rotation-cooldowns-abilities
local armsRotationRules = {
    _name = "armsRotationRules",
    {
        -- impending victory if below health ratio
        spellInfo = warriorSpellInfo.impendingVictory,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- victory rush if below health ratio
        spellInfo = warriorSpellInfo.victoryRush,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.victorious, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            -- prevent this skill from showing up just as the unit dies
            { aType = HoJ.TYPE_UNIT_IS_DEAD, aConst = "target", bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- rend if target has < 4 seconds on rend debuff, and colossus smash debuff isn't already on the target
        spellInfo = warriorSpellInfo.rend,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.rendDebuff, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF, aConst = warriorSpellInfo.colossusSmashDebuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- skull splitter if < 60 power and we aren't "about to use" bladestorm
        -- about to use bladestorm is hard to quantify, but we shouldn't use it unless we have colossus smash active
        -- so we'll make sure bladestorm is on cooldown OR colossus smash debuff isn't on the target before 
        -- choosing skullsplitter
        spellInfo = warriorSpellInfo.skullSplitter,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = warriorSpellInfo.armsBladestorm, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF, aConst = warriorSpellInfo.colossusSmashDebuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 60, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.avatarArms,
    },
    {
        spellInfo = warriorSpellInfo.ravager,
        _rangeCheck = true,
    },
    {
        spellInfo = warriorSpellInfo.warbreaker,
        _rangeCheck = true,
    },
    {
        -- colossus smash if warbreaker isn't talented
        spellInfo = warriorSpellInfo.colossusSmash,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.warbreaker, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.deadlyCalm,
    },
    {
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        spellInfo = warriorSpellInfo.executeArmsProt,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.armsMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            -- sometimes (varies based on lag) prevent this skill from showing up just as the unit dies
            { aType = HoJ.TYPE_UNIT_IS_DEAD, aConst = "target", bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.armsSuddenDeathBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.overpower,
        _rangeCheck = true,
        _debugName = "overpower wower",
    },
    {
        spellInfo = warriorSpellInfo.mortalStrike,
        _debugName = "mortalStrike_to_refresh_deepWounds",
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.deepWoundsDebuff, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
        }
    },
    {
        -- bladestorm during colossus smash, if ravagaer isn't talented and colossus smash is on the target
        spellInfo = warriorSpellInfo.armsBladestorm,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.ravager, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF, aConst = warriorSpellInfo.colossusSmashDebuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- whirlwind if fervor of battle is talented and we have the power to do it
        spellInfo = warriorSpellInfo.whirlwind,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.fervorOfBattle, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- slam if we have the power for it and the target is in range
        spellInfo = warriorSpellInfo.slam,
        _rangeCheck = true,
        _powerCheck = true,
    },
    -- if we don't have enough power for slam, consider some other options that may be talent-based
    {
        -- storm bolt if it's talented and within range
        spellInfo = warriorSpellInfo.stormBolt,
        _rangeCheck = true,
    },
    { 
        -- charge if the target is out of range, but further than 8 yards 
        spellInfo = warriorSpellInfo.charge,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- heroic leap if the target is out of range, but further than 8 yards (and we couldn't charge in the previous test)
        spellInfo = warriorSpellInfo.heroicLeap,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- heroic throw if within range, but further than 8 yards
        spellInfo = warriorSpellInfo.heroicThrow,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    -- note that we don't check power for the final options because, 
    -- since we don't have any other option right now, we're fine with waiting for the power to pool
    {
        -- rend without power check if target has < 4 seconds on rend debuff, and colossus smash debuff isn't already on the target
        spellInfo = warriorSpellInfo.rend,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.rendDebuff, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF, aConst = warriorSpellInfo.colossusSmashDebuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- mortal strike without power check, if deep wounds is about to expire
        spellInfo = warriorSpellInfo.mortalStrike,
        _debugName = "mortalStrike_to_refresh_deepWounds",
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.deepWoundsDebuff, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
        }
    },
    {
        -- whirlwind if fervor of battle is talented
        spellInfo = warriorSpellInfo.whirlwind,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.fervorOfBattle, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- if fervor of battle wasn't selected, fall back to slam
        spellInfo = warriorSpellInfo.slam,
        _rangeCheck = true,
    },
}

local armsMultiTargetRotationRules = {
    _name = "armsMultiTargetRotationRules",
    {
        -- impending victory if below health ratio
        spellInfo = warriorSpellInfo.impendingVictory,
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.impendingVictory, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.impendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- victory rush if below health ratio
        spellInfo = warriorSpellInfo.victoryRush,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.victorious, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            -- prevent this skill from showing up just as the unit dies
            { aType = HoJ.TYPE_UNIT_IS_DEAD, aConst = "target", bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- cleave if not about to use bladestorm
        spellInfo = warriorSpellInfo.cleave,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = warriorSpellInfo.armsBladestorm, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER },
        }
    },
    {
        -- sweeping strikes if not about to use bladestorm and cleave is not talented
        spellInfo = warriorSpellInfo.sweepingStrikes,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.cleave, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = warriorSpellInfo.armsBladestorm, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- rend, unless we have 4 or more targets
        spellInfo = warriorSpellInfo.rend,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.rendDebuff, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },

        }
    },
    {
        -- skull splitter if < 60 power and we aren't less than 5 seconds from using bladestorm
        spellInfo = warriorSpellInfo.skullSplitter,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 60, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = warriorSpellInfo.armsBladestorm, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.avatarArms,
    },
    {
        spellInfo = warriorSpellInfo.ravager,
        _rangeCheck = true,
    },
    {
        spellInfo = warriorSpellInfo.warbreaker,
        _rangeCheck = true,
    },
    {
        -- colossus smash if warbreaker isn't talented
        spellInfo = warriorSpellInfo.colossusSmash,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.warbreaker, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.deadlyCalm,
    },
    {
        -- cleave to maintain deep wounds debuff
        spellInfo = warriorSpellInfo.cleave,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.deepWoundsDebuff, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS },
        }
    },
    {
        -- execute when < 4 targets
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        spellInfo = warriorSpellInfo.executeArmsProt,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.armsMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.armsSuddenDeathBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        -- execute when > 4 targets and sweeping strikes is active
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        spellInfo = warriorSpellInfo.executeArmsProt,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.armsMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.armsSuddenDeathBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.sweepingStrikes, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- whirlwind if colossus smash debuff is on target and 4 or more attackers
        spellInfo = warriorSpellInfo.whirlwind,
        _rangeCheck = true,
        -- _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF, aConst = warriorSpellInfo.colossusSmashDebuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.overpower,
        _rangeCheck = true,
    },
    {
        -- mortal strike only if < 4 attackers
        spellInfo = warriorSpellInfo.mortalStrike,
        _debugName = "mortalStrike",
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS },
        }
    },
    {
        -- slam if sweeping strikes is active and < 4 attackers, with power check so that if we don't
        -- have enough power we'll check for other options like stormbolt, charge or leap if possible
        spellInfo = warriorSpellInfo.slam,
        _rangeCheck = true,
        _powerCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.sweepingStrikes, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- whirlwind with power check so that if we don't
        -- have enough power we'll check for other options like stormbolt, charge or leap if possible
        spellInfo = warriorSpellInfo.whirlwind,
        _rangeCheck = true,
        _powerCheck = true,
    },
    {
        -- storm bolt if it's talented and within range
        spellInfo = warriorSpellInfo.stormBolt,
        _rangeCheck = true,
    },
    {  
        spellInfo = warriorSpellInfo.charge,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicThrow,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- slam if sweeping strikes is active and < 4 attackers, without power check so that we'll wait on pooling
        spellInfo = warriorSpellInfo.slam,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.sweepingStrikes, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- whirlwind without power check so that we'll wait on pooling
        spellInfo = warriorSpellInfo.whirlwind,
        _rangeCheck = true,
    },
}

local armsInterruptRules = {
    _name = "armsInterruptRules",
    {
        spellInfo = warriorSpellInfo.pummel,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.pummel.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.intimidatingShout.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local armsMitigationRules = {
    _name = "armsMitigationRules",
    -- emergency mitigations
    -- these are selected if:
    --   - the amount of damage predicted over the next 5 seconds exceeds the player's current health
    --   - predicted death time is < 10 seconds
    -- NOTE: non-instant healing is not considered a mitigation
    {
        spellInfo = warriorSpellInfo.dieByTheSword,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.defensiveStance,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.stoneform,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.rallyingCry,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.berserkerRage,
        tests = {
            { aType = HoJ.TYPE_UNIT_IS_SLOWED, aConst = "player", bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.avatar,
        tests = {
            { aType = HoJ.TYPE_UNIT_IS_SLOWED, aConst = "player", bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local armsDpsCooldownRules = {
    {
        spellInfo = warriorSpellInfo.avatarArms,
        tests = {
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 1, bType = HoJ.TYPE_SPELL_DURATION, bConst = warriorSpellInfo.avatar, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_COMBAT_ATTACKS_HIT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_TOUGH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.recklessness, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

local armsCooldownSortPriorities = {
    _name = "armsCooldownSortPriorities",
    [warriorSpellInfo.dieByTheSword.id] = 8,
    [warriorSpellInfo.stoneform.id] = 7,
    [warriorSpellInfo.pummel.id] = 6,
    [warriorSpellInfo.avatarArms.id] = 5,
    [warriorSpellInfo.intimidatingShout.id] = 4,
    [warriorSpellInfo.heroicLeap.id] = 3,
    [warriorSpellInfo.rallyingCry.id] = 2,
    [warriorSpellInfo.berserkerRage.id] = 1,
}

local armsActiveBuffList = { 
    [warriorSpellInfo.overpower.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.armsSuddenDeathBuff.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.victorious.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.bladestorm.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.sweepingStrikes.id] = HoJ:ConstructActiveBuff(),
}

-- this is a list of buffs we want to check for showing on the buffs bar
local armsProcdBuffList = {
    { _spellInfo = warriorSpellInfo.victorious },
}

local armsActiveBlessingList = {
    [warriorSpellInfo.battleShout.id] = {
        spellInfo = warriorSpellInfo.battleShout,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
}

local armsBuffSortPriorities = {
    _name = "armsBuffSortPriorities",
    [warriorSpellInfo.victorious.id] = 6,
    [warriorSpellInfo.armsSuddenDeathBuff.id] = 5,
    [warriorSpellInfo.overpower.id] = 4,
    [warriorSpellInfo.bladestorm.id] = 3,
    [warriorSpellInfo.sweepingStrikes.id] = 2,
    [warriorSpellInfo.battleShout.id] = 1,
}

--------------------------------------------------------------------------------
-- FURY
--------------------------------------------------------------------------------

-- https://www.icy-veins.com/wow/fury-warrior-pve-dps-rotation-cooldowns-abilities
local furyRotationRules = {
    _name = "furyRotationRules",
    {
        spellInfo = warriorSpellInfo.impendingVictory,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.impendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.victoryRush,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.impendingVictory, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.victoryRush.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.shieldSlam,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.shieldSlam.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_IS_SHIELD_EQUIPPED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.onslaught,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.onslaught.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.rampage,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 90, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    --[[
    {
        spellInfo = warriorSpellInfo.recklessness, -- if not on cooldown
        tests = {
            -- TODO: allow this to be removed from the rotation and used only as a cooldown?
            { aType = HoJ.TYPE_CONSTANT, aConst = true, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    --]]
    {
        spellInfo = warriorSpellInfo.siegebreaker,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.siegebreaker.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.executeFury,
        -- _debugSuccess = true,
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.furyMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furySuddenDeathBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.executeFury.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        _debugName = "bloodthirst_no_enrage",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.ragingBlow,
        _debugName = "ragingBlow_2_charges",
        tests = {
            { aType = HoJ.TYPE_SKILL_CHARGES, aConst = warriorSpellInfo.ragingBlow, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ragingBlow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- bloodthirst only if below minor heal health ratio
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        _debugName = "bloodthirst_low_health",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.ragingBlow,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ragingBlow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    -- bloodthirst as a filler if it's available
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.dragonRoar,
        _debugName = "dragonRoar_during_enrage",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.dragonRoar.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.charge,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    -- prefer whirlwind as the first filler, if it is in range
    {
        spellInfo = warriorSpellInfo.whirlwind,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_SKILL_POWER_COST, bConst = warriorSpellInfo.whirlwind, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind2.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_LEVEL, aConst = 0, bType = HoJ.TYPE_SPELL_MIN_LEVEL, bConst = warriorSpellInfo.whirlwind, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_LEVEL, aConst = 0, bType = HoJ.TYPE_SPELL_MIN_LEVEL, bConst = warriorSpellInfo.whirlwind2, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    -- prefer whirlwind as the first filler, if it is in range
    {
        spellInfo = warriorSpellInfo.whirlwind2,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind2.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    -- prefer whirlwind as the first filler, if it is in range
    {
        spellInfo = warriorSpellInfo.whirlwind3,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind3.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- use slam if we can't use raging blow yet (pre-level 12)
        spellInfo = warriorSpellInfo.slam,
        _rangeCheck = true,
        _powerCheck = true,
    },
    -- storm bolt if it's talented and within range
    {
        spellInfo = warriorSpellInfo.stormBolt,
        _rangeCheck = true,
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    -- heroic throw if within range
    {
        spellInfo = warriorSpellInfo.heroicThrow,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    -- whirlwind for the heck of it, maybe another target is in range
    {
        spellInfo = warriorSpellInfo.whirlwind2,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind2.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },    
    {
        spellInfo = warriorSpellInfo.whirlwind3,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind3.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

-- https://www.icy-veins.com/wow/fury-warrior-pve-dps-rotation-cooldowns-abilities
-- note that furious slash is not used in the multi-target rotation based on the icy-veins recommendation
local furyMultiTargetRotationRules = {
    _name = "furMultiTargetRotationRules",
    {
        spellInfo = warriorSpellInfo.impendingVictory,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.impendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },        
    {
        spellInfo = warriorSpellInfo.victoryRush,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.impendingVictory, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.victoryRush.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.onslaught,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.onslaught.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- only use rank 1 whirlwind when we have enough rage
    {
        spellInfo = warriorSpellInfo.whirlwind,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_SKILL_POWER_COST, bConst = warriorSpellInfo.whirlwind, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind2.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_LEVEL, aConst = 1, bType = HoJ.TYPE_SPELL_MIN_LEVEL, bConst = warriorSpellInfo.whirlwind, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_LEVEL, aConst = 1, bType = HoJ.TYPE_SPELL_MIN_LEVEL, bConst = warriorSpellInfo.whirlwind2, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    -- whirlwind as the first filler, if the player doesn't have the final rank of whirlwind and if it is in range
    {
        spellInfo = warriorSpellInfo.whirlwind2,
        tests = {
            { aType = HoJ.TYPE_PLAYER_LEVEL, aConst = 1, bType = HoJ.TYPE_SPELL_MIN_LEVEL, bConst = warriorSpellInfo.whirlwind3, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind2.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- whirlwind if player has rank 3 whirlwind and the buff is expired
    {
        spellInfo = warriorSpellInfo.whirlwind3,
        tests = {
            { aType = HoJ.TYPE_PLAYER_LEVEL, aConst = 1, bType = HoJ.TYPE_SPELL_MIN_LEVEL, bConst = warriorSpellInfo.whirlwind3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind3.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.whirlwindBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },    
    {
        spellInfo = warriorSpellInfo.siegebreaker,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.siegebreaker.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.rampage,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 75, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    {
        spellInfo = warriorSpellInfo.bladestorm,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bladestorm.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL }
        }
    },
    {
        spellInfo = warriorSpellInfo.dragonRoar,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.dragonRoar.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL }
        }
    },
    -- NOTE: we don't included recklessness in the rotation -- it's a cooldown
    {
        spellInfo = warriorSpellInfo.execute,
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.furyMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furySuddenDeathBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.execute.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        _debugName = "bloodthirst_no_enrage",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.ragingBlow,
        tests = {
            { aType = HoJ.TYPE_SKILL_CHARGES, aConst = warriorSpellInfo.ragingBlow, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ragingBlow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    -- bloodthirst only if below minor heal health ratio
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.ragingBlow,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ragingBlow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    -- bloodthirst as a filler if it's available
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.charge,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    -- heroic throw if within range
    {
        spellInfo = warriorSpellInfo.heroicThrow,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    -- prefer whirlwind as a filler in multi-target situations, even if our target is out of range
    {
        spellInfo = warriorSpellInfo.whirlwind2,
        tests = {
            { aType = HoJ.TYPE_CONSTANT, aConst = true, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.whirlwind3,
        tests = {
            { aType = HoJ.TYPE_CONSTANT, aConst = true, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local furyInterruptRules = {
    _name = "furyInterruptRules",
    {
        spellInfo = warriorSpellInfo.pummel,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.stormBolt,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
}

local furyMitigationRules = {
    _name = "furyMitigationRules",
    -- emergency mitigations
    -- these are selected if:
    --   - the amount of damage predicted over the next 5 seconds exceeds the player's current health
    --   - predicted death time is < 10 seconds
    -- NOTE: non-instant healing is not considered a mitigation
    {
        spellInfo = warriorSpellInfo.ignorePain,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_SKILL_POWER_COST, bConst = warriorSpellInfo.ignorePain, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.enragedRegeneration,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.stoneform,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.rallyingCry,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.dragonRoar,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.piercingHowl,
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.berserkerRage,
        tests = {
            { aType = HoJ.TYPE_UNIT_IS_SLOWED, aConst = "player", bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local furyDpsCooldownRules = {
    _name = "furyDpsCooldownRules",
    {
        spellInfo = warriorSpellInfo.recklessness,
        tests = {
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 1, bType = HoJ.TYPE_SPELL_DURATION, bConst = warriorSpellInfo.recklessness, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_COMBAT_ATTACKS_HIT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_TOUGH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.recklessness, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

local furyCooldownSortPriorities = {
    _name = "furyCooldownSortPriorities",
    [warriorSpellInfo.ignorePain.id] = 12,
    [warriorSpellInfo.pummel.id] = 11,
    [warriorSpellInfo.stormBolt.id] = 10,
    [warriorSpellInfo.enragedRegeneration.id] = 9,
    [warriorSpellInfo.stoneform.id] = 8,
    [warriorSpellInfo.recklessness.id] = 7,
    [warriorSpellInfo.rallyingCry.id] = 6,
    [warriorSpellInfo.heroicLeap.id] = 5,
    [warriorSpellInfo.intimidatingShout.id] = 4,
    [warriorSpellInfo.dragonRoar.id] = 2,
    [warriorSpellInfo.piercingHowl.id] = 3,
    [warriorSpellInfo.berserkerRage.id] = 1,
}

local furyActiveBuffList = { 
    [warriorSpellInfo.furyEnrageBuff.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.furySuddenDeathBuff.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.whirlwindBuff.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.victorious.id] = HoJ:ConstructActiveBuff(),
}

-- this is the list of buffs that we want to show on the buffs bar
-- in general we just show buffs that proc randomly here, or buffs with a stack count that needs tracking
local furyProcdBuffList = {
    { _spellInfo = warriorSpellInfo.furyEnrageBuff },
    { _spellInfo = warriorSpellInfo.furySuddenDeathBuff },
    { _spellInfo = warriorSpellInfo.whirlwindBuff },
    { _spellInfo = warriorSpellInfo.victorious },
}

local furyActiveBlessingList = {
    [warriorSpellInfo.battleShout.id] = {
        spellInfo = warriorSpellInfo.battleShout,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
}

local furyBuffSortPriorities = {
    _name = "furyBuffSortPriorities",
    [warriorSpellInfo.victorious.id] = 5,
    [warriorSpellInfo.furyEnrageBuff.id] = 4,
    [warriorSpellInfo.furySuddenDeathBuff.id] = 3,
    [warriorSpellInfo.whirlwindBuff.id] = 2,
    [warriorSpellInfo.battleShout.id] = 1,
}

--------------------------------------------------------------------------------
-- PROTECTION
--------------------------------------------------------------------------------
local protRotationRules = {
    _name = "protRotationRules",
    -- 1) avatar on cooldown
    -- 2) demoralizing shout on cooldown (only with booming voice)
    -- 3) shield slam on cooldown
    -- 4) thunder clap on cooldown
    -- 5) revenge if free
    -- 6) devastate
    {
        spellInfo = warriorSpellInfo.victoryRush,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.protImpendingVictory, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.victoryRush.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.impendingVictory,
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.protImpendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.avatar,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.avatar.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.demoralizingShout,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.demoralizingShout.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.executeArmsProt,
        tests = {
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.20, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.executeArmsProt.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.protDragonRoar,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.protDragonRoar.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.shieldSlam,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.shieldSlam.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.thunderClap,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.thunderClap.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.revenge,
        _debugName = "revenge_if_free",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.revenge.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.revengeBuff, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.devastate,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.devastate.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        -- storm bolt if it's talented and within range
        spellInfo = warriorSpellInfo.protStormBolt,
        _rangeCheck = true,
    },
    {  
        spellInfo = warriorSpellInfo.charge,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicThrow,
        _rangeCheck = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
}

local protMitigationRules = {
    _name = "protMitigationRules",
    {
        spellInfo = warriorSpellInfo.lastStand,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
        }
    },
    {
        spellInfo = warriorSpellInfo.shieldWall,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 20.0, op = HoJ.OP_LESS },
        }
    },
    {
        spellInfo = warriorSpellInfo.ignorePain,
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 20.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.ignorePain, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.shieldBlock,
        tests = {
            { aType = HoJ.TYPE_SKILL_CHARGES, aConst = warriorSpellInfo.shieldBlock, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.shieldBlockBuff, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_PCT, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.25, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.shockwave,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.shockwave.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_PCT, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.25, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_NEAR_DEATH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 20.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {  
        spellInfo = warriorSpellInfo.intervene,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.intervene.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_FRIEND, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_PLAYER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            -- { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND }
        }
    },
}

local protInterruptRules = {
    {
        spellInfo = warriorSpellInfo.pummel,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.pummel.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.shockwave,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.shockwave.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.protStormBolt,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.protStormBolt.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.intimidatingShout.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local protCooldownSortPriorities = {
    _name = "protCooldownSortPriorities",
    [warriorSpellInfo.shieldWall.id] = 10,
    [warriorSpellInfo.lastStand.id] = 9,
    [warriorSpellInfo.ignorePain.id] = 8,
    [warriorSpellInfo.spellReflection.id] = 7,
    [warriorSpellInfo.shieldBlock.id] = 6,
    [warriorSpellInfo.pummel.id] = 5,
    [warriorSpellInfo.intervene.id] = 4,
    [warriorSpellInfo.shockwave.id] = 3,
    [warriorSpellInfo.protStormBolt.id] = 2,
    [warriorSpellInfo.intimidatingShout.id] = 1,
}

local protActiveBuffList = { 
    [warriorSpellInfo.revengeBuff.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.victorious.id] = HoJ:ConstructActiveBuff(),
    [warriorSpellInfo.shieldBlockBuff.id] = HoJ:ConstructActiveBuff(),
}

local protProcdBuffList = {
    { _spellInfo = warriorSpellInfo.revengeBuff },
    { _spellInfo = warriorSpellInfo.victorious },
    { _spellInfo = warriorSpellInfo.shieldBlockBuff },
}

local protActiveBlessingList = {
    [warriorSpellInfo.battleShout.id] = {
        spellInfo = warriorSpellInfo.battleShout,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
}

local protBuffSortPriorities = {
    _name = "protBuffSortPriorities",
    [warriorSpellInfo.shieldBlockBuff.id] = 4,
    [warriorSpellInfo.revengeBuff.id] = 3,
    [warriorSpellInfo.victorious.id] = 2,
    [warriorSpellInfo.battleShout.id] = 1,
}

HoJ[HoJ.CLASS_NAME_WARRIOR] = {
    _spellInfos = HoJ._warriorSpellInfos,
    _className = "WARRIOR",
    _defaultSpecId = HoJ.SPEC_WARRIOR_ARMS,
    _powerColor = { r = 0.81, g = 0.01, b = 0.01, a = 1.0 },
    [HoJ.SPEC_NAME_WARRIOR_ARMS] = {
        -- properties
        _name = HoJ._text["Arms"],
        _spec = HoJ.SPEC_WARRIOR_ARMS,
        _role = HoJ.ROLE_DPS,
        _enabled = true,
        _multiTargetRange = 8,
        _multiTargetThreshold = 1,
        _minorHealHealthRatio = 0.99,
        _majorHealHealthRatio = 0.99,
        _defaultPowerPerSec = 20, -- testing shows around 10 pps
        _autoPowerGen = true,
        _powerType = HoJ.POWER_TYPE_RAGE,
        _powerGainStep = 25,
        _powerMax = 100,
        _usesPowerPooling = false,
        -- functions
        _updateFn = HoJ.Warrior_Update,
        _updateRenderFn = nil,
        _resetFn = nil,
        _selectSingleOrMultiTargetFn = HoJ.Warrior_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        _checkComboPointsFn = nil,
        _checkPowerFn = HoJ.CheckPower,
        -- tables
        _meleeAoESpells = armsMeleeAoESpells,
        _activeBuffList = armsActiveBuffList,
        _procdBuffList = armsProcdBuffList,
        _activeBlessings = armsActiveBlessingList,
        _rotationRules = { _singleTarget = armsRotationRules, _multiTarget = armsMultiTargetRotationRules },
        _interruptRules = { _singleTarget = armsInterruptRules, _multiTarget = armsInterruptRules },
        _mitigationRules = { _singleTarget = armsMitigationRules, _multiTarget = armsMitigationRules },
        _dpsCooldownRules = { _singleTarget = armsDpsCooldownRules, _multiTarget = armsDpsCooldownRules },
        _cooldownSortPriorities = armsCooldownSortPriorities,
        _buffSortPriorities = armsBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = warriorSpellInfo.mortalStrike
            },
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = warriorSpellInfo.charge
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
        },
    },
    [HoJ.SPEC_NAME_WARRIOR_FURY] = {
        -- properties
        _name = HoJ._text["Fury"],
        _spec = HoJ.SPEC_WARRIOR_FURY,
        _role = HoJ.ROLE_DPS,
        _enabled = true,    -- addon enabled for this spec?
        _multiTargetRange = 8,
        _multiTargetThreshold = 1,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 10, -- testing with my 114 shows about 8 pps
        _autoPowerGen = false,
        _powerType = HoJ.POWER_TYPE_RAGE,
        _powerGainStep = 0,
        _powerMax = 100,
        _usesPowerPooling = false,
        -- functions
        _updateFn = HoJ.Warrior_Update,
        _updateRenderFn = nil,
        _resetFn = nil,
        _selectSingleOrMultiTargetFn = HoJ.Warrior_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        _checkComboPointsFn = nil,
        _checkPowerFn = HoJ.CheckPower,
        -- tables
        _meleeAoESpells = furyMeleeAoESpells,
        _activeBuffList = furyActiveBuffList,
        _procdBuffList = furyProcdBuffList,
        _activeBlessings = furyActiveBlessingList,
        _rotationRules = { _singleTarget = furyRotationRules, _multiTarget = furyMultiTargetRotationRules },
        _interruptRules = { _singleTarget = furyInterruptRules, _multiTarget = furyInterruptRules },
        _mitigationRules = { _singleTarget = furyMitigationRules, _multiTarget = furyMitigationRules },
        _dpsCooldownRules = { _singleTarget = furyDpsCooldownRules, _multiTarget = furyDpsCooldownRules },
        _cooldownSortPriorities = furyCooldownSortPriorities,
        _buffSortPriorities = furyBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = warriorSpellInfo.bloodthirst
            },
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = warriorSpellInfo.charge
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
        },          
    },
    [HoJ.SPEC_NAME_WARRIOR_PROTECTION] = {
        -- properties
        _name = HoJ._text["Protection"],
        _spec = HoJ.SPEC_WARRIOR_PROTECTION,
        _role = HoJ.ROLE_TANK,            
        _enabled = true,    -- enabled for this spec?
        _multiTargetRange = 8,
        _multiTargetThreshold = 1,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.99,
        _defaultPowerPerSec = 30,
        _autoPowerGen = false,
        _powerType = HoJ.POWER_TYPE_RAGE,
        _powerGainStep = 0,
        _powerMax = 100,
        _usesPowerPooling = false,
        -- functions
        _updateFn = HoJ.Warrior_Update,
        _updateRenderFn = nil,
        _resetFn = nil,
        _selectSingleOrMultiTargetFn = HoJ.Warrior_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        _checkComboPointsFn = nil,
        _checkPowerFn = HoJ.CheckPower,
        -- tables
        _meleeAoESpells = protMeleeAoESpells,
        _activeBuffList = protActiveBuffList,
        _procdBuffList = protProcdBuffList,
        _activeBlessings = protActiveBlessingList,
        _rotationRules = { _singleTarget = protRotationRules, _multiTarget = protRotationRules },
        _interruptRules = { _singleTarget = protInterruptRules, _multiTarget = protInterruptRules },
        _mitigationRules = { _singleTarget = protMitigationRules, _multiTarget = protMitigationRules },
        _dpsCooldownRules = { _singleTarget = nil, _multiTarget = nil },
        _cooldownSortPriorities = protCooldownSortPriorities,
        _buffSortPriorities = protBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = warriorSpellInfo.intervene
            },
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = warriorSpellInfo.pummel
            },
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = warriorSpellInfo.charge
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
        },
    }
}
    
-- -----------------------------------------------------------------------------
-- HoJ:LoadWarrior
-- -----------------------------------------------------------------------------
function HoJ:LoadWarrior()
    -- HoJ:Log( "LoadWarrior" )

    if ( HoJ._classId ~= HoJ.CLASS_ID_WARRIOR ) then
        HoJ:Error( "LoadWarrior called but _classId is not warrior!" )
    end

    HoJ.SpellInfo = HoJ._warriorSpellInfos
end

